//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
*******************************************************************************
* @file  sdcard.c
* @brief Implementation of the SDcard protocol
*
* Copyright (C) 2007 Parrot S.A.
*
* @author     florent.bayendrian@parrot.fr
* @date       2007-01-08
*******************************************************************************
*/

#include <stdio.h>
#include <cyg/kernel/kapi.h>
#include <cyg/io/sdcard.h>
#include <cyg/hal/hal_cache.h>
#include <cyg/utils/mmu.h>
#include <cyg/hal/hal_diag.h>

//At least P6/P6i controller need this buffer to be aligned on his size
cyg_uint8 sdcard_buffer[SDCARD_MAX_TRANSFER_SIZE]
        CYGBLD_ATTRIB_ALIGN(SDCARD_MAX_TRANSFER_SIZE);

#define SDCARD_R1_TYPE          (SDCARD_COMMAND_RESPONSE |      \
                                 SDCARD_COMMAND_ASK_CRC |       \
                                 SDCARD_COMMAND_ASK_OPCODE)
#define SDCARD_R1b_TYPE         (SDCARD_R1_TYPE |               \
                                 SDCARD_COMMAND_ASK_BUSY)
#define SDCARD_R2_TYPE          (SDCARD_COMMAND_RESPONSE |      \
                                 SDCARD_COMMAND_LONG_RESPONSE | \
                                 SDCARD_COMMAND_ASK_CRC)
#define SDCARD_R3_TYPE          (SDCARD_COMMAND_RESPONSE)
#define SDCARD_R4_TYPE          (SDCARD_COMMAND_RESPONSE)
#define SDCARD_R5_TYPE          (SDCARD_COMMAND_RESPONSE |      \
                                 SDCARD_COMMAND_ASK_CRC)
#define SDCARD_R5b_TYPE         (SDCARD_R5_TYPE |               \
                                 SDCARD_COMMAND_ASK_BUSY)
#define SDCARD_R6_TYPE          (SDCARD_COMMAND_RESPONSE |      \
                                 SDCARD_COMMAND_ASK_CRC |       \
                                 SDCARD_COMMAND_ASK_OPCODE)
#define SDCARD_R7_TYPE          (SDCARD_COMMAND_RESPONSE |      \
                                 SDCARD_COMMAND_ASK_CRC)

//Basic command list
//Reset (go idle state)
#define CMDINDEX_CMD_RESET              ( 0)
//Request OP condition (MMC)
#define CMDINDEX_CMD_REQUEST_OP_COND    ( 1 | SDCARD_R3_TYPE)
//Request CID (SD), ALL_SEND_CID (MMC)
#define CMDINDEX_CMD_REQUEST_ALL_CID    ( 2 | SDCARD_R2_TYPE)
//Request RCA (SD)
#define CMDINDEX_CMD_REQUEST_RCA        (3 | SDCARD_R6_TYPE)
//Set relative address (MMC)
#define CMDINDEX_CMD_SET_RCA            (3 | SDCARD_R1_TYPE)
//Switch function
#define CMDINDEX_CMD_SWITCH_FUNC_SD     (6 | SDCARD_R1_TYPE)
#define CMDINDEX_CMD_SWITCH_FUNC_MMC    (6 | SDCARD_R1b_TYPE)
//Select/Deselect card (SD)
#define CMDINDEX_CMD_SELECT_CARD        ( 7 | SDCARD_R1b_TYPE)
//Send interface condition
#define CMDINDEX_CMD_SEND_IF_CONDITION  ( 8 | SDCARD_R7_TYPE)
//Request CSD
#define CMDINDEX_CMD_REQUEST_CSD        ( 9 | SDCARD_R2_TYPE)
//Request CID
#define CMDINDEX_CMD_REQUEST_CID        (10 | SDCARD_R2_TYPE)
//Switch voltage
#define CMDINDEX_CMD_VOLTAGE_SWITCH     (11 | SDCARD_R1_TYPE)
//Stop transmission
#define CMDINDEX_CMD_STOP_TRANS         (12 | SDCARD_R1b_TYPE \
                                            | SDCARD_COMMAND_ASK_ABORT)
//HC request SDCard status
#define CMDINDEX_CMD_STATUS_REQUEST     (13 | SDCARD_R1_TYPE)
//Set the block length
#define CMDINDEX_CMD_SET_BLOCK_LEN      (16 | SDCARD_R1_TYPE)
//Read single block
#define CMDINDEX_CMD_READ_SINGLE        (17 | SDCARD_R1_TYPE \
                                            | SDCARD_COMMAND_WITH_DATA)
//Read multiple block
#define CMDINDEX_CMD_READ_MULT          (18 | SDCARD_R1_TYPE \
                                            | SDCARD_COMMAND_WITH_DATA)
//Set block count
#define CMDINDEX_CMD_SET_BLOCK_COUNT    (23 | SDCARD_R1_TYPE)
//Write single block
#define CMDINDEX_CMD_WRITE_SINGLE       (24 | SDCARD_R1_TYPE \
                                            | SDCARD_COMMAND_WITH_DATA)
//Write multiple block
#define CMDINDEX_CMD_WRITE_MULT         (25 | SDCARD_R1_TYPE \
                                            | SDCARD_COMMAND_WITH_DATA)
//Application specific command
#define CMDINDEX_CMD_APP_SPECIFIC       (55 | SDCARD_R1_TYPE)

//Must be precede by CMDINDEX_CMD_APP_SPECIFIC
//Set card bus width
#define CMDINDEX_ACMD_SEND_BUS_WIDTH            ( 6 | SDCARD_R1_TYPE)
//Request card status
#define CMDINDEX_ACMD_REQUEST_STATUS            (13 | SDCARD_R1_TYPE)
//Request the number of the written (without errors) blocks
#define CMDINDEX_ACMD_REQUEST_WRITTEN_BLOCKS    (22 | SDCARD_R1_TYPE)
//Set the number of write blocks to be pre-erased before writing
#define CMDINDEX_ACMD_PRE_ERASE_BLOCKS          (23 | SDCARD_R1_TYPE)
//Sends host capacity support information (HCS) and asks the accessed
//card to send its operating condition register (OCR)
#define CMDINDEX_ACMD_SEND_HC_CAPACITY          (41 | SDCARD_R3_TYPE)

#define CMD8_ARG_CHECK_PATTERN      (0xAA)
#define CMD8_ARG_VHS_3V3            (0x1 << 8)     // 2.7-3.6V
#define CMD8_ARG_VHS_1V8            (0x2 << 8)     // low voltage range

#define ACMD6_ARG_BUS_WIDTH_4BITS   (0x2)

#define ACMD41_ARG_HIGH_CAPACITY    (1<<30)
#define ACMD41_ARG_MAX_PERF         (1<<28)
#define ACMD41_ARG_S18R             (1<<24)
#define ACMD41_ARG_3V3              (3<<20)
#define ACMD41_ARG_LOW_VOLTAGE      (1<<7)

//VDD voltage window [2.7:3.6]
#define CMD1_VOLTAGE_3V3            (0xFF8000)
//VDD voltage window [1.65:1.95]
#define CMD1_VOLTAGE_1V8            (0x000080)
#define CMD1_ANSWER_BUSY            (1<<31)

#define ACMD41_ANSWER_BUSY          (1<<31)
#define ACMD41_ANSWER_CCS           (1<<30)
#define ACMD41_ANSWER_S18A          (1<<24)

//Call this before any lookup
static void
sdcard_init_default(cyg_sdcard_disk_info_t *sdcard_info)
{
    DEBUG3("sdcard_init_default\n");

    sdcard_info->sdcard_freq = SDCARD_CLK_SLOW;
    sdcard_info->sdcard_voltage = SDCARD_VOLTAGE_3V3;

    ll_sdcard_card_init(sdcard_info);
}


SDCARD_eError
sdcard_init(cyg_sdcard_disk_info_t *sdcard_info)
{
    SDCARD_eError err;

    DEBUG3("sdcard_init\n");

    err = ll_sdcard_hw_init(sdcard_info);

    if (err == SDCARD_ERROR_NONE) {

    }

    return err;
}


SDCARD_eError
sdcard_lookup(cyg_sdcard_disk_info_t *sdcard_info)
{
    cyg_uint32 cmd_param;
    cyg_uint32 answer[4];
    cyg_bool flag_f8 = false;

    //Long answer bits
    //135        104|103        72|71         40|39         8
    //   answer[3]  |  answer[2]  |  answer[1]  |  answer[0]

    //Short answer bits
    //31         0
    //  answer[0]

    SDCARD_eError err = SDCARD_ERROR_NOT_CONNECTED;
    cyg_bool switch_to_1_8v = false;

    DEBUG3("sdcard_lookup\n");

    //Restore default for getting connected sdcard info
    sdcard_init_default(sdcard_info);

    sdcard_info->sdcard_card_type = SDCARD_TYPE_NONE;

    //Reset card, some cards need this twice
    err = ll_sdcard_send_command(sdcard_info, CMDINDEX_CMD_RESET, 0, NULL);
    // some cards need CMD0 2 times ...
    ll_sdcard_send_command(sdcard_info, CMDINDEX_CMD_RESET, 0, NULL);

    if (err == SDCARD_ERROR_NONE) {
        cmd_param = CMD8_ARG_CHECK_PATTERN;
        if (sdcard_info->ctrl_voltage == SDCARD_VOLTAGE_1V8) {
            cmd_param |= CMD8_ARG_VHS_1V8;
        } else {
            cmd_param |= CMD8_ARG_VHS_3V3;
        }

        err = ll_sdcard_send_command(sdcard_info,
            CMDINDEX_CMD_SEND_IF_CONDITION, cmd_param, answer);
        if (err == SDCARD_ERROR_NONE) {
            //Vers 2.00 or later Memory card
            flag_f8 = true;
        } else if (err == SDCARD_ERROR_TIMEOUT) {
            //Vers 2.00 or later Memory card but voltage mismatch
            //Vers 1.x SD Memory card
            //Not a Memory card
            flag_f8  = false;
        } else {
            return err;
        }
    } else {
        DEBUG1("Bad answer to CMD RESET %d\n", err);
        return err;
    }


    if (flag_f8) {
        if ((answer[0] & CMD8_ARG_CHECK_PATTERN) != CMD8_ARG_CHECK_PATTERN) {
            return SDCARD_ERROR_HARDWARE;
        }
    } else {
        //Nothing
    }

    if ((cmd_param & CMD8_ARG_VHS_1V8) == CMD8_ARG_VHS_1V8) {
        cmd_param = ACMD41_ARG_S18R | ACMD41_ARG_LOW_VOLTAGE;
    } else {
        cmd_param = ACMD41_ARG_3V3;
    }
    cmd_param |= ACMD41_ARG_HIGH_CAPACITY;
    cmd_param |= ACMD41_ARG_MAX_PERF;

    //Check answer validity
    do {
        err = ll_sdcard_send_command(sdcard_info,
            CMDINDEX_CMD_APP_SPECIFIC, 0, answer);

        if (err == SDCARD_ERROR_TIMEOUT) {
            //Maybe a MMC card
            sdcard_info->sdcard_card_type = SDCARD_TYPE_MMC;
            sdcard_info->sdcard_freq = SDCARD_CLK_MMC_FULL;
            DEBUG2("Detect MMC card or No card\n");
        } else {
            err = ll_sdcard_send_command(sdcard_info,
                CMDINDEX_ACMD_SEND_HC_CAPACITY, cmd_param, answer);

            if (err == SDCARD_ERROR_TIMEOUT) {
                //Maybe a MMC card
                sdcard_info->sdcard_card_type = SDCARD_TYPE_MMC;
                sdcard_info->sdcard_freq = SDCARD_CLK_MMC_FULL;
                DEBUG2("Detect MMC card\n");
            } else if (err == SDCARD_ERROR_NONE) {
                //Answer is ready
                if (answer[0] & ACMD41_ANSWER_BUSY) {
                    if (flag_f8 && (answer[0] & ACMD41_ANSWER_CCS)) {
                            //High capacity card
                            sdcard_info->sdcard_card_type = SDCARD_TYPE_SDHC;
                            sdcard_info->sdcard_freq = SDCARD_CLK_SD_HIGH;
                            DEBUG2("Detect SDHC card\n");
                    } else {
                        //Standart capacity card
                        sdcard_info->sdcard_card_type = SDCARD_TYPE_SDCARD;
                        sdcard_info->sdcard_freq = SDCARD_CLK_SD_FULL;
                        DEBUG2("Detect SD card\n");
                    }

                    if (answer[0] & ACMD41_ANSWER_S18A) {
                        switch_to_1_8v = true;
                    }
                    break;
                } else {
                    //Card busy, wait for max 1s
                    static cyg_uint32 timeout = 20;
                    HAL_DELAY_US(50000);
                    if (timeout-- == 0) {
                        err = SDCARD_ERROR_TIMEOUT;
                        timeout = 20;
                    }
                }
            }
        }
    } while (err == SDCARD_ERROR_NONE);

    if (sdcard_info->sdcard_card_type == SDCARD_TYPE_MMC) {
        cmd_param = CMD1_VOLTAGE_3V3;

        if (sdcard_info->ctrl_voltage == SDCARD_VOLTAGE_1V8) {
            cmd_param |= CMD1_VOLTAGE_1V8;
        }

        do {
            err = ll_sdcard_send_command(sdcard_info,
                CMDINDEX_CMD_REQUEST_OP_COND, cmd_param, answer);

            if (err == SDCARD_ERROR_NONE) {
                //Answer is ready
                if (answer[0] & CMD1_ANSWER_BUSY) {
                    if (answer[0] & CMD1_VOLTAGE_1V8) {
                        switch_to_1_8v = true;
                    }
                    break;
                } else {
                    //Card busy
                }
            } else if (err == SDCARD_ERROR_TIMEOUT) {
                sdcard_info->sdcard_card_type = SDCARD_TYPE_NONE;
            }
        } while (err == SDCARD_ERROR_NONE);
    }

    if ((err == SDCARD_ERROR_NONE) && switch_to_1_8v) {
        cmd_param = 0;
        err = ll_sdcard_send_command(sdcard_info, CMDINDEX_CMD_VOLTAGE_SWITCH,
            cmd_param, answer);
        if (err == SDCARD_ERROR_NONE) {
            //Card is now in 1.8V
            //Change host controller to 1.8V
            //For now host controller does not change voltage
        }

        //It's not an issue if the card does not change voltage
        err = SDCARD_ERROR_NONE;
    }

    if (err == SDCARD_ERROR_NONE) {
        cmd_param = 0;
        err = ll_sdcard_send_command(sdcard_info, CMDINDEX_CMD_REQUEST_ALL_CID,
                    cmd_param, answer);

        if (err == SDCARD_ERROR_NONE) {
            sdcard_info->sdcard_serial_number =
                (answer[1] << 16) | (answer[0] >> 16);
        } else {
            sdcard_info->sdcard_serial_number = 0;
        }
    }

    if (err == SDCARD_ERROR_NONE) {
        if (sdcard_info->sdcard_card_type != SDCARD_TYPE_MMC) {
            cmd_param = 0;
            err = ll_sdcard_send_command(sdcard_info, CMDINDEX_CMD_REQUEST_RCA,
                    cmd_param, answer);

            if (err == SDCARD_ERROR_NONE) {
                sdcard_info->sdcard_rca = (answer[0] >> 16) & 0xFFFF;
            }
        } else {
            sdcard_info->sdcard_rca = 1; //Any number could be anythings
            cmd_param = sdcard_info->sdcard_rca << 16;
            err = ll_sdcard_send_command(sdcard_info, CMDINDEX_CMD_SET_RCA,
                    cmd_param, answer);

            if (err != SDCARD_ERROR_NONE) {
                sdcard_info->sdcard_rca = 0;
            }
        }
    }

    if (err == SDCARD_ERROR_NONE) {
        cmd_param = sdcard_info->sdcard_rca << 16;
        err = ll_sdcard_send_command(sdcard_info, CMDINDEX_CMD_REQUEST_CSD,
                cmd_param, answer);

        if (err == SDCARD_ERROR_NONE) {
            //answer[0] -> csd_structure[ 39:  8]
            //answer[1] -> csd_structure[ 71: 40]
            //answer[2] -> csd_structure[103: 72]
            //answer[3] -> csd_structure[127:104]
            cyg_uint8 csd_structure = (answer[3] >> 22) & 0x3;
            cyg_uint8 tran_speed = answer[2] >> 24;

            if ((csd_structure == 0) ||
                (sdcard_info->sdcard_card_type == SDCARD_TYPE_MMC)) {
                //CSD structure v1.0, Standart capacity
                cyg_uint32 c_size, c_size_mult, read_bl_len;
                //         [73:62] [49:47]      [83:80]

                c_size_mult = (answer[1] >> 7) & 0x7;
                read_bl_len = (answer[2] >> 8) & 0xF;
                c_size = (answer[1] >> 22) & 0x3FF;
                c_size |= (answer[2] & 0x3) << 10;

                sdcard_info->sdcard_block_count = ((c_size + 1) <<
                    (c_size_mult + read_bl_len + 2)) / SDCARD_BLOCK_SIZE;
            } else if (csd_structure == 1) {
                //CSD structure v2.0, High and Extended capacity
                cyg_uint32 c_size;
                //         [69:48]
                c_size = (answer[1] >> 8) & 0x3FFFFF;

                sdcard_info->sdcard_block_count = (c_size + 1) *
                    (512 * 1024) / SDCARD_BLOCK_SIZE;
            } else {
                //Not supported
                sdcard_info->sdcard_block_count = 0;
            }

            //Check card supported speed
            if ((sdcard_info->sdcard_freq > SDCARD_CLK_SD_FULL) &&
                (tran_speed != 0x5A)) {
                DEBUG1("SD does not support max speed\n");
                sdcard_info->sdcard_freq = SDCARD_CLK_SD_FULL;
            }

        } else {
            sdcard_info->sdcard_block_count = 0;
        }
    }

    if (err == SDCARD_ERROR_NONE) {
        cmd_param = sdcard_info->sdcard_rca << 16;
        err = ll_sdcard_send_command(sdcard_info, CMDINDEX_CMD_SELECT_CARD,
                cmd_param, answer);
    }

    if (err == SDCARD_ERROR_NONE) {
        //Set bus width and frequency
        if (sdcard_info->sdcard_card_type != SDCARD_TYPE_MMC) {
            cmd_param = sdcard_info->sdcard_rca << 16;
            err = ll_sdcard_send_command(sdcard_info,
                CMDINDEX_CMD_APP_SPECIFIC, cmd_param, answer);

            cmd_param = ACMD6_ARG_BUS_WIDTH_4BITS;
            err = ll_sdcard_send_command(sdcard_info,
                CMDINDEX_ACMD_SEND_BUS_WIDTH, cmd_param, answer);
        }
    }

    if (err == SDCARD_ERROR_NONE) {
        ll_sdcard_card_init(sdcard_info);
    }

    return err;
}


SDCARD_eError
sdcard_read(cyg_sdcard_disk_info_t *sdcard_info, cyg_uint8 *buffer,
    cyg_uint32 block, cyg_uint32 *len)
{
    SDCARD_eError err = SDCARD_ERROR_NONE;
    cyg_uint32 block_nb = *len;
    CYG_ADDRESS paddr = arm_mmu_v2p((CYG_ADDRESS)sdcard_buffer);

    DEBUG3("sdcard_read\n");

    while (block_nb > 0) {
        cyg_uint32 answer[4];
        cyg_uint32 cmd, read_block, data_address;

        read_block = CYG_MIN(SDCARD_MAX_TRANSFER_SIZE / SDCARD_BLOCK_SIZE,
            block_nb);

        //Invalidate cache will be filled by a DMA request
        arm_mmu_invalidate_dcache_range((CYG_ADDRESS)sdcard_buffer,
            read_block * SDCARD_BLOCK_SIZE);

        if (read_block == 1) {
            //Start read card (single block mode)
            cmd = CMDINDEX_CMD_READ_SINGLE;
        } else {
            //Start read card (multi block mode)
            cmd = CMDINDEX_CMD_READ_MULT;
        }

        err = ll_sdcard_receive_data(sdcard_info, paddr, read_block);
        if (err != SDCARD_ERROR_NONE) {
            DEBUG1("sdcard_read, receive data failed %d\n", err);
            break;
        }

        // SDHC and SDXC Cards use block unit address
        // SDSC Card uses byte unit address
        data_address = sdcard_info->sdcard_card_type ==
            SDCARD_TYPE_SDHC ? block : block * SDCARD_BLOCK_SIZE;

        err = ll_sdcard_send_command(sdcard_info, cmd, data_address, answer);
        if (err == SDCARD_ERROR_NONE) {
            block_nb -= read_block;
            block += read_block;
            memcpy(buffer, sdcard_buffer, read_block * SDCARD_BLOCK_SIZE);
            buffer += read_block * SDCARD_BLOCK_SIZE;
        } else {
            DEBUG1("sdcard_read, send_command failed %d\n", err);
            block_nb = 0;
        }

        if (cmd == CMDINDEX_CMD_READ_MULT) {
            //Stop transfert
            ll_sdcard_send_command(sdcard_info, CMDINDEX_CMD_STOP_TRANS,
                0, answer);
        }
    }

    //return the number of block read
    *len = *len - block_nb;

    return err;
}


SDCARD_eError
sdcard_write(cyg_sdcard_disk_info_t *sdcard_info, const cyg_uint8 *buffer,
    cyg_uint32 block, cyg_uint32 *len)
{
    SDCARD_eError err = SDCARD_ERROR_NONE;
    cyg_uint32 block_nb = *len, cmd_param;
    cyg_uint32 answer[4];
    CYG_ADDRESS paddr = arm_mmu_v2p((CYG_ADDRESS)sdcard_buffer);

    DEBUG3("sdcard_write\n");

    while (block_nb > 0) {
        cyg_uint32 cmd, write_block, data_address;

        // SDHC and SDXC Cards use block unit address
        // SDSC Card uses byte unit address
        data_address = sdcard_info->sdcard_card_type ==
            SDCARD_TYPE_SDHC ? block : block * SDCARD_BLOCK_SIZE;

        write_block = CYG_MIN(SDCARD_MAX_TRANSFER_SIZE / SDCARD_BLOCK_SIZE,
            block_nb);

        //Update the buffer
        memcpy(sdcard_buffer, buffer, write_block * SDCARD_BLOCK_SIZE);
        arm_mmu_clean_dcache_range((CYG_ADDRESS)sdcard_buffer,
            write_block * SDCARD_BLOCK_SIZE);

        if (write_block == 1) {
            cmd = CMDINDEX_CMD_WRITE_SINGLE;
        } else {
            cmd = CMDINDEX_CMD_WRITE_MULT;
            if (sdcard_info->sdcard_card_type == SDCARD_TYPE_MMC) {
                //Set block count
                cmd_param = write_block;
                err = ll_sdcard_send_command(sdcard_info,
                    CMDINDEX_CMD_SET_BLOCK_COUNT, cmd_param, answer);
            } else {
                cmd_param = sdcard_info->sdcard_rca << 16;
                err = ll_sdcard_send_command(sdcard_info,
                    CMDINDEX_CMD_APP_SPECIFIC, cmd_param, answer);

                if (err == SDCARD_ERROR_NONE) {
                    // pre erase command
                    cmd_param = write_block;
                    err = ll_sdcard_send_command(sdcard_info,
                        CMDINDEX_ACMD_PRE_ERASE_BLOCKS, cmd_param, answer);
                }
            }
        }

        if (err == SDCARD_ERROR_NONE) {
            err = ll_sdcard_send_data(sdcard_info, paddr, write_block);
            if (err != SDCARD_ERROR_NONE) {
                DEBUG1("sdcard_write, send data failed %d\n", err);
                break;
            }
        }

        if (err == SDCARD_ERROR_NONE) {
            err = ll_sdcard_send_command(sdcard_info, cmd,
                data_address, answer);
        }

        //Stop transfert
        if (write_block > 1) {
            ll_sdcard_send_command(sdcard_info, CMDINDEX_CMD_STOP_TRANS,
                0, answer);
        }

        if (err == SDCARD_ERROR_NONE) {
            block_nb -= write_block;
            block += write_block;
            buffer += write_block * SDCARD_BLOCK_SIZE;
        } else {
            DEBUG1("sdcard_write, send command failed %d\n", err);
            break;
        }
    }

    //return the number of block written
    *len = *len - block_nb;

    return err;
}


SDCARD_eError
sdcard_state(cyg_sdcard_disk_info_t *sdcard_info, SDCARD_eState *state)
{
    SDCARD_eError err = SDCARD_ERROR_NONE;
    cyg_uint32 cmd_param;
    cyg_uint32 answer[4];

    DEBUG3("sdcard_state\n");

    //By default we don't know
    *state = SDCARD_STATE_UNKNOW;

    cmd_param = sdcard_info->sdcard_rca << 16;
    err = ll_sdcard_send_command(sdcard_info, CMDINDEX_CMD_SELECT_CARD,
        cmd_param, answer);

    if (err == SDCARD_ERROR_NONE) {
        *state = (answer[0] >> 9) & SDCARD_STATE_MASK;
    }

    return err;
}
