//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
*******************************************************************************
* @file  sdcard_block.c
* @brief SDcard eCos interface
*
* Copyright (C) 2007 Parrot S.A.
*
* @author     florent.bayendrian@parrot.fr
* @date       2007-01-08
*******************************************************************************
*/

#include <stdio.h>
#include <errno.h>
#include <cyg/io/io.h>
#include <cyg/io/devtab.h>
#include <cyg/io/disk.h>
#include <cyg/io/sdcard.h>
#include <cyg/infra/diag.h>
#include <cyg/hal/hal_diag.h>
#include <cyg/infra/cyg_type.h>
#include <assert.h>

static cyg_mutex_t sdcard_block_mutex;

static Cyg_ErrNo
sdcard_block_disk_read(disk_channel *chan, void *buf, cyg_uint32 len,
    cyg_uint32 block_num)
{
    cyg_sdcard_disk_info_t *disk = (cyg_sdcard_disk_info_t*)chan->dev_priv;
    Cyg_ErrNo ret = ENOERR;
    SDCARD_eError err = SDCARD_ERROR_NONE;

    DEBUG3("sdcard_block_disk_read: block_num = %d, len = %d\n",
           block_num, len);

    if (block_num >= disk->sdcard_block_count)
        return -EINVAL;

    cyg_mutex_lock(&sdcard_block_mutex);

    if (disk->sdcard_connected) {
        err = sdcard_read(disk, (cyg_uint8*)buf, block_num, &len);
    } else {
        err = SDCARD_ERROR_NOT_CONNECTED;
    }

    if (err != SDCARD_ERROR_NONE) {
        DEBUG1("%d) Read error %d\n", disk->device_type, err);
        if (disk->device_type == SDCARD_DEVICE_SDCARD) {
            // The card is gone
            DEBUG1("Disabling access to device %d\n", disk->device_type);
            disk->sdcard_connected = false;
            (*chan->callbacks->disk_disconnected)(chan);
            ret = -ENODEV;
        } else {
            ret = -EIO;
        }
    }

    cyg_mutex_unlock(&sdcard_block_mutex);

    return ret;
}

static Cyg_ErrNo
sdcard_block_disk_write(disk_channel *chan, const void *buf, cyg_uint32 len,
    cyg_uint32 block_num)
{
    cyg_sdcard_disk_info_t * disk = (cyg_sdcard_disk_info_t *) chan->dev_priv;
    Cyg_ErrNo ret = ENOERR;
    SDCARD_eError err = SDCARD_ERROR_NONE;
    int retry_count = 10;

    DEBUG3("sdcard_block_disk_write: block_num = %d, len = %d\n",
           block_num, len);

    if (block_num >= disk->sdcard_block_count)
        return -EINVAL;

    cyg_mutex_lock(&sdcard_block_mutex);

    if (disk->sdcard_connected) {
        cyg_uint32 blocks_written = len;
        cyg_uint8 *data = (cyg_uint8*)buf;

        while ((len > 0) && (retry_count > 0)) {
            blocks_written = len;
            err = sdcard_write(disk, data, block_num, &blocks_written);

            if (len > blocks_written)
                len -= blocks_written;
            else
                len = 0;
            data += (blocks_written * SDCARD_BLOCK_SIZE);
            block_num += blocks_written;

            if (err == SDCARD_ERROR_TIMEOUT) {
                retry_count--;
                cyg_thread_delay(1);
            } else if (err != SDCARD_ERROR_NONE) {
                break;
            } else {
                retry_count = 10;
            }
        };
    } else {
        err = SDCARD_ERROR_NOT_CONNECTED;
    }

    if (err != SDCARD_ERROR_NONE) {
        DEBUG1("%d) Write error %d\n", disk->device_type, err);
        if (disk->device_type == SDCARD_DEVICE_SDCARD) {
            // The card is gone
            DEBUG1("Disabling access to device %d\n", disk->device_type);
            disk->sdcard_connected = false;
            (*chan->callbacks->disk_disconnected)(chan);
            ret = -ENODEV;
        } else {
            ret = -EIO;
        }
    }

    cyg_mutex_unlock(&sdcard_block_mutex);

    return ret;
}

static Cyg_ErrNo
sdcard_block_disk_get_config(disk_channel *chan, cyg_uint32 key,
    const void *xbuf, cyg_uint32 *len)
{
    Cyg_ErrNo result = ENOERR;
    cyg_sdcard_disk_info_t *disk = (cyg_sdcard_disk_info_t*)chan->dev_priv;

    DEBUG3("sdcard_block_disk_get_config\n");

    cyg_mutex_lock(&sdcard_block_mutex);

    switch(key) {
        case CYG_IO_GET_CONFIG_DISK_SERIAL_NUMBER:
            if (disk->sdcard_connected) {
                CYG_ASSERT(xbuf != NULL, "Transfert buffer is NULL");
                *((cyg_uint32*) xbuf) = disk->sdcard_serial_number;
            } else {
                result = -ENODEV;
            }
            break;

        default:
            result = -EINVAL;
            break;
    }

    cyg_mutex_unlock(&sdcard_block_mutex);

    return result;
}

static Cyg_ErrNo
sdcard_block_disk_set_config(disk_channel *chan, cyg_uint32 key,
    const void *xbuf, cyg_uint32 *len)
{
    Cyg_ErrNo result = ENOERR;
    cyg_sdcard_disk_info_t *disk = (cyg_sdcard_disk_info_t*)chan->dev_priv;

    DEBUG3("sdcard_block_disk_set_config\n");

    cyg_mutex_lock(&sdcard_block_mutex);

    switch(key) {
        case CYG_IO_SET_CONFIG_DISK_MOUNT:
        // There will have been a successful lookup(), so there's
        // little point in checking the disk again.
        break;

        case CYG_IO_SET_CONFIG_DISK_UMOUNT:
            if (chan->info->mounts == 0) {
                // If this is the last unmount of the card, mark it as
                // disconnected. If the user then removes the card and
                // plugs in a new one everything works cleanly.
                disk->sdcard_connected = false;
                result = (chan->callbacks->disk_disconnected)(chan);
                DEBUG1("umounted\n");
            }
            break;

        default:
            result = -EINVAL;
            break;
    }

    cyg_mutex_unlock(&sdcard_block_mutex);

    return result;
}

static cyg_bool
sdcard_block_disk_init(struct cyg_devtab_entry *tab)
{
    cyg_bool result;
    SDCARD_eError err;

    disk_channel *chan = (disk_channel*)tab->priv;
    cyg_sdcard_disk_info_t* disk = (cyg_sdcard_disk_info_t*) chan->dev_priv;

    DEBUG1("sdcard_block_disk_init\n");
    cyg_mutex_init(&sdcard_block_mutex);

    //Init sdcard and low level
    err = sdcard_init(disk);

    if (err == SDCARD_ERROR_NONE) {
        cyg_mutex_lock(&sdcard_block_mutex);
        result = (*chan->callbacks->disk_init)(tab);
        cyg_mutex_unlock(&sdcard_block_mutex);
    } else {
        result = false;
    }

    return result;
}

static Cyg_ErrNo
sdcard_block_disk_lookup(struct cyg_devtab_entry **tab,
                   struct cyg_devtab_entry *sub_tab,
                   const char *name)
{
    disk_channel* chan = (disk_channel*) (*tab)->priv;
    cyg_sdcard_disk_info_t* disk = (cyg_sdcard_disk_info_t*) chan->dev_priv;
    SDCARD_eError err = SDCARD_ERROR_NONE;
    Cyg_ErrNo ret = ENOERR;

    DEBUG2("sdcard_block_disk_lookup\n");

    if (disk->sdcard_connected) {
        SDCARD_eState state;
        // There was a card plugged in last time we looked. Is it still there?
        err = sdcard_state(disk, &state);

        if ((err != SDCARD_ERROR_NONE) || (state != SDCARD_STATE_IDLE)) {
            // The old card is gone. Either there is no card plugged in, or
            // it has been replaced with a different one. If the latter the
            // existing mounts must be removed before anything sensible
            // can be done
            disk->sdcard_connected = false;
            (*chan->callbacks->disk_disconnected)(chan);
            if (0 != chan->info->mounts) {
                ret = -ENODEV;
            }
        } else {
            ret = (*chan->callbacks->disk_lookup)(tab, sub_tab, name);
            return ret;
        }
    }

    if ((chan->info->mounts != 0) && !disk->sdcard_connected) {
        // There are still mount points to an old card. We cannot accept
        // new mount requests until those have been cleaned out.
        ret = -ENODEV;
    }

    if (ret == ENOERR) {
        err = sdcard_lookup(disk);
        if (err == SDCARD_ERROR_NONE) {
            DEBUG1("Found a %d MB Card\n",
                   disk->sdcard_block_count * SDCARD_BLOCK_SIZE >> 20);
            disk->sdcard_connected = true;
        } else if (err == SDCARD_ERROR_TIMEOUT) {
            DEBUG2("No SDcard found\n");
            ret = -ENODEV;
        } else {
            DEBUG2("sdcard_block_disk_lookup: failed with sdcard_lookup()\n");
            ret = -ENODEV;
        }
    }

    if ((ret == ENOERR) &&
        (err == SDCARD_ERROR_NONE) &&
        (disk->sdcard_block_count > 0)) {
        cyg_disk_identify_t ident;

        sprintf(ident.serial, "%u", disk->sdcard_serial_number);
        ident.firmware_rev[0]   = '\0';
        ident.model_num[0]      = '\0';
        ident.cylinders_num     = 0; // not used
        ident.heads_num         = 0; // not used
        ident.sectors_num       = 0; // not used
        ident.lba_sectors_num   = disk->sdcard_block_count;
        ident.phys_block_size   = SDCARD_BLOCK_SIZE;
        ident.max_transfer      = SDCARD_MAX_TRANSFER_SIZE/SDCARD_BLOCK_SIZE;

        DEBUG3("Calling disk_connected(): serial %s, lba_sectors_num %d\n",
            ident.serial, ident.lba_sectors_num);
        (*chan->callbacks->disk_connected)(*tab, &ident);

        // We now have a valid card and higher-level code knows about it
        ret = (*chan->callbacks->disk_lookup)(tab, sub_tab, name);
    } else {
        ret = -ENODEV;
    }

    DEBUG3("sdcard_block_disk_lookup return %d\n", ret);

    return ret;
}

#if 0
static cyg_sdcard_disk_info_t cyg_sdcard_disk0_info_polling = {
    .sdcard_connected   = false,
    .device_type        = SDCARD_DEVICE_SDCARD,
    .ctrl_voltage       = SDCARD_VOLTAGE_3V3,
    .driver_mode        = SDCARD_DRIVERMODE_POLLING,
    .priv               = NULL
};
#else
static cyg_sdcard_disk_info_t cyg_sdcard_disk0_info_interrupt = {
    .sdcard_connected   = false,
    .device_type        = SDCARD_DEVICE_SDCARD,
    .ctrl_voltage       = SDCARD_VOLTAGE_3V3,
    .driver_mode        = SDCARD_DRIVERMODE_INTERRUPT,
    .priv               = NULL
};
#endif

DISK_FUNS(cyg_sdcard_block_funs,
          sdcard_block_disk_read,
          sdcard_block_disk_write,
          sdcard_block_disk_get_config,
          sdcard_block_disk_set_config);

#if 0
//Polling
DISK_CONTROLLER(cyg_sdcard_disk_controller,
                cyg_sdcard_disk0_info_polling.priv);

DISK_CHANNEL(cyg_sdcard_disk0_channel,
             cyg_sdcard_block_funs,
             cyg_sdcard_disk0_info_polling,
             cyg_sdcard_disk_controller,
             true,
             1);

BLOCK_DEVTAB_ENTRY(cyg_sdcard_block0_devtab_entry,
                   CYGPKG_DEVS_DISK_GENERIC_SDCARD_NAME,
                   0,
                   &cyg_io_disk_devio,
                   sdcard_block_disk_init,
                   sdcard_block_disk_lookup,
                   &cyg_sdcard_disk0_channel);

#else
//Interrupt
DISK_CONTROLLER(cyg_sdcard_disk_controller_int,
                cyg_sdcard_disk0_info_interrupt.priv);

DISK_CHANNEL(cyg_sdcard_disk0_channel_int,
             cyg_sdcard_block_funs,
             cyg_sdcard_disk0_info_interrupt,
             cyg_sdcard_disk_controller_int,
             true,
             1);

BLOCK_DEVTAB_ENTRY(cyg_sdcard_block0_devtab_entry_int,
                   CYGPKG_DEVS_DISK_GENERIC_SDCARD_NAME,
                   0,
                   &cyg_io_disk_devio,
                   sdcard_block_disk_init,
                   sdcard_block_disk_lookup,
                   &cyg_sdcard_disk0_channel_int);
#endif

#if defined(CYGPKG_DEVS_DISK_GENERIC_SDCARD_NAME_INAND)
static cyg_sdcard_disk_info_t cyg_sdcard_disk1_info = {
    .sdcard_connected   = false,
    .device_type        = SDCARD_DEVICE_INAND,
    .ctrl_voltage       = SDCARD_VOLTAGE_3V3
    .driver_mode        = SDCARD_DRIVERMODE_INTERRUPT,
    .priv               = NULL
};

DISK_CHANNEL(cyg_sdcard_disk1_channel,
             cyg_sdcard_block_funs,
             cyg_sdcard_disk1_info,
             cyg_sdcard_disk_controller,
             true,
             4);

BLOCK_DEVTAB_ENTRY(cyg_sdcard_block1_devtab_entry,
                   CYGPKG_DEVS_DISK_GENERIC_SDCARD_NAME_INAND,
                   0,
                   &cyg_io_disk_devio,
                   sdcard_block_disk_init,
                   sdcard_block_disk_lookup,
                   &cyg_sdcard_disk1_channel);
#endif
