//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
*******************************************************************************
* @file  mci.c
* @brief SDcard low level driver
*
* Copyright (C) 2007 Parrot S.A.
*
* @author     florent.bayendrian@parrot.fr
* @date       2007-01-08
*******************************************************************************
*/

#include <cyg/hal/cpu_types.h>
#include <cyg/hal/drv_api.h>
#include <cyg/hal/hal_io.h>
#include <cyg/kernel/kapi.h>
#include <cyg/io/sdcard.h>
#include <assert.h>
#include <cyg/hal/hal_diag.h>

#include "mci.h"
#include <pkgconf/devs_disk_sdcard_parrot5.h>

#if defined(CYGHWR_HAL_CUSTOM_CONFIG_CK5300P)
#include <cyg/io/gpio.h>
#endif

#define MCI_ISR_PRIORITY    (4)
#define MCI0_INTERRUPT      (5)
#define MCI1_INTERRUPT      (6)

#define read_mci_uint32(_ctrl_, _off_, _val_)               \
    HAL_READ_UINT32(_ctrl_->controller_base + _off_, _val_)

#define write_mci_uint32(_ctrl_, _off_, _val_)              \
    HAL_WRITE_UINT32(_ctrl_->controller_base + _off_, _val_)

#define read_dma_uint32(_off_, _val_)               \
    HAL_READ_UINT32(P5P_DMA_BASE_ADDRESS + _off_, _val_)

#define write_dma_uint32(_off_, _val_)              \
    HAL_WRITE_UINT32(P5P_DMA_BASE_ADDRESS + _off_, _val_)

/*
 * Description
 * Clock speed dividers
 */
#define MCI_MCLK_SPEED          (20800000)          /* 20.8 MHz */
#define MCI_TRANSFERT_TIMEOUT   ((cyg_uint64)250)   /* 250 ms */

typedef struct MCI_privData {
    cyg_uint32  mci_datatimer; //Value to write to MCI_REG_DATATIMER
    cyg_uint32  mci_datactrl;  //Value to write to MCI_REG_DATACTRL
    cyg_uint32  mci_freq;
    cyg_sem_t   mci_sem_done;
    cyg_uint32  mci_interrupt;
    cyg_bool    mci_data_transfert;
    cyg_uint32  (*wait_end)(cyg_sdcard_disk_info_t *sdcard_info);
    cyg_bool    mci_inhibit_crc;
} MCI_privData;

static MCI_privData g_privData;

// interrupt handlers
static cyg_handle_t     sdcard_int_handle;
static cyg_interrupt    intr;
static cyg_uint32       mci_isr(cyg_vector_t vector, cyg_addrword_t data);
static void             mci_dsr(cyg_vector_t vector, cyg_ucount32 count,
                                cyg_addrword_t data);

#define COPY32_TO_MCI_FIFO(__data, __ctrl)                  \
    asm volatile (                                          \
        "ldmia %0,{r0-r7}\n\t"                              \
        "stmia %1,{r0-r7}\n"                                \
        :                                                   \
        : "r" ((__data)),                                   \
          "r" ((__ctrl->controller_base + MCI_REG_FIFO))    \
        : "r0","r1","r2","r3","r4","r5","r6","r7" );

#define COPY32_FROM_MCI_FIFO(__data, __ctrl)                \
    asm volatile (                                          \
        "ldmia %1,{r0-r7}\n\t"                              \
        "stmia %0,{r0-r7}\n"                                \
        :                                                   \
        : "r" ((__data)),                                   \
          "r" ((__ctrl->controller_base + MCI_REG_FIFO))    \
        : "r0","r1","r2","r3","r4","r5","r6","r7" );

static void
mci_dma_init(void)
{
    // Enable DMA Clock on the system controller
    HAL_WRITE_UINT32(PARROT5_SYS + _SYS_CEN, SYS_CEN_DMACLK);

    // Enable DMA Operation
    write_dma_uint32(P5P_DMA_REG_CONFIG, 1);
}

static void
mci_dma_config(cyg_sdcard_disk_info_t *sdcard_info, CYG_ADDRESS phys_addr,
    cyg_uint32 nb_block, cyg_bool read_nWrite)
{
    cyg_uint32 reg_dma;
    cyg_uint32 reg_ctrl, reg_cfg, reg_src, reg_dst;

    //Check if DMA is available
    read_dma_uint32(P5P_DMA_REG_ENBLDCHNS, reg_dma);
    if (reg_dma & (1 << MCI_DMA_CHANNEL)) {
        // DMA is enable, disable it for the next transfer
        write_dma_uint32(P5P_DMA_REG_CHX_CONFIG, 0);
    }

    reg_ctrl = P5P_DMA_CH_CTRL_IT           |
        P5P_DMA_CH_CTRL_SRC_WIDTH_WORD      |
        P5P_DMA_CH_CTRL_DST_WIDTH_WORD      |
        P5P_DMA_CH_CTRL_SRC_BURST_SIZE_8    |
        P5P_DMA_CH_CTRL_DST_BURST_SIZE_8;

    //Set transfert size
    reg_ctrl |= (nb_block * SDCARD_BLOCK_SIZE / sizeof(cyg_uint32));

    reg_cfg = P5P_DMA_CH_CONFIG_ENABLE;

    if (read_nWrite) {
        reg_src = (cyg_uint32)(sdcard_info->controller_base + MCI_REG_FIFO);
        reg_dst = phys_addr;

        reg_ctrl |= P5P_DMA_CH_CTRL_DI;
        reg_cfg |= P5P_DMA_CH_CONFIG_FLOWCNTRL_PERIPH2MEM;

        switch ((cyg_uint32)sdcard_info->controller_base) {
            case (cyg_uint32)MCI0_BASE_ADDRESS:
                reg_cfg |= 0 << P5P_DMA_CH_CONFIG_SRCPERIPH_SHIFT;
                break;

            case (cyg_uint32)MCI1_BASE_ADDRESS:
                reg_cfg |= 1 << P5P_DMA_CH_CONFIG_SRCPERIPH_SHIFT;
                break;

            default:
                break;
        }
    } else {
        reg_src = phys_addr;
        reg_dst = (cyg_uint32)(sdcard_info->controller_base + MCI_REG_FIFO);

        reg_ctrl |= P5P_DMA_CH_CTRL_SI;
        reg_cfg |= P5P_DMA_CH_CONFIG_FLOWCNTRL_MEM2PERIPH;

        switch ((cyg_uint32)sdcard_info->controller_base) {
            case (cyg_uint32)MCI0_BASE_ADDRESS:
                reg_cfg |= 0 << P5P_DMA_CH_CONFIG_DSTPERIPH_SHIFT;
                break;

            case (cyg_uint32)MCI1_BASE_ADDRESS:
                reg_cfg |= 1 << P5P_DMA_CH_CONFIG_DSTPERIPH_SHIFT;
                break;

            default:
                break;
        }
    }

    write_dma_uint32(P5P_DMA_REG_CHX_SRCADDR, reg_src);
    write_dma_uint32(P5P_DMA_REG_CHX_DESTADDR, reg_dst);
    write_dma_uint32(P5P_DMA_REG_CHX_LLI, 0);

    write_dma_uint32(P5P_DMA_REG_CHX_CTRL, reg_ctrl);
    write_dma_uint32(P5P_DMA_REG_CHX_CONFIG, reg_cfg);
}

static SDCARD_eError
mci_handle_end_command_transfert(cyg_uint32 status)
{
    SDCARD_eError err;

    // check errors and response
    if (status & MCI_STATUS_CMD_TIMEOUT) {
        err = SDCARD_ERROR_TIMEOUT;
    } else if (status & MCI_STATUS_CMD_CRC_FAIL) {
        err = SDCARD_ERROR_CRCFAIL;
    } else if (status & MCI_STATUS_START_BIT_ERR) {
        err = SDCARD_ERROR_CMD_HW_ERROR;
    } else if (status & MCI_STATUS_CMD_SENT) {
        err = SDCARD_ERROR_NONE;
    } else if (status & MCI_STATUS_CMD_RESP_END) {
        err = SDCARD_ERROR_NONE;
    } else {
        err = SDCARD_ERROR_NONE;
    }

    return err;
}


static SDCARD_eError
mci_handle_end_data_transfert(cyg_uint32 status)
{
    SDCARD_eError err;

    if (status & MCI_STATUS_DATA_TIMEOUT) {
        err = SDCARD_ERROR_DATA_TIMEOUT;
    } else if (status & MCI_STATUS_DATA_CRC_FAIL) {
        err = SDCARD_ERROR_CRCFAIL;
    } else if (status & MCI_STATUS_DATA_BLOCK_END) {
        err = SDCARD_ERROR_NONE;
    } else {
        err = SDCARD_ERROR_NONE;
    }

    return err;
}


static SDCARD_eError
mci_handle_end_transfert(cyg_uint32 status)
{
    SDCARD_eError err = SDCARD_ERROR_NONE;

    if (status) {
        //Data interrupt
        if (status & MCI_INT_DATA_MASK) {
            err |= mci_handle_end_data_transfert(status & MCI_INT_DATA_MASK);
        }

        //Command interrupt
        if (status & MCI_INT_CMD_MASK) {
            err |= mci_handle_end_command_transfert(status & MCI_INT_CMD_MASK);
        }
    }

    return err;
}


static cyg_uint32
mci_wait_end_polling(cyg_sdcard_disk_info_t *sdcard_info)
{
    cyg_uint32 status = 0;

    //Poll Normal interrupt status
    do {
        read_mci_uint32(sdcard_info, MCI_REG_STATUS, status);
        status &= MCI_INT_ALL_MASK;
    } while (status == 0);

    return status;
}


static cyg_uint32
mci_wait_end_irq(cyg_sdcard_disk_info_t *sdcard_info)
{
    MCI_privData *priv = (MCI_privData*)sdcard_info->priv;
    cyg_uint32 status = 0;

    cyg_semaphore_wait(&(priv->mci_sem_done));
    read_mci_uint32(sdcard_info, MCI_REG_STATUS, status);

    return status;
}


static SDCARD_eError
mci_wait_end(cyg_sdcard_disk_info_t *sdcard_info)
{
    MCI_privData *priv = (MCI_privData*)sdcard_info->priv;
    cyg_uint32 status = 0;
    SDCARD_eError ret;

    status = priv->wait_end(sdcard_info);

    DEBUG3("sdhci_wait_end: status 0x%x\n", status);

    if (status & MCI_STATUS_DATA_BLOCK_END)
        priv->mci_data_transfert = false;

    //Clear error
    write_mci_uint32(sdcard_info, MCI_REG_CLEAR, status);

    if ((status & MCI_STATUS_CMD_CRC_FAIL) &&
        (priv->mci_inhibit_crc == true)) {
        status &= ~MCI_STATUS_CMD_CRC_FAIL;
        status |= MCI_STATUS_CMD_RESP_END;

        DEBUG3("sdhci_wait_end: Inhibit crc new status 0x%x\n", status);
    }

    status &= MCI_INT_ALL_MASK;

    ret = mci_handle_end_transfert(status);

    if (sdcard_info->driver_mode == SDCARD_DRIVERMODE_INTERRUPT) {
        //We have clear our interrupt so unmask it
        cyg_drv_interrupt_unmask(priv->mci_interrupt);
    }

    return ret;
}

/**
 * Set the current frequency of the interface
 *
 * @param freq frequency in Hz
 */
static void
mci_set_frequency(cyg_sdcard_disk_info_t *sdcard_info, cyg_uint32 freq)
{
    cyg_uint32 reg_clock;
    MCI_privData *priv = (MCI_privData*)sdcard_info->priv;
    int clkdiv = MCI_MCLK_SPEED / (2 * freq) - 1;
    cyg_uint32 ctrl_freq;

    read_mci_uint32(sdcard_info, MCI_REG_CLOCK, reg_clock);
    reg_clock |= MCI_CLOCK_ENABLE;

    //Clear Freq bits
    reg_clock &= ~MCI_CLOCK_CLKDIV_MASK;

    if ((freq >= MCI_MCLK_SPEED) || (clkdiv < 0)) {
        reg_clock |= MCI_CLOCK_BYPASS;
        ctrl_freq = MCI_MCLK_SPEED;
    } else {
        reg_clock &= ~MCI_CLOCK_BYPASS;
        reg_clock |= clkdiv & MCI_CLOCK_CLKDIV_MASK;
        ctrl_freq = MCI_MCLK_SPEED / (2 *
            ((clkdiv & MCI_CLOCK_CLKDIV_MASK) + 1));
    }

    priv->mci_datatimer = MCI_TRANSFERT_TIMEOUT * (ctrl_freq / 1000);

    write_mci_uint32(sdcard_info, MCI_REG_CLOCK, reg_clock);
}

static void
mci_prepare_transfert(cyg_sdcard_disk_info_t *sdcard_info,
    cyg_bool read_nWrite, CYG_ADDRESS phys_addr, cyg_uint32 block_number)
{
    cyg_uint32 bytes_length;
    MCI_privData *priv = (MCI_privData*)sdcard_info->priv;

    //Set Data Timer
    write_mci_uint32(sdcard_info, MCI_REG_DATATIMER, priv->mci_datatimer);

    //Set Data length
    bytes_length = block_number * SDCARD_BLOCK_SIZE;
    write_mci_uint32(sdcard_info, MCI_REG_DATALENGTH, bytes_length);

    //data controller
    priv->mci_datactrl = MCI_DATACTRL_DMA_ENABLE |
        MCI_DATACTRL_BLOCK_SIZE_512;
    if (read_nWrite) {
        priv->mci_datactrl |= MCI_DATACTRL_READ;
    }

    priv->mci_data_transfert = true;

    //Set DMA address
    mci_dma_config(sdcard_info, phys_addr, block_number, read_nWrite);
}


static cyg_uint32
mci_isr(cyg_vector_t vector, cyg_addrword_t data)
{
    cyg_drv_interrupt_mask(vector);
    cyg_drv_interrupt_acknowledge(vector);

    return (CYG_ISR_CALL_DSR | CYG_ISR_HANDLED);
}


static void
mci_dsr(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data)
{
    cyg_sdcard_disk_info_t *sdcard_info = (cyg_sdcard_disk_info_t*)data;
    MCI_privData *priv = (MCI_privData*)sdcard_info->priv;

    cyg_semaphore_post(&(priv->mci_sem_done));

    //Do not unmask interrupt now
}


SDCARD_eError
ll_sdcard_hw_init(cyg_sdcard_disk_info_t *sdcard_info)
{
    cyg_uint32 reg_power;

    if (sdcard_info->init == true)
        return SDCARD_ERROR_NONE;

	//Set private data
    sdcard_info->priv = (void*)&g_privData;

    g_privData.mci_freq = CYG_MIN(MCI_MCLK_SPEED,
        CYGPKG_DEVS_DISK_SDCARD_PARROT5_MAX_FREQUENCY);

    //Set base controller address
#if (CYGPKG_DEVS_DISK_SDCARD_PARROT5_SD_INTERFACE == 0)
    sdcard_info->controller_base = MCI0_BASE_ADDRESS;
    g_privData.mci_interrupt = MCI0_INTERRUPT;
#elif (CYGPKG_DEVS_DISK_SDCARD_PARROT5_SD_INTERFACE == 1)
    sdcard_info->controller_base = MCI1_BASE_ADDRESS;
    g_privData.mci_interrupt = MCI1_INTERRUPT;
#else
# Warning Parrot5 SD interface not set
#endif

    if (sdcard_info->controller_base == MCI0_BASE_ADDRESS) {
        SELECT_PIN(SD0_CLK);
        SELECT_PIN(SD0_CMD);
        SELECT_PIN(SD0_DAT0);
        SELECT_PIN(SD0_DAT1);
        SELECT_PIN(SD0_DAT2);
        SELECT_PIN(SD0_DAT3);

        HAL_WRITE_UINT32(PARROT5_SYS + _SYS_SDCARD, SYS_SDCARD0_20MHZ);
        HAL_WRITE_UINT32(PARROT5_SYS + _SYS_CEN, SYS_CEN_MCICLK);
    } else if (sdcard_info->controller_base == MCI1_BASE_ADDRESS) {
        cyg_uint32 reg;

        SELECT_PIN(SD1_CLK);
        SELECT_PIN(SD1_CMD);
        SELECT_PIN(SD1_DAT0);
        SELECT_PIN(SD1_DAT1);
        SELECT_PIN(SD1_DAT2);
        SELECT_PIN(SD1_DAT3);

        // select radio bidir mode to enable SD pins
        HAL_READ_UINT32(PARROT5_SYS + _SYS_IOCR4, reg);
        reg |= SYS_IOCR4_RADIO_BIDIR | SYS_IOCR4_EDR;
        HAL_WRITE_UINT32(PARROT5_SYS + _SYS_IOCR4, reg);

        HAL_WRITE_UINT32(PARROT5_SYS + _SYS_SDCARD, SYS_SDCARD1_20MHZ);
        HAL_WRITE_UINT32(PARROT5_SYS + _SYS_CEN, SYS_CEN_MCI1CLK);
    } else {
        CYG_ASSERT(0, "MCI_State.Base error");
    }

    // Power off
    reg_power = 0;
    write_mci_uint32(sdcard_info, MCI_REG_POWER, reg_power);
    hal_delay_us(10000);
    // Power up
    reg_power = MCI_POWER_UP | (13 << MCI_POWER_VOLTAGE_SHIFT);
    write_mci_uint32(sdcard_info, MCI_REG_POWER, reg_power);
    hal_delay_us(10000);
    // Power on
    reg_power = MCI_POWER_ON | (13 << MCI_POWER_VOLTAGE_SHIFT);
    write_mci_uint32(sdcard_info, MCI_REG_POWER, reg_power);
    hal_delay_us(10000);

    mci_dma_init();

    //Set signal we will manage
    write_mci_uint32(sdcard_info, MCI_REG_MASK0, MCI_INT_ALL_MASK);
    write_mci_uint32(sdcard_info, MCI_REG_MASK1, MCI_INT_ALL_MASK);

    if (sdcard_info->driver_mode == SDCARD_DRIVERMODE_INTERRUPT) {
        //For using interrupt
        g_privData.wait_end = mci_wait_end_irq;

        // interrupt init
        cyg_drv_interrupt_create(
                g_privData.mci_interrupt,
                MCI_ISR_PRIORITY,
                (cyg_addrword_t)sdcard_info,
                (cyg_ISR_t*)&mci_isr,
                (cyg_DSR_t*)&mci_dsr,
                &sdcard_int_handle,
                &intr);
        cyg_drv_interrupt_attach(sdcard_int_handle);
        cyg_drv_interrupt_unmask(g_privData.mci_interrupt);
    } else if (sdcard_info->driver_mode == SDCARD_DRIVERMODE_POLLING) {
        //For using polling
        g_privData.wait_end = mci_wait_end_polling;
    } else {
        return SDCARD_ERROR_DATA_HW_ERROR;
    }

    //Init done and OK
    sdcard_info->init = true;

    return SDCARD_ERROR_NONE;
}

SDCARD_eError
ll_sdcard_card_init(cyg_sdcard_disk_info_t *sdcard_info)
{
    SDCARD_eError err = SDCARD_ERROR_NONE;
    MCI_privData *priv = (MCI_privData*)sdcard_info->priv;
    cyg_uint32 freq;
    cyg_uint32 reg_clock = MCI_CLOCK_ENABLE | MCI_CLOCK_PWRSAVE;

    //Set frequency according to sdcard connected
    freq = CYG_MIN(priv->mci_freq, sdcard_info->sdcard_freq);
    mci_set_frequency(sdcard_info, freq);

    read_mci_uint32(sdcard_info, MCI_REG_CLOCK, reg_clock);

    //Set bus width
    switch (sdcard_info->sdcard_card_type) {
        case SDCARD_TYPE_MMC:
            //nothing to do
            break;

        case SDCARD_TYPE_SDCARD:
        case SDCARD_TYPE_SDHC:
            reg_clock |= MCI_CLOCK_WIDEBUS;
            break;

        default:
            err = SDCARD_ERROR_HARDWARE;
            break;
    }
    write_mci_uint32(sdcard_info, MCI_REG_CLOCK, reg_clock);

    //Set voltage
    //TODO

    return err;
}

SDCARD_eError
ll_sdcard_send_command(cyg_sdcard_disk_info_t *sdcard_info,
    cyg_uint32 command, cyg_uint32 argument, cyg_uint32 *response)
{
    MCI_privData *priv = (MCI_privData*)sdcard_info->priv;
    cyg_uint32 status_reg;
    cyg_uint16 command_reg = 0;
    cyg_uint16 answer_reg;
    SDCARD_eError ret;

    //Wait controller ready
    do {
        read_mci_uint32(sdcard_info, MCI_REG_STATUS, status_reg);
    } while (status_reg & MCI_STATUS_ACTIVE);

    write_mci_uint32(sdcard_info, MCI_REG_ARGUMENT, argument);

    //Set response type
    if (command & SDCARD_COMMAND_RESPONSE)
        command_reg |= MCI_COMMAND_RESPONSE;

    if (command & SDCARD_COMMAND_LONG_RESPONSE)
        command_reg |= MCI_COMMAND_LONG_RESPONSE;

    if (command & SDCARD_COMMAND_ASK_BUSY) {
//        write_mci_uint32(sdcard_info, MCI_REG_DATATIMER, priv->mci_datatimer);
//        priv->mci_data_transfert = true;
    }

    if (command & SDCARD_COMMAND_WITH_DATA) {
//        write_mci_uint32(sdcard_info, MCI_REG_DATACTRL, 0);
        priv->mci_datactrl |= MCI_DATACTRL_ENABLE;
        write_mci_uint32(sdcard_info, MCI_REG_DATACTRL, priv->mci_datactrl);
    }

    if (command & SDCARD_COMMAND_ASK_RESET)
        command_reg |= MCI_COMMAND_RESET;

    if (command & SDCARD_COMMAND_ASK_ABORT)
        command_reg |= MCI_COMMAND_ABORT;

    if (command & SDCARD_COMMAND_ASK_MULTI)
        command_reg |= MCI_COMMAND_MULTI;

    if (command & SDCARD_COMMAND_ASK_CRC)
        priv->mci_inhibit_crc = false;
    else
        priv->mci_inhibit_crc = true;

    command_reg |= command & SDCARD_COMMAND_MASK;
    command_reg |= MCI_COMMAND_ENABLE;

    write_mci_uint32(sdcard_info, MCI_REG_COMMAND, command_reg);

    // wait for transfer end
    ret = mci_wait_end(sdcard_info);

    //Check if the answer received is coherent with command send
    if ((command & SDCARD_COMMAND_ASK_OPCODE) && (ret == SDCARD_ERROR_NONE)) {
        read_mci_uint32(sdcard_info, MCI_REG_RESPCMD, answer_reg);
        answer_reg &= SDCARD_COMMAND_MASK;
        if (answer_reg != (command & SDCARD_COMMAND_MASK)) {
            ret = SDCARD_ERROR_DATA_HW_ERROR;
        }
    }

    if (ret == SDCARD_ERROR_NONE) {
        if (response) {
            if (command & SDCARD_COMMAND_LONG_RESPONSE) {
                int reg;
                cyg_uint32 resp[4];
                for (reg = 0; reg < 4; reg++) {
                    read_mci_uint32(sdcard_info, (MCI_REG_RESPONSE0 +
                        reg * 4), resp[reg]);
                }

                for (reg = 0; reg < 3; reg++) {
                    response[reg] = resp[3 - reg] >> 8;
                    response[reg] |= (resp[2 - reg] & 0xFF) << 24;
                }
                response[3] = resp[0] >> 8;
            } else if (command & SDCARD_COMMAND_RESPONSE) {
                cyg_uint32 resp;
                read_mci_uint32(sdcard_info, MCI_REG_RESPONSE0, resp);
                response[0] = resp;
            }
        }
    } else {
//        mci_reset(sdcard_info, SDHCI_SOFT_RESET_CMD);
//        if (command & SDCARD_COMMAND_WITH_DATA) {
//            mci_reset(sdcard_info, SDHCI_SOFT_RESET_DATA);
//        }
    }

    while ((ret == SDCARD_ERROR_NONE) && (priv->mci_data_transfert)) {
        ret = mci_wait_end(sdcard_info);
    }

    return ret;
}


SDCARD_eError
ll_sdcard_send_data(cyg_sdcard_disk_info_t *sdcard_info,
    CYG_ADDRESS phys_addr, cyg_uint32 block_to_write)
{
    SDCARD_eError ret = SDCARD_ERROR_NONE;

    mci_prepare_transfert(sdcard_info, false, phys_addr, block_to_write);

    return ret;
}


SDCARD_eError
ll_sdcard_receive_data(cyg_sdcard_disk_info_t *sdcard_info,
    CYG_ADDRESS phys_addr, cyg_uint32 block_to_read)
{
    SDCARD_eError ret = SDCARD_ERROR_NONE;

    mci_prepare_transfert(sdcard_info, true, phys_addr, block_to_read);

    return ret;
}

