//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
*******************************************************************************
* @file     sdhci.c
* @brief    SDcard low level driver for P6/P6i
*
* Copyright (C) 2007 Parrot S.A.
*
* @author   florent.bayendrian@parrot.com
* @date     2007-12-19
*******************************************************************************
*/

#include <cyg/hal/cpu_types.h>
#include <cyg/hal/drv_api.h>
#include <cyg/hal/hal_io.h>
#include <cyg/kernel/kapi.h>
#include <cyg/io/sdcard.h>
#include <assert.h>

#include "sdhci.h"
#include <pkgconf/devs_disk_sdcard_parrot6.h>

#define SDHCI_ISR_PRIORITY      (4)
#define SDHCI_TRANSFERT_TIMEOUT (250)   //250 ms

#define read_sdhci_uint32(_ctrl_, _off_, _val_)                 \
    HAL_READ_UINT32(_ctrl_->controller_base + _off_, _val_)

#define write_sdhci_uint32(_ctrl_, _off_, _val_)                \
    HAL_WRITE_UINT32(_ctrl_->controller_base + _off_, _val_)

#define read_sdhci_uint16(_ctrl_, _off_, _val_)                 \
    HAL_READ_UINT16(_ctrl_->controller_base + _off_, _val_)

#define write_sdhci_uint16(_ctrl_, _off_, _val_)                \
    HAL_WRITE_UINT16(_ctrl_->controller_base + _off_, _val_)

#define read_sdhci_uint8(_ctrl_, _off_, _val_)                  \
    HAL_READ_UINT8(_ctrl_->controller_base + _off_, _val_)

#define write_sdhci_uint8(_ctrl_, _off_, _val_)                 \
    HAL_WRITE_UINT8(_ctrl_->controller_base + _off_, _val_)

typedef struct SDHCI_privData {
    cyg_uint32  sdhci_freq;
    cyg_sem_t   sdhci_sem_done;
    cyg_bool    sdhci_data_transfert;
    cyg_uint32  (*wait_end)(cyg_sdcard_disk_info_t *sdcard_info);
} SDHCI_privData;

static SDHCI_privData g_privData;

// interrupt handlers
static cyg_handle_t     sdcard_int_handle;
static cyg_interrupt    intr;
static cyg_uint32       sdhci_isr(cyg_vector_t vector, cyg_addrword_t data);
static void             sdhci_dsr(cyg_vector_t vector, cyg_ucount32 count,
                                  cyg_addrword_t data);


static void
sdhci_reset(cyg_sdcard_disk_info_t *sdcard_info, cyg_uint8 reset)
{
    cyg_uint8 reg;

    write_sdhci_uint8(sdcard_info, SDHCI_REG_SOFT_RESET, reset);
    do {
        read_sdhci_uint8(sdcard_info, SDHCI_REG_SOFT_RESET, reg);
    } while (reg & reset);
}


/**
 * Set the current frequency of the interface
 *
 * @param freq frequency in Hz
 */
static void
sdhci_set_frequency(cyg_sdcard_disk_info_t *sdcard_info, cyg_uint32 freq)
{
    SDHCI_privData *priv = (SDHCI_privData*)sdcard_info->priv;
    cyg_uint8 ctrl;
    cyg_uint16 clk, reg;
    cyg_uint32 div;
    cyg_uint32 current_timeout;

    //temporary disable the clock
    write_sdhci_uint16(sdcard_info, SDHCI_REG_CLOCKCTRL, 0);

    // high speed mode or not
    read_sdhci_uint8(sdcard_info, SDHCI_REG_HOSTCTRL, ctrl);
    if (freq > SDCARD_CLK_SD_FULL)
        ctrl |= SDHCI_HOSTCTRL_HIGHSPEED;
    else
        ctrl &= ~SDHCI_HOSTCTRL_HIGHSPEED;
    write_sdhci_uint8(sdcard_info, SDHCI_REG_HOSTCTRL, ctrl);

    //Select divider for sdclk
    for (div = 1; div < 256; div *= 2) {
        if ((priv->sdhci_freq / div) <= freq)
            break;
    }

    //Real frequency of the controller
    freq = (priv->sdhci_freq / div);

    div >>= 1;
    clk = (div << SDHCI_DIVIDER_SHIFT) | SDHCI_CLOCKCTRL_INT_CLOCK_ENABLE;

    //Enable internal clock
    write_sdhci_uint16(sdcard_info, SDHCI_REG_CLOCKCTRL, clk);

    //And wait for stable bit
    do {
        read_sdhci_uint16(sdcard_info, SDHCI_REG_CLOCKCTRL, reg);
    } while (!(reg & SDHCI_CLOCKCTRL_INT_CLOCK_STABLE));

    //Now enable the clock to the card
    clk |= SDHCI_CLOCKCTRL_CLOCK_ENABLE;
    write_sdhci_uint16(sdcard_info, SDHCI_REG_CLOCKCTRL, clk);

    //current_timeout in ms
    current_timeout = ((1 << (13 + SDHCI_MAX_TIMEOUT)) / freq) * 1000;
    reg = SDHCI_MAX_TIMEOUT;
    while (current_timeout > SDHCI_TRANSFERT_TIMEOUT) {
        reg--;
        current_timeout >>= 1;
    }

    if (reg != SDHCI_MAX_TIMEOUT) {
        reg++;
    }

    DEBUG3("Timeout reg: %d\n", reg);

    write_sdhci_uint8(sdcard_info, SDHCI_REG_TIMEOUT_CTRL, reg);
}


static void
sdhci_prepare_transfert(cyg_sdcard_disk_info_t *sdcard_info,
    cyg_bool read_nWrite, CYG_ADDRESS phys_addr, cyg_uint32 block_number)
{
    cyg_uint16 mode;

    //Set transfert mode
    mode = SDHCI_TRANS_DMA_EN | SDHCI_TRANS_BLOCK_COUNT_EN;
    mode |= read_nWrite ? SDHCI_TRANS_READ : 0;
    mode |= block_number > 1 ? SDHCI_TRANS_MULTIBLOCK : 0;
    write_sdhci_uint16(sdcard_info, SDHCI_REG_TRANS_MODE, mode);

    write_sdhci_uint16(sdcard_info, SDHCI_REG_BLOCK_COUNT, block_number);

    //Set DMA address
    write_sdhci_uint32(sdcard_info, SDHCI_REG_DMA_ADDRESS,
        (cyg_uint32)phys_addr);
}


static SDCARD_eError
sdhci_handle_end_command_transfert(cyg_uint32 status)
{
    SDCARD_eError err;

    // check errors and response
    if (status & SDHCI_INT_ERR_CMD_TIMEOUT) {
        err = SDCARD_ERROR_TIMEOUT;
    } else if (status & SDHCI_INT_ERR_CMD_CRC) {
        err = SDCARD_ERROR_CRCFAIL;
    } else if (status & SDHCI_INT_ERR_CMD_ENDBIT) {
        err = SDCARD_ERROR_CMD_HW_ERROR;
    } else if (status & SDHCI_INT_ERR_CMD_INDEX) {
        err = SDCARD_ERROR_INDEX;
    } else if (status & SDHCI_INT_CMD_COMPLETE) {
        err = SDCARD_ERROR_NONE;
    } else {
        err = SDCARD_ERROR_NONE;
    }

    return err;
}


/**
 * Handle the dpsm
 */
static SDCARD_eError
sdhci_handle_end_data_transfert(cyg_uint32 status)
{
    SDCARD_eError err;

    if (status & SDHCI_INT_ERR_DATA_TIMEOUT) {
        err = SDCARD_ERROR_DATA_TIMEOUT;
    } else if (status & SDHCI_INT_ERR_DATA_CRC) {
        err = SDCARD_ERROR_DATA_HW_ERROR;
    } else if (status & SDHCI_INT_ERR_DATA_ENDBIT) {
        err = SDCARD_ERROR_DATA_HW_ERROR;
    } else if (status & SDHCI_INT_DATA_END) {
        err = SDCARD_ERROR_NONE;
    } else {
        err = SDCARD_ERROR_NONE;
    }

    return err;
}


static SDCARD_eError
sdhci_handle_end_transfert(cyg_uint32 status)
{
    SDCARD_eError err = SDCARD_ERROR_NONE;

    if (status) {
        //Data interrupt
        if (status & SDHCI_INT_DATA_MASK) {
            err |= sdhci_handle_end_data_transfert(status & SDHCI_INT_DATA_MASK);
        }

        //Command interrupt
        if (status & SDHCI_INT_CMD_MASK) {
            err |= sdhci_handle_end_command_transfert(status & SDHCI_INT_CMD_MASK);
        }
    }

    return err;
}


static cyg_uint32
sdhci_wait_end_polling(cyg_sdcard_disk_info_t *sdcard_info)
{
    cyg_uint32 status = 0;

    //Poll Normal interrupt status
    do {
        read_sdhci_uint32(sdcard_info, SDHCI_REG_INT_STATUS, status);
    } while (status == 0);

    return status;
}


static cyg_uint32
sdhci_wait_end_irq(cyg_sdcard_disk_info_t *sdcard_info)
{
    SDHCI_privData *priv = (SDHCI_privData*)sdcard_info->priv;
    cyg_uint32 status = 0;

    cyg_semaphore_wait(&(priv->sdhci_sem_done));
    read_sdhci_uint32(sdcard_info, SDHCI_REG_INT_STATUS, status);

    return status;
}


static SDCARD_eError
sdhci_wait_end(cyg_sdcard_disk_info_t *sdcard_info)
{
    SDHCI_privData *priv = (SDHCI_privData*)sdcard_info->priv;
    cyg_uint32 status = 0;
    SDCARD_eError ret;

    status = priv->wait_end(sdcard_info);

    DEBUG3("sdhci_wait_end: status 0x%x\n", status);

    if (status & SDHCI_INT_DATA_END)
        priv->sdhci_data_transfert = false;

    status &= SDHCI_INT_ALL_MASK;

    ret = sdhci_handle_end_transfert(status);

    //Clear error
    write_sdhci_uint32(sdcard_info, SDHCI_REG_INT_STATUS, status);

    if (sdcard_info->driver_mode == SDCARD_DRIVERMODE_INTERRUPT) {
        //We have clear our interrupt so unmask it
        cyg_drv_interrupt_unmask(CYGNUM_HAL_INTERRUPT_SDIO);
    }

    return ret;
}

static cyg_uint32
sdhci_isr(cyg_vector_t vector, cyg_addrword_t data)
{
    cyg_drv_interrupt_mask(vector);
    cyg_drv_interrupt_acknowledge(vector);

    return (CYG_ISR_CALL_DSR | CYG_ISR_HANDLED);
}


static void
sdhci_dsr(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data)
{
    cyg_sdcard_disk_info_t *sdcard_info = (cyg_sdcard_disk_info_t*)data;
    SDHCI_privData *priv = (SDHCI_privData*)sdcard_info->priv;

    cyg_semaphore_post(&(priv->sdhci_sem_done));

    //Do not unmask interrupt now
}


SDCARD_eError
ll_sdcard_hw_init(cyg_sdcard_disk_info_t *sdcard_info)
{
    cyg_uint64 caps CYGBLD_ATTRIB_UNUSED; //Just used with assert
    cyg_uint16 block_size = SDCARD_BLOCK_SIZE;
    cyg_uint32 host_dma_bytes = 4096; //Min dma size
    cyg_uint8 host_dma_bits = 0;
    unsigned int ahb_freq;
    cyg_uint32 sdhci_freq = CYGPKG_DEVS_DISK_SDCARD_PARROT6_MAX_FREQUENCY;
    cyg_uint8 sdhci_div = 3;
    cyg_uint32 intmask = SDHCI_INT_ALL_MASK;

    if (sdcard_info->init == true)
        return SDCARD_ERROR_NONE;

	//Set private data
    sdcard_info->priv = (void*)&g_privData;

    //Set base controller address

    sdcard_info->controller_base = SDHCI0_BASE_ADDRESS +
        0x100 * CYGPKG_DEVS_DISK_SDCARD_PARROT6_SD_INTERFACE;

    ahb_freq = pclock_get_rate(NULL); //AHB freq

    //freq(sdhci) = freq(AHB)/(SDHCI_DIV + 3)
    while ((ahb_freq / sdhci_div) > sdhci_freq) {
        sdhci_div++;

        if (sdhci_div >= (8 + 3))
            return SDCARD_ERROR_HARDWARE;
    }

    HAL_WRITE_UINT32(PARROT6_SYS + _P6_SYS_SDCARD, (sdhci_div - 3));
    HAL_WRITE_UINT32(PARROT6_SYS + _P6_SYS_CDIS, P6_SYS_CLK_SDIO);
    HAL_WRITE_UINT32(PARROT6_SYS + _P6_SYS_CEN, P6_SYS_CLK_SDIO);

    if (PARROT_CHIP_IS_P6()) {
#if (CYGPKG_DEVS_DISK_SDCARD_PARROT6_SD_INTERFACE == 0)
        SELECT_PIN(SD0_CLK);
        SELECT_PIN(SD0_CMD);
        SELECT_PIN(SD0_DAT0);
        SELECT_PIN(SD0_DAT1);
        SELECT_PIN(SD0_DAT2);
        SELECT_PIN(SD0_DAT3);
#elif (CYGPKG_DEVS_DISK_SDCARD_PARROT6_SD_INTERFACE == 1)
        SELECT_PIN(SD1_CLK);
        SELECT_PIN(SD1_CMD);
        SELECT_PIN(SD1_DAT0);
        SELECT_PIN(SD1_DAT1);
        SELECT_PIN(SD1_DAT2);
        SELECT_PIN(SD1_DAT3);
#elif (CYGPKG_DEVS_DISK_SDCARD_PARROT6_SD_INTERFACE == 2)
        SELECT_PIN(SD2_CLK);
        SELECT_PIN(SD2_CMD);
        SELECT_PIN(SD2_DAT0);
        SELECT_PIN(SD2_DAT1);
        SELECT_PIN(SD2_DAT2);
        SELECT_PIN(SD2_DAT3);
#else
# Warning Parrot6 SD interface not set
#endif
    } else if (PARROT_CHIP_IS_P6I()) {
        SELECT_PIN(P6I_SD0_CLK);
        SELECT_PIN(P6I_SD0_CMD);
        SELECT_PIN(P6I_SD0_DAT0);
        SELECT_PIN(P6I_SD0_DAT1);
        SELECT_PIN(P6I_SD0_DAT2);
        SELECT_PIN(P6I_SD0_DAT3);
    }

    //Check capabilities
    read_sdhci_uint32(sdcard_info, SDHCI_REG_CAPABILITIES, caps);

    CYG_ASSERT(caps & SDHCI_CAPS_SUPPORT_DMA,
        "DMA NOT supported");
    CYG_ASSERT(caps & SDHCI_CAPS_SUPPORT_HIGHSPEED,
        "High Speed mode NOT supported");
    CYG_ASSERT(caps & SDHCI_CAPS_SUPPORT_1V8,
        "1.8V NOT supported");
    CYG_ASSERT(caps & SDHCI_CAPS_SUPPORT_3V0,
        "3.0V NOT supported");
    CYG_ASSERT(caps & SDHCI_CAPS_SUPPORT_3V3,
        "3.3V NOT supported");
    CYG_ASSERT(caps & SDHCI_CAPS_SUPPORT_INT,
        "Interrupt NOT supported");

    //Set block size information
    for (host_dma_bits = 0; host_dma_bits < 8; host_dma_bits++)
        if (host_dma_bytes == SDCARD_MAX_TRANSFER_SIZE)
            break;
        else
            host_dma_bytes <<= 1;

    if (host_dma_bits != 8)
        block_size |= (host_dma_bits << SDHCI_MAX_DMA_BUFFER_SHIFT);
    else
        return SDCARD_ERROR_HARDWARE;

    //Set block size
    write_sdhci_uint16(sdcard_info, SDHCI_REG_BLOCK_SIZE, block_size);

    //Set signal we will manage
    write_sdhci_uint32(sdcard_info, SDHCI_REG_SIGNAL_ENABLE, intmask);
    write_sdhci_uint32(sdcard_info, SDHCI_REG_INT_ENABLE, intmask);

    if (sdcard_info->driver_mode == SDCARD_DRIVERMODE_INTERRUPT) {
        //For using interrupt
        g_privData.wait_end = sdhci_wait_end_irq;

        // interrupt init
        cyg_drv_interrupt_create(
                CYGNUM_HAL_INTERRUPT_SDIO,
                SDHCI_ISR_PRIORITY,
                (cyg_addrword_t)sdcard_info,
                (cyg_ISR_t*)&sdhci_isr,
                (cyg_DSR_t*)&sdhci_dsr,
                &sdcard_int_handle,
                &intr);
        cyg_drv_interrupt_attach(sdcard_int_handle);
        cyg_drv_interrupt_unmask(CYGNUM_HAL_INTERRUPT_SDIO);
    } else if (sdcard_info->driver_mode == SDCARD_DRIVERMODE_POLLING) {
        //For using polling
        g_privData.wait_end = sdhci_wait_end_polling;
    } else {
        return SDCARD_ERROR_DATA_HW_ERROR;
    }

    //Set controller base frequency
    g_privData.sdhci_freq = (ahb_freq / sdhci_div);

    //Reset controller for starting clean
    //sdhci_reset(sdcard_info, SDHCI_SOFT_RESET_CMD /*| SDHCI_SOFT_RESET_DATA*/);

    //Init done and OK
    sdcard_info->init = true;

    return SDCARD_ERROR_NONE;
}

SDCARD_eError
ll_sdcard_card_init(cyg_sdcard_disk_info_t *sdcard_info)
{
    SDCARD_eError err = SDCARD_ERROR_NONE;
    SDHCI_privData *priv = (SDHCI_privData*)sdcard_info->priv;
    cyg_uint32 freq;
    cyg_uint8 ctrl;

    //Set frequency according to sdcard connected
    freq = CYG_MIN(priv->sdhci_freq, sdcard_info->sdcard_freq);
    sdhci_set_frequency(sdcard_info, freq);

    if (err == SDCARD_ERROR_NONE) {
        //Set bus width
        read_sdhci_uint8(sdcard_info, SDHCI_REG_HOSTCTRL, ctrl);
        ctrl &= ~(SDHCI_HOSTCTRL_4BITBUS | SDHCI_HOSTCTRL_8BITBUS);
        switch (sdcard_info->sdcard_card_type) {
            case SDCARD_TYPE_MMC:
                //nothing to do
                break;

            case SDCARD_TYPE_SDCARD:
            case SDCARD_TYPE_SDHC:
                ctrl |= SDHCI_HOSTCTRL_4BITBUS;
                break;

            default:
                err = SDCARD_ERROR_HARDWARE;
                break;
        }
    }

    if (err == SDCARD_ERROR_NONE) {
        write_sdhci_uint8(sdcard_info, SDHCI_REG_HOSTCTRL, ctrl);

        //Set voltage
        switch (sdcard_info->sdcard_voltage) {
            case SDCARD_VOLTAGE_NONE:
                //Turn off power
                ctrl = 0;
                break;

            case SDCARD_VOLTAGE_1V8:
                ctrl = SDHCI_POWER_180 | SDHCI_POWER_ON;
                break;

            case SDCARD_VOLTAGE_3V0:
            default:
                err = SDCARD_ERROR_HARDWARE;
                break;

            case SDCARD_VOLTAGE_3V3:
                ctrl = SDHCI_POWER_330 | SDHCI_POWER_ON;
                break;
        }
    }

    if (err == SDCARD_ERROR_NONE) {
        write_sdhci_uint8(sdcard_info, SDHCI_REG_POWERCTRL, ctrl);
    }

    return err;
}

SDCARD_eError
ll_sdcard_send_command(cyg_sdcard_disk_info_t *sdcard_info,
    cyg_uint32 command, cyg_uint32 argument, cyg_uint32 *response)
{
    SDHCI_privData *priv = (SDHCI_privData*)sdcard_info->priv;
    cyg_uint32 present_state_reg;
    cyg_uint32 present_state_mask = SDHCI_PRESENT_STATE_CMDINHIB;
    cyg_uint16 command_reg;
    SDCARD_eError ret;

    write_sdhci_uint32(sdcard_info, SDHCI_REG_ARGUMENT, argument);

    //Set response type
    if (!(command & SDCARD_COMMAND_RESPONSE)) {
        command_reg = SDHCI_COMMAND_RSP_NONE;
    } else if (command & SDCARD_COMMAND_LONG_RESPONSE) {
        command_reg = SDHCI_COMMAND_RSP136;
    } else if (command & SDCARD_COMMAND_ASK_BUSY) {
        command_reg = SDHCI_COMMAND_RSP48_BUSY;
        priv->sdhci_data_transfert = true;
    } else {
        command_reg = SDHCI_COMMAND_RSP48;
    }

    if (command & SDCARD_COMMAND_ASK_CRC) {
        command_reg |= SDHCI_COMMAND_CHECK_CRC;
    }
    if (command & SDCARD_COMMAND_ASK_OPCODE) {
        command_reg |= SDHCI_COMMAND_CHECK_INDEX;
    }
    if (command & SDCARD_COMMAND_WITH_DATA) {
        priv->sdhci_data_transfert = true;
        command_reg |= SDHCI_COMMAND_DATA;
        present_state_mask |= SDHCI_PRESENT_STATE_DATAINHIB;
    }
    if (command & SDCARD_COMMAND_ASK_ABORT)
        command_reg |= SDHCI_COMMAND_ABORT;

    command_reg |= ((command & SDCARD_COMMAND_MASK) <<
        SDHCI_COMMAND_INDEX_SHIFT);

    //The Host Driver shall check the Command Inhibit (DAT) bit
    //and Command Inhibit (CMD) bit in the Present State register
    //before writing to this register
    do {
        read_sdhci_uint32(sdcard_info, SDHCI_REG_PRESENT_STATE,
            present_state_reg);
    } while (present_state_reg & present_state_mask);

    write_sdhci_uint16(sdcard_info, SDHCI_REG_COMMAND, command_reg);

    // wait for transfer end
    ret = sdhci_wait_end(sdcard_info);

    if (ret == SDCARD_ERROR_NONE) {
        if (response && (command & SDCARD_COMMAND_RESPONSE)) {
            // short response
            cyg_uint32 answer;

            read_sdhci_uint32(sdcard_info, SDHCI_REG_RESPONSE0, answer);
            response[0] = answer;

            if (command & SDCARD_COMMAND_LONG_RESPONSE) {
                int resp;
                for (resp = 1; resp < 4; resp++) {
                    read_sdhci_uint32(sdcard_info, (SDHCI_REG_RESPONSE0 +
                        resp * 4), answer);
                    response[resp] = answer;
                }
            }
        }
    } else {
        sdhci_reset(sdcard_info, SDHCI_SOFT_RESET_CMD);
        if (command & SDCARD_COMMAND_WITH_DATA) {
            sdhci_reset(sdcard_info, SDHCI_SOFT_RESET_DATA);
        }
    }

    while ((ret == SDCARD_ERROR_NONE) && (priv->sdhci_data_transfert)) {
        ret = sdhci_wait_end(sdcard_info);
    }

    return ret;
}


SDCARD_eError
ll_sdcard_send_data(cyg_sdcard_disk_info_t *sdcard_info,
    CYG_ADDRESS phys_addr, cyg_uint32 block_to_write)
{
    SDCARD_eError ret = SDCARD_ERROR_NONE;

    sdhci_prepare_transfert(sdcard_info, false, phys_addr, block_to_write);

    return ret;
}


SDCARD_eError
ll_sdcard_receive_data(cyg_sdcard_disk_info_t *sdcard_info,
    CYG_ADDRESS phys_addr, cyg_uint32 block_to_read)
{
    SDCARD_eError ret = SDCARD_ERROR_NONE;

    sdhci_prepare_transfert(sdcard_info, true, phys_addr, block_to_read);

    return ret;
}
