/**
 ****************************************************************************
 * Driver for Cirrus Logic CS42XX stereo audio codec.
 *
 * @file   cs42xx.h
 * @brief  Main header of driver.
 *
 * Code is based on the following datasheets:
 * http://www.cirrus.com/en/pubs/proDatasheet/CS4245_F1.pdf
 * http://www.cirrus.com/en/pubs/proDatasheet/CS4265_F1.pdf
 * http://www.cirrus.com/en/pubs/proDatasheet/CS5345_F1.pdf
 *  
 * @author David Tylski
 * @date   Jan. 2007
 * 
 * Copyright (C) 2007, Parrot S.A. - www.parrot.biz
 ****************************************************************************
 */
#ifndef _CS42XX_H_
#define _CS42XX_H_

#include <cyg/io/io.h>
#include <cyg/io/i2c.h>
#include <cyg/io/gpio.h>
#include <pkgconf/devices_cs42xx.h>
#include <cyg/hal/hal_diag.h>          // HAL_DELAY_US()



/*
 * Defines.
 ***********/


/** Number of concurrent devices. */
#define CS42XX_NBDEVICES		CYGPKG_DEVICES_CS42XX_NBDEVICES


/**
 * @defgroup cs42xx_bitmacros  Macros to work on bits of integer.
 * @{
 *
 * These macros are local to the file, so:
 * - they can be located in C file, and
 * - considering their generic meaning and the local use, they can have
 *   a name which is not prefixed by CS42XX (this is clearer in source
 *   code below).
 */

/** Get a bit from integer.
 * E.g. with x="....0...", CS42XX_GETBIT(x,3) returns 0 (false).
 * and  with x="....1...", CS42XX_GETBIT(x,3) returns 1 (true).
 ****************************************************************************/
#define CS42XX_GETBIT(reg,bit) \
	(((reg) >> (bit)) & 1)

/** Get a multibit value from integer.
 * E.g. with x=".yyy....", CS42XX_GETVAL(x,0x07,4) returns yyy.
 ****************************************************************************/
#define CS42XX_GETVAL(reg,valmask,shift) \
	(((reg) >> (shift)) & (valmask))

/** Set a bit to the given boolean value.
 * E.g. CS42XX_SETBIT(x,3,false) writes "....0..." in x.
 ****************************************************************************/
#define CS42XX_SETBIT(reg,bit,val) \
	( (val)? (reg |= (1 << (bit))) : (reg &= ~(1 << (bit))) )

/** Apply a bitmask to set or clear bits.
 * E.g. CS42XX_SETBITMASK(x,0x1C,false) writes "...000.." in x,
 * and  CS42XX_SETBITMASK(x,0x1C,true)  writes "...111.." in x.
 ****************************************************************************/
#define CS42XX_SETBITMASK(reg,bitmask,val) \
	( (val)? (reg |= (bitmask)) : (reg &= ~(bitmask)) )

/** Set a shifted multibit value in integer.
 * E.g. CS42XX_SETVAL(x,0x06,0x07,4) writes ".110...." in x.
 ****************************************************************************/
#define CS42XX_SETVAL(reg,val,valmask,shift) \
	( reg = ((reg) & ~((valmask) << (shift))) \
		| (((val) & (valmask)) << (shift)) )

// !cs42xx_bitmacros
/** @} */


/*
 * Timers.
 */

/** Period of system timer, in milliseconds.
 * (Extracted from initial ParrotOS version.)
 */
#define CS42XX_TIMER_PERIOD_MS \
        (CYGNUM_HAL_RTC_NUMERATOR / (1000000 * CYGNUM_HAL_RTC_DENOMINATOR))

/** Insert a time delay in execution, of at least the given number of
 *  milliseconds.
 * 
 * This macro either makes the current thread sleep for a given time delay
 * (threaded context), or actively waits for this time delay (initial i.e.
 * cyg_user_start() context).
 *
 * @param[in] _delayms_  Time delay to insert
 ****************************************************************************/
#define CS42XX_DELAY_MS(_delayms_)                                           \
    do {                                                                     \
        if (cyg_thread_self() != cyg_thread_idle_thread())                   \
        {                                                                    \
            cyg_thread_delay(1 + ((_delayms_) / CS42XX_TIMER_PERIOD_MS));    \
        }                                                                    \
        else                                                                 \
        {                                                                    \
            HAL_DELAY_US((_delayms_) * 1000);                                \
        }                                                                    \
    } while(0)

/** Insert a time delay in execution, of at least the given number of
 *  microseconds.
 * 
 * This inserts an active wait corresponding to the given number of
 * milliseconds.
 *
 * @param[in] _delayus_  Time delay to insert
 ****************************************************************************/
#define CS42XX_DELAY_US(_delayus_)     HAL_DELAY_US(_delayus_)



/*
 * Types.
 *********/


/** Type for device id. */
typedef cyg_uint8 CS42XX_DEVID;


/*
 * These types are useful as arguments of API.
 */


/** Chip identifier.
 * @typedef CS42XX_CHIPID
 */
typedef enum CS42XX_CHIPID_TAG CS42XX_CHIPID;
enum CS42XX_CHIPID_TAG
{
	CS42XX_CHIPID_CS4245 = 0xC,      ///< Part number Id for CS4245.
	CS42XX_CHIPID_CS4265 = 0xD,      ///< Part number Id for CS4265.
	CS42XX_CHIPID_CS5345 = 0xE,      ///< Part number Id for CS5345.
	CS42XX_CHIPID_NOTSUPPORTED = 0xFF ///< Unsupported device.
};

/** Chip revision.
 * @typedef CS42XX_CHIPREV
 */
typedef enum CS42XX_CHIPREV_TAG CS42XX_CHIPREV;
enum CS42XX_CHIPREV_TAG
{
	CS42XX_CHIPREV_A  = 0x01,             ///< Part revision A.
	CS42XX_CHIPREV_B  = 0x02,             ///< Part revision B.
	CS42XX_CHIPREV_C0 = CS42XX_CHIPREV_B, ///< Part revision C0 (same as B).
	CS42XX_CHIPREV_C1 = 0x03,             ///< Part revision C1.
	CS42XX_CHIPREV_UNKNOWN = 0xFF         ///< Unknown revision.
};

/** Sample rates.
 * @typedef CS42XX_SAMPLERATE
 */
typedef enum CS42XX_SAMPLERATE_TAG CS42XX_SAMPLERATE;
enum CS42XX_SAMPLERATE_TAG
{
	CS42XX_SAMPLERATE_SINGLE = 0x00, ///< Single speed (4-50 kHz).
	CS42XX_SAMPLERATE_DOUBLE,        ///< Double speed (50-100 kHz).
	CS42XX_SAMPLERATE_QUAD,          ///< Quad speed (100-200 kHz).
	CS42XX_SAMPLERATE_NB
};

/** Digital interface format.
 * @typedef CS42XX_DIF
 */
typedef enum CS42XX_DIF_TAG CS42XX_DIF;
enum CS42XX_DIF_TAG
{
	CS42XX_DIF_LEFT24 = 0x00, ///< Left-justified, up to 24 bits.
	CS42XX_DIF_I2S,           ///< I2S, up to 24 bits.
	CS42XX_DIF_RIGHT16,       ///< Right-just., 16 bits (DAC only).
	CS42XX_DIF_RIGHT24,       ///< Right-just., 24 bits (DAC only).
	
	CS42XX_DIF_NB_DAC = CS42XX_DIF_RIGHT24+1,
	CS42XX_DIF_NB_ADC = CS42XX_DIF_I2S+1,
};

/** Master / slave selection.
 * @typedef CS42XX_DIFMS
 */
typedef enum CS42XX_DIFMS_TAG CS42XX_DIFMS;
enum CS42XX_DIFMS_TAG
{
	CS42XX_DIFMS_SLAVE  = false, ///< Device is slave on the interface.
	CS42XX_DIFMS_MASTER = true   ///< Device is master on the interface.
};

/** MCLK dividers.
 * @typedef CS42XX_MCLK_DIV
 */
typedef enum CS42XX_MCLK_DIV_TAG CS42XX_MCLK_DIV;
enum CS42XX_MCLK_DIV_TAG
{
	CS42XX_MCLK_DIV_1 = 0x00, ///< Divided by 1.
	CS42XX_MCLK_DIV_1_5,      ///< Divided by 3/2.
	CS42XX_MCLK_DIV_2,        ///< Divided by 2.
	CS42XX_MCLK_DIV_3,        ///< Divided by 3.
	CS42XX_MCLK_DIV_4,        ///< Divided by 4.
	CS42XX_MCLK_DIV_NB
};

/** Possible sources for auxiliary output.
 * @typedef CS42XX_AOUTSEL
 */
typedef enum CS42XX_AOUTSEL_TAG CS42XX_AOUTSEL;
enum CS42XX_AOUTSEL_TAG
{
	CS42XX_AOUTSEL_HIGHZ = 0x00, ///< Auxiliary output disabled.
	CS42XX_AOUTSEL_DAC,          ///< DAC signal on auxiliary output.
	CS42XX_AOUTSEL_PGA,          ///< PGA signal on auxiliary output.
	CS42XX_AOUTSEL_NB
};

/** Minimal PGA gain: -12dB. */
#define CS42XX_PGAGAIN_MIN		-24
/** Maximal PGA gain: +12dB. */
#define CS42XX_PGAGAIN_MAX		+24
/** PGA gain for 0dB. */
#define CS42XX_PGAGAIN_0DB		0

/** Minimal output volume level (maximal attenuation). */
#define CS42XX_VOLUME_MIN		0
/** Maximal output volume level (minimal attenuation). */
#define CS42XX_VOLUME_MAX		0xFF

/** Analog input multiplexer selection.
 * @typedef CS42XX_AINSEL
 */
typedef enum CS42XX_AINSEL_TAG CS42XX_AINSEL;
enum CS42XX_AINSEL_TAG
{
	CS42XX_AINSEL_MIC = 0x00, ///< Mic-level inputs (+32 dB).
	CS42XX_AINSEL_IN1,        ///< Line-level input pair 1.
	CS42XX_AINSEL_IN2,        ///< Line-level input pair 2.
	CS42XX_AINSEL_IN3,        ///< Line-level input pair 3.
	CS42XX_AINSEL_IN4,        ///< Line-level input pair 4.
	CS42XX_AINSEL_IN5,        ///< Line-level input pair 5.
	CS42XX_AINSEL_IN6,        ///< Line-level input pair 6.
	CS42XX_AINSEL_NB
};

/** Policies to deal with level changes on PGA and DAC volume: apply a soft
 *  ramp and/or change only on zero-cross.
 * @typedef CS42XX_LVLCHGPOLICY
 */
typedef enum CS42XX_LVLCHGPOLICY_TAG CS42XX_LVLCHGPOLICY;
enum CS42XX_LVLCHGPOLICY_TAG
{
	CS42XX_LVLCHGPOLICY_NONE = 0x00,     ///< Immediate change.
	CS42XX_LVLCHGPOLICY_ZCROSS,          ///< Zero-cross.
	CS42XX_LVLCHGPOLICY_SOFTRAMP,        ///< Soft ramp.
	CS42XX_LVLCHGPOLICY_SOFTRAMP_ZCROSS, ///< Soft ramp and zero-cross.
	CS42XX_LVLCHGPOLICY_NB
};

/** State of INT pin according to interrupt condition.
 * @typedef CS42XX_TRIGMODE
 */
typedef enum CS42XX_TRIGMODE_TAG CS42XX_TRIGMODE;
enum CS42XX_TRIGMODE_TAG
{
	CS42XX_TRIGMODE_RISING = 0x00, ///< On rising edge of condition.
	CS42XX_TRIGMODE_FALLING,       ///< On falling edge of condition.
	CS42XX_TRIGMODE_LEVEL,         ///< During all interrupt condition.
	CS42XX_TRIGMODE_RESERVED,
};

/** Register caching: synchronize function immediately or wait.
 * @typedef CS42XX_CACHESYNC
 */
typedef enum CS42XX_CACHESYNC_TAG CS42XX_CACHESYNC;
enum CS42XX_CACHESYNC_TAG
{
	CS42XX_CACHE_NOSYNC = false, ///< Store change, but delay it.
	CS42XX_CACHE_SYNC   = true   ///< Synchronize all changes to registers.
};



/*
 * Configuration structures.
 */


/** Structure to configure a digital interface (input or output)
 *  (part of CS42XX_CONFIG).
 */
struct CS42XX_CONFIG_DIO_TAG
{
	CS42XX_MCLK_DIV   mclkdiv;   ///< MCLK divider.
	CS42XX_SAMPLERATE srate;     ///< Sample rate.
	CS42XX_DIFMS      master;    ///< Master or slave interface.
	CS42XX_DIF        format;    ///< Format of frames.
};

/** Structure to set all global configuration in once.
 * @typedef CS42XX_CONFIG
 */
typedef struct CS42XX_CONFIG_TAG CS42XX_CONFIG;
struct CS42XX_CONFIG_TAG
{
	cyg_bool async;  ///< True if MCLK1 and MCLK2 can be asynchronous.

	struct CS42XX_CONFIG_DIO_TAG in;  ///< Configuration of input.
	struct CS42XX_CONFIG_DIO_TAG out; ///< Configuration of output.

	CS42XX_AINSEL  inputsel;  ///< Analog input source.
	CS42XX_AOUTSEL auxoutsel; ///< Signal on auxiliary output.
	cyg_bool       loopmode;  ///< Loop mode activated.
};


/** Union to describe either a PGA gain or an output volume (part of
 *  struct CS42XX_SETTINGS_CHANNEL_TAG).
 */
union CS42XX_SETTINGS_VOL_TAG
{
	cyg_int8  gain; ///< PGA gain      (ADC only).
	cyg_uint8 vol;  ///< Output volume (DAC only).
};

/** Structure to set channel settings (part of CS42XX_SETTINGS). */
struct CS42XX_SETTINGS_CHANNEL_TAG
{
	union CS42XX_SETTINGS_VOL_TAG a; ///< Gain or volume, channel A.
	union CS42XX_SETTINGS_VOL_TAG b; ///< Gain or volume, channel B.

	cyg_bool mute;   ///< Mute.
	cyg_bool filter; ///< High-pass filter (ADC), deemph. (DAC).

	CS42XX_LVLCHGPOLICY lvlchgpcy; ///< Policy to apply on level changes.
};

/** Structure to set all settings in once.
 * Settings are applied more often than global configuration.
 * @typedef CS42XX_SETTINGS
 */
typedef struct CS42XX_SETTINGS_TAG CS42XX_SETTINGS;
struct CS42XX_SETTINGS_TAG
{
	struct CS42XX_SETTINGS_CHANNEL_TAG adc; ///< Settings of ADC.
	struct CS42XX_SETTINGS_CHANNEL_TAG dac; ///< Settings of DAC.
};



/*
 * Exported functions.
 **********************/


// --- General stuff ---


// Initialize the whole driver.
extern Cyg_ErrNo
CS42XX_Init(void);

// Attach a device to an I2C bus.
extern Cyg_ErrNo
CS42XX_Init_AttachBus(CS42XX_DEVID, cyg_i2c_bus *, cyg_bool ad0, cyg_bool ad1);

// Get the part number ID and revision of device.
extern cyg_uint8
CS42XX_GetChipId(CS42XX_DEVID, Cyg_ErrNo *);


// --- Configurations in once ---


// Fill a config structure according to current configuration.
extern Cyg_ErrNo
CS42XX_Config_Get(CS42XX_DEVID, CS42XX_CONFIG *);

// Fill a settings structure according to current settings.
extern Cyg_ErrNo
CS42XX_Settings_Get(CS42XX_DEVID, CS42XX_SETTINGS *);


// Apply all configuration.
extern Cyg_ErrNo
CS42XX_Config_Set(CS42XX_DEVID, CS42XX_CONFIG *);

// Apply all settings in once.
extern Cyg_ErrNo
CS42XX_Settings_Set(CS42XX_DEVID, CS42XX_SETTINGS *);


// --- Interrupts ---


// Read the current interrupt status register.
extern CYG_BYTE
CS42XX_GetIntStatus(CS42XX_DEVID, Cyg_ErrNo *);

// Get the interrupt mask as an integer value.
extern CYG_BYTE
CS42XX_GetIntMaskVal(CS42XX_DEVID, Cyg_ErrNo *);

// Configure the INT output pin as active high or low.
extern Cyg_ErrNo
CS42XX_ConfigInt_OutputPin(CS42XX_DEVID, cyg_bool activeh, CS42XX_CACHESYNC);

// Set the interrupt mask (mask given as integer).
extern Cyg_ErrNo
CS42XX_ConfigInt_MaskVal(CS42XX_DEVID, CYG_BYTE mask, CS42XX_CACHESYNC);

// Set interrupt triggering mode (the same for all events).
extern Cyg_ErrNo
CS42XX_ConfigInt_TrigMode(CS42XX_DEVID, CS42XX_TRIGMODE trigmode,
		CS42XX_CACHESYNC);


// --- Power control ---


// Freeze or enable changes on control ports of codec.
extern Cyg_ErrNo
CS42XX_FreezePorts(CS42XX_DEVID, cyg_bool freeze);

// Set microphone preamplifier block in low-power mode.
extern Cyg_ErrNo
CS42XX_SetPowerDown_Mic(CS42XX_DEVID, cyg_bool lowpow, CS42XX_CACHESYNC);

// Set ADC pair in low-power mode.
extern Cyg_ErrNo
CS42XX_SetPowerDown_ADC(CS42XX_DEVID, cyg_bool lowpow, CS42XX_CACHESYNC);

// Set DAC pair in low-power mode.
extern Cyg_ErrNo
CS42XX_SetPowerDown_DAC(CS42XX_DEVID, cyg_bool lowpow, CS42XX_CACHESYNC);

// Set device in power-down mode (direct register access).
extern Cyg_ErrNo
CS42XX_SetPowerDown(CS42XX_DEVID, cyg_bool powdown);


// Power the device up (blocking function).
extern Cyg_ErrNo
CS42XX_PowerUp(CS42XX_DEVID);

// Power the device down(blocking function).
extern Cyg_ErrNo
CS42XX_PowerDown(CS42XX_DEVID);


// --- Settings of digital interfaces ---


// Get format on digital output (related to ADC).
extern CS42XX_DIF
CS42XX_GetFrameFormat_Out(CS42XX_DEVID, Cyg_ErrNo *);

// Get format on digital input (related to DAC).
extern CS42XX_DIF
CS42XX_GetFrameFormat_In(CS42XX_DEVID, Cyg_ErrNo *);

// Get if device is master of digital output interface.
extern cyg_bool
CS42XX_IsMaster_Out(CS42XX_DEVID, Cyg_ErrNo *);

// Get if device is master of digital input interface.
extern cyg_bool
CS42XX_IsMaster_In(CS42XX_DEVID, Cyg_ErrNo *);

// Get MCLK1 frequency divider (related to output interface).
extern CS42XX_MCLK_DIV
CS42XX_GetMCLKFreq_Out(CS42XX_DEVID, Cyg_ErrNo *);

// Get MCLK2 frequency divider (related to input interface).
extern CS42XX_MCLK_DIV
CS42XX_GetMCLKFreq_In(CS42XX_DEVID, Cyg_ErrNo *);

// Get if sample rates of digital I/O (thus DAC/ADC) are asynchronous.
extern cyg_bool
CS42XX_IsAsyncMode(CS42XX_DEVID, Cyg_ErrNo *);

// Get sample rate of digital output (related to ADC).
extern CS42XX_SAMPLERATE
CS42XX_GetSampleRate_Out(CS42XX_DEVID, Cyg_ErrNo *);

// Get sample rate of digital input (related to DAC).
extern CS42XX_SAMPLERATE
CS42XX_GetSampleRate_In(CS42XX_DEVID, Cyg_ErrNo *);


// Set format on digital output (related to ADC).
extern Cyg_ErrNo
CS42XX_ConfigFrameFormat_Out(CS42XX_DEVID, CS42XX_DIF format, CS42XX_CACHESYNC);

// Set format on digital input (related to DAC).
extern Cyg_ErrNo
CS42XX_ConfigFrameFormat_In(CS42XX_DEVID, CS42XX_DIF format, CS42XX_CACHESYNC);

// Set digital output interface of device in either master or slave mode.
extern Cyg_ErrNo
CS42XX_ConfigMaster_Out(CS42XX_DEVID, CS42XX_DIFMS master, CS42XX_CACHESYNC);

// Set digital input interface of device in either master or slave mode.
extern Cyg_ErrNo
CS42XX_ConfigMaster_In(CS42XX_DEVID, CS42XX_DIFMS master, CS42XX_CACHESYNC);

// Set MCLK(1,2) frequency dividers (MCLK1: output, MCLK2: input).
extern Cyg_ErrNo
CS42XX_ConfigMCLKFreq(CS42XX_DEVID,
		CS42XX_MCLK_DIV divout, CS42XX_MCLK_DIV divin, CS42XX_CACHESYNC);

// Set the relation between sample rates of digital interfaces (thus ADC/DAC).
extern Cyg_ErrNo
CS42XX_ConfigAsyncMode(CS42XX_DEVID, cyg_bool async, CS42XX_CACHESYNC);

// Set sample rate of digital output (related to ADC).
extern Cyg_ErrNo
CS42XX_ConfigSampleRate_Out(CS42XX_DEVID, CS42XX_SAMPLERATE, CS42XX_CACHESYNC);

// Set sample rate of digital input (related to DAC).
extern Cyg_ErrNo
CS42XX_ConfigSampleRate_In(CS42XX_DEVID, CS42XX_SAMPLERATE, CS42XX_CACHESYNC);

// Set sample rate for both digital I/O.
extern Cyg_ErrNo
CS42XX_ConfigSampleRate(CS42XX_DEVID, CS42XX_SAMPLERATE, CS42XX_CACHESYNC);


// --- Mute ---


// Get mute state of ADC.
extern cyg_bool
CS42XX_IsMuted_ADC(CS42XX_DEVID, Cyg_ErrNo *);

// Get mute state of DAC.
extern cyg_bool
CS42XX_IsMuted_DAC(CS42XX_DEVID, Cyg_ErrNo *);


// Mute or unmute ADC digital outputs (signal caught from ADC inputs).
extern Cyg_ErrNo
CS42XX_Mute_ADC(CS42XX_DEVID, cyg_bool mute, CS42XX_CACHESYNC);

// Mute or unmute DAC analog outputs.
extern Cyg_ErrNo
CS42XX_Mute_DAC(CS42XX_DEVID, cyg_bool mute, CS42XX_CACHESYNC);


// --- Input gain ---


// Get PGA gain (of ADC) for channel A.
extern cyg_int8
CS42XX_GetGain_ADCA(CS42XX_DEVID, Cyg_ErrNo *);

// Get PGA gain (of ADC) for channel B.
extern cyg_int8
CS42XX_GetGain_ADCB(CS42XX_DEVID, Cyg_ErrNo *);


// Set PGA gain (of ADC) for channel A.
extern Cyg_ErrNo
CS42XX_SetGain_ADCA(CS42XX_DEVID, cyg_int8 gain, CS42XX_CACHESYNC);

// Set PGA gain (of ADC) for channel B.
extern Cyg_ErrNo
CS42XX_SetGain_ADCB(CS42XX_DEVID, cyg_int8 gain, CS42XX_CACHESYNC);

// Set PGA gain (of ADC) for channels A and B.
extern Cyg_ErrNo
CS42XX_SetGain_ADC(CS42XX_DEVID,
			cyg_int8 gaina, cyg_int8 gainb, CS42XX_CACHESYNC);

// Increase PGA gain on both channels.
extern Cyg_ErrNo
CS42XX_SetGain_ADC_Up(CS42XX_DEVID, cyg_uint8 incr, CS42XX_CACHESYNC);

// Decrease PGA gain on both channels.
extern Cyg_ErrNo
CS42XX_SetGain_ADC_Down(CS42XX_DEVID, cyg_uint8 decr, CS42XX_CACHESYNC);


// --- Output volume ---


// Get output volume on channel A.
extern cyg_uint8
CS42XX_GetVolume_DACA(CS42XX_DEVID, Cyg_ErrNo *);

// Get output volume on channel B.
extern cyg_uint8
CS42XX_GetVolume_DACB(CS42XX_DEVID, Cyg_ErrNo *);


// Set output volume on channel A.
extern Cyg_ErrNo
CS42XX_SetVolume_DACA(CS42XX_DEVID, cyg_uint8 volume, CS42XX_CACHESYNC);

// Set output volume on channel B.
extern Cyg_ErrNo
CS42XX_SetVolume_DACB(CS42XX_DEVID, cyg_uint8 volume, CS42XX_CACHESYNC);

// Set output volume on channels A and B.
extern Cyg_ErrNo
CS42XX_SetVolume_DAC(CS42XX_DEVID,
		cyg_uint8 vola, cyg_uint8 volb, CS42XX_CACHESYNC);

// Increase output volume on both channels.
extern Cyg_ErrNo
CS42XX_SetVolume_DAC_Up(CS42XX_DEVID, cyg_uint8 incr, CS42XX_CACHESYNC);

// Decrease output volume on both channels.
extern Cyg_ErrNo
CS42XX_SetVolume_DAC_Down(CS42XX_DEVID, cyg_uint8 decr, CS42XX_CACHESYNC);


// --- Level change policy ---


// Get current policy about level changes on ADC.
extern CS42XX_LVLCHGPOLICY
CS42XX_GetLCPolicy_ADC(CS42XX_DEVID, Cyg_ErrNo *);

// Get current policy about level changes on DAC.
extern CS42XX_LVLCHGPOLICY
CS42XX_GetLCPolicy_DAC(CS42XX_DEVID, Cyg_ErrNo *);


// Set policy to apply to level changes on ADC.
extern Cyg_ErrNo
CS42XX_ConfigLCPolicy_ADC(CS42XX_DEVID, CS42XX_LVLCHGPOLICY, CS42XX_CACHESYNC);

// Set policy to apply to level changes on DAC.
extern Cyg_ErrNo
CS42XX_ConfigLCPolicy_DAC(CS42XX_DEVID, CS42XX_LVLCHGPOLICY, CS42XX_CACHESYNC);


// --- In-codec filters ---


// Get if high-pass filter is activated on ADC input.
extern cyg_bool
CS42XX_IsFiltered_ADC(CS42XX_DEVID, Cyg_ErrNo *perr);

// Get if de-emphasis filter is activated on DAC output.
extern cyg_bool
CS42XX_IsFiltered_DAC(CS42XX_DEVID, Cyg_ErrNo *perr);


// Activate or deactivate high-pass filter on ADC input.
extern Cyg_ErrNo
CS42XX_SetFilter_ADC(CS42XX_DEVID, cyg_bool filter, CS42XX_CACHESYNC);

// Activate or deactivate de-emphasis filter on DAC output.
extern Cyg_ErrNo
CS42XX_SetFilter_DAC(CS42XX_DEVID, cyg_bool filter, CS42XX_CACHESYNC);


// --- Signal selection ---


// Get the analog input used for ADC.
extern CS42XX_AINSEL
CS42XX_GetADCInput(CS42XX_DEVID, Cyg_ErrNo *);

// Get the source of signal played on auxiliary output.
extern CS42XX_AOUTSEL
CS42XX_GetAuxOut(CS42XX_DEVID, Cyg_ErrNo *);

// Get if "Loop mode" is activated.
extern cyg_bool
CS42XX_GetLoopMode(CS42XX_DEVID, Cyg_ErrNo *);


// Select the appropriate analog input for ADC.
extern Cyg_ErrNo
CS42XX_ConfigADCInput(CS42XX_DEVID, CS42XX_AINSEL source, CS42XX_CACHESYNC);

// Switch the analog input for ADC; mute input during switch.
extern Cyg_ErrNo
CS42XX_SwitchADCInput(CS42XX_DEVID, CS42XX_AINSEL source);

// Select the signal to be played on auxiliary output.
extern Cyg_ErrNo
CS42XX_ConfigAuxOut(CS42XX_DEVID, CS42XX_AOUTSEL source, CS42XX_CACHESYNC);

// Activate or deactivate "Loop mode".
extern Cyg_ErrNo
CS42XX_ConfigLoopMode(CS42XX_DEVID, cyg_bool loopmode, CS42XX_CACHESYNC);



#endif // _CS42XX_H_
