//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
 ****************************************************************************
 * Driver for Cirrus Logic CS42XX stereo audio codec.
 *
 * @file   cs42xx_i2c.c
 * @brief  Interface with IC driver.
 *
 * Code is based on the following datasheets:
 * http://www.cirrus.com/en/pubs/proDatasheet/CS4245_F1.pdf
 * http://www.cirrus.com/en/pubs/proDatasheet/CS4265_F1.pdf
 * http://www.cirrus.com/en/pubs/proDatasheet/CS5345_F1.pdf
 *  
 * @author David Tylski
 * @date   Jan. 2007
 * 
 * Copyright (C) 2007, Parrot S.A. - www.parrot.biz
 ****************************************************************************
 */

#include <cyg/infra/cyg_ass.h>
#include <cyg/io/i2c.h>
#include <cyg/devs/cs42xx.h>
#include <cyg/devs/cs42xx_i2c.h>


/*
 * Macros.
 **********/

/** Declare a CS42XX device. */
#define CS42XX_I2C_DEVICE()						\
	{											\
		.i2c_bus     = NULL,					\
		.i2c_address = 0x00,					\
		.i2c_flags   = 0x00,					\
		.i2c_delay   = CYG_I2C_DEFAULT_DELAY	\
	}

/** Get the IC address of a CS42XX device on IC bus from ad0 and ad1.
 *
 * @def   CS42XX_I2C_ADDR
 * @param ad1  State of AD1 pin on device (boolean)
 * @param ad0  State of AD0 pin on device (boolean)
 ****************************************************************************/
#define CS42XX_I2C_ADDR(ad1,ad0)		(0x4C | ((ad1)<<1) | (ad0))


/*
 * Variables.
 *************/

/** List of CS42XX devices.
 *
 * @note  Naming convention: the name contains '_' for better readability.
 *        CS42XXI2CDevice would really not be helpful...
 *
 * @var CS42XX_I2C_Device
 */
cyg_i2c_device CS42XX_I2C_Device[CS42XX_NBDEVICES] = {
#if CS42XX_NBDEVICES < 1
#error "Invalid number of CS42XX devices configured."
#endif
		CS42XX_I2C_DEVICE(),
#if CS42XX_NBDEVICES > 1
		CS42XX_I2C_DEVICE(),
#endif
#if CS42XX_NBDEVICES > 2
		CS42XX_I2C_DEVICE(),
#endif
#if CS42XX_NBDEVICES > 3
		CS42XX_I2C_DEVICE(),
#endif
#if CS42XX_NBDEVICES > 4
#error "Too many CS42XX devices configured; please update driver."
#endif
};


/*
 * Definition of functions.
 ***************************/


/** Get the IC device associated to a symbolic device Id.
 *
 * @param devid  Symbolic device Id
 *
 * @return       Pointer to an IC device
 * @retval NULL  Invalid device Id
 ****************************************************************************/
static inline cyg_i2c_device *
CS42XX_I2C_GetDevice(CS42XX_DEVID devid)
{
	return ((devid < CS42XX_NBDEVICES)?
		&CS42XX_I2C_Device[devid] : NULL);
}


/** Attach a symbolic device to a bus.
 * This function "fills a slot" by setting address of device and bus to
 * which it is connected.
 *
 * @param devid  Symbolic device Id
 * @param bus    Bus to attach device to
 * @param ad0    State of ad0 pin on audio device
 * @param ad1    State of ad1 pin on audio device (set it to 1 for CS4265)
 *
 * @retval  ENOERR  Operation successful
 * @retval -EINVAL  Invalid parameters
 * @retval -ENODEV  No such device
 ****************************************************************************/
Cyg_ErrNo
CS42XX_I2C_AttachBus(CS42XX_DEVID devid, const cyg_i2c_bus *bus,
		cyg_bool ad0, cyg_bool ad1)
{
	Cyg_ErrNo       err = -EINVAL;
	cyg_i2c_device *dev;

	if (NULL != bus)
	{
		err = -ENODEV;
		dev = CS42XX_I2C_GetDevice(devid);
		if (dev)
		{
			dev->i2c_bus     = (cyg_i2c_bus *)bus;
			dev->i2c_address = CS42XX_I2C_ADDR(ad1,ad0);
			err = ENOERR;
		}
		// else invalid device Id.
	}
	// else invalid parameter.
	
	return err;
}


/** Write one or successive bytes in device memory, at given location.
 *
 * @param dev      IC device to access to
 * @param mapbyte  Address where to locate the byte(s)
 * @param data     Pointer to the byte(s) to be sent
 * @param datalen  Number of bytes to be written
 *
 * @retval  ENOERR   Operation successful
 * @retval -ENODEV   No such device
 * @retval -EIO      Error during transmission
 ****************************************************************************/
Cyg_ErrNo
CS42XX_I2C_Write(CS42XX_DEVID devid,
		CYG_BYTE addr, const CYG_BYTE *buf, cyg_count8 len)
{
	Cyg_ErrNo       err = -ENODEV;
	cyg_i2c_device *dev;
	cyg_uint32      cnt;
	
	dev = CS42XX_I2C_GetDevice(devid);
	if (dev && dev->i2c_address)
	{
		CYG_ASSERT(dev->i2c_bus, "Call CS42XX_Init() first");
		CYG_ASSERT(buf, "buf=NULL");
		
		// Lock IC
		cyg_i2c_transaction_begin(dev);
		
		{
			// START(, devaddr), address (no STOP)
			cnt = cyg_i2c_transaction_tx(dev, true, &addr, 1, false);
		}
		if (1 == cnt)
		{
			// (no START) buf[0], ..., STOP
			cnt = cyg_i2c_transaction_tx(dev, false, buf, len, true);
			if (len == cnt)
			{
				err = ENOERR;
			}
			else
			{
				// I/O error occurred on transmission
				err = -EIO;
			}
		}
		// else no device answered to this address (-ENODEV).
		
		// Unlock IC
		cyg_i2c_transaction_end(dev);
	}
	// else no such device (no slot or wrong IC address).

	return err;
}


/** Read one or successive bytes from device memory, at given location.
 *
 * Procedure begins by a WRITE, interrupted after the ACK of mapbyte.
 * Then a "new" frame begins with READ, *not followed by mapbyte*, and
 * data is sent from device until it is not acknowledged by us.
 *
 * @param dev   IC device to access to
 * @param addr  Address where to read byte(s)
 * @param buf   Pointer to a buffer to receive data
 * @param len   Number of bytes to be read
 * 
 * @retval  ENOERR   Operation successful
 * @retval -ENODEV   No such device
 * @retval -EIO      Error during transmission or reception
 ****************************************************************************/
Cyg_ErrNo
CS42XX_I2C_Read(CS42XX_DEVID devid,
		CYG_BYTE addr, CYG_BYTE *buf, cyg_count8 len)
{
	Cyg_ErrNo       err = -ENODEV;
	cyg_i2c_device *dev;
	cyg_uint32      cnt;
	
	dev = CS42XX_I2C_GetDevice(devid);
	if (dev && dev->i2c_address)
	{
		CYG_ASSERT(dev->i2c_bus, "Call CS42XX_Init() first");
		CYG_ASSERT(buf, "buf=NULL");
		
		// Lock IC
		cyg_i2c_transaction_begin(dev);
		
		{
			// START(, devaddr), address, STOP
			cnt = cyg_i2c_transaction_tx(dev, true, &addr, 1, true);
		}
		if (1 == cnt)
		{
			// START(, devaddr), buf[0], ..., STOP
			cnt = cyg_i2c_transaction_rx(dev, true, buf, len, true, true);
			if (len == cnt)
			{
				err = ENOERR;
			}
			else
			{
				// I/O error occurred on transmission
				err = -EIO;
			}
		}
		// else no device answered to this address (-ENODEV).
		
		// Unlock IC
		cyg_i2c_transaction_end(dev);
	}
	// else no such device (no slot or wrong IC address).

	return err;
}


/**
@mainpage  CS42XX audio codec driver for eCos

This doxygen-generated documentation explains the internals of CS42XX audio
codec driver, both for the application implementing the driver and also for
driver maintaining.

@section howtouse_doc  How to use this driver in application

Please refer to the document below, available on CVS in the eCos repository:

<a href='http://canari/cgi-bin/cvsweb/~checkout~/SoftParrot5/ecos/ecos-current/packages/devs/i2c/components/cs42xx/current/doc/CS42XX%20driver%20-%20Description%20and%20use.doc?rev=1.4;content-type=application%2Fmsword'>
CS42XX driver - Description and use.doc</a>

*/

