//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
 ****************************************************************************
 * Driver for Cirrus Logic CS42XX stereo audio codec.
 *
 * @file   cs42xx_testctrl.c
 * @brief  Tests applicable on CS42XX driver - Driver control.
 *
 * @author David Tylski
 * @date   Mar. 2007
 *
 * Copyright (C) 2007, Parrot S.A. - www.parrot.biz
 ****************************************************************************
 */


#include <cyg/devs/cs42xx_api.h>

#include <cyg/devs/test/cs42xx_test.h>
#include <cyg/devs/test/cs42xx_testhmi.h>
#include <cyg/devs/test/cs42xx_testctrl.h>



#ifdef CS42XX_TEST_SUPPORT



/*
 * Defines.
 ***********/

#define CS42XX_TESTCTRL_THRPRIORITY			4
#define CS42XX_TESTCTRL_STACKSIZE			4096


/*
 * Types.
 *********/

/** Structure for the storage of data related to I�C testing.
 * @typedef CS42XX_TESTCTRL_DATA
 */
typedef struct CS42XX_TESTCTRL_DATA_TAG CS42XX_TESTCTRL_DATA;
struct CS42XX_TESTCTRL_DATA_TAG
{
	CS42XX_TEST_ID testcurr;  ///< Id of test currently running.
	CS42XX_TEST_ID testreq;   ///< Id of test requested.
	cyg_mutex_t    testmutex; ///< Mutex for testcurr and testreq.

	cyg_thread     threadobj;                        ///< Thread object.
	cyg_handle_t   thread;                           ///< Handle of thread.
	char           stack[CS42XX_TESTCTRL_STACKSIZE]; ///< Stack.

	cyg_mbox       mailboxobj; ///< Mailbox object.
	cyg_handle_t   mailbox;    ///< Handle of mailbox.
};



/*
 * Variables.
 *************/

/** Data related to control testing. */
static CS42XX_TESTCTRL_DATA CS42XXTestCtrlData;



/*
 * Definition of functions.
 ***************************/



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                        TESTS ON DRIVER CONTROLS                         *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/** Switch to the next analog input.
 *
 * In case of error detected, test score is affected.
 *
 * @retval CYGNUM_TEST_PASS  Test passed
 * @retval CYGNUM_TEST_FAIL  Failure during test
 ****************************************************************************/
static Cyg_test_code
CS42XX_TestCtrl_SwitchInput(void)
{
	Cyg_test_code result;
	CS42XX_DEVID  devid;
	Cyg_ErrNo     err;
	CS42XX_AINSEL ainput;

	devid  = CS42XX_Test_GetDeviceId();
	ainput = CS42XX_GetADCInput(devid, &err);
	if (ENOERR == err)
	{
		// Step to the next analog input
		if (ainput+1 < CS42XX_Supp_GetNbAnalogInputs(devid))
		{
			ainput++;
		}
		else
		{
			ainput = 0;
		}

		err = CS42XX_ConfigADCInput(devid, ainput, CS42XX_CACHE_SYNC);
	}

	// Score the test
	if (ENOERR == err)
	{
		CS42XX_TEST_PRINTF("Switch Input: %d", ainput);
		result = CYGNUM_TEST_PASS;
	}
	else
	{
		result = CYGNUM_TEST_FAIL;
	}

	return result;
}


/** Mute/unmute input.
 *
 * In case of error detected, test score is affected.
 *
 * @retval CYGNUM_TEST_PASS  Test passed
 * @retval CYGNUM_TEST_FAIL  Failure during test
 ****************************************************************************/
static Cyg_test_code
CS42XX_TestCtrl_MuteInput(void)
{
	Cyg_test_code result = CYGNUM_TEST_FAIL;
	CS42XX_DEVID  devid;
	Cyg_ErrNo     err;
	cyg_bool          mute, newmute;

	devid = CS42XX_Test_GetDeviceId();
	mute  = CS42XX_IsMuted_ADC(devid, &err);
	if (ENOERR == err)
	{
		// Switch MUTE state
		mute = !mute;
		err = CS42XX_Mute_ADC(devid, mute, CS42XX_CACHE_SYNC);

		newmute = CS42XX_IsMuted_ADC(devid, NULL);
		CS42XX_TEST_PRINTF("MuteADC=%d", newmute);

		// Score the test
		if ((ENOERR == err) && (newmute == mute))
		{
			result = CYGNUM_TEST_PASS;
		}
		// else test failed.
	}
	// else test failed.

	return result;
}


/** Mute/unmute output.
 *
 * In case of error detected, test score is affected.
 *
 * @retval CYGNUM_TEST_PASS  Test passed
 * @retval CYGNUM_TEST_FAIL  Failure during test
 ****************************************************************************/
static Cyg_test_code
CS42XX_TestCtrl_MuteOutput(void)
{
	Cyg_test_code result = CYGNUM_TEST_FAIL;
	CS42XX_DEVID  devid;
	Cyg_ErrNo     err;
	cyg_bool          mute, newmute;

	devid = CS42XX_Test_GetDeviceId();
	mute  = CS42XX_IsMuted_DAC(devid, &err);
	if (ENOERR == err)
	{
		// Switch MUTE state
		mute = !mute;
		err = CS42XX_Mute_DAC(devid, mute, CS42XX_CACHE_SYNC);

		newmute = CS42XX_IsMuted_DAC(devid, NULL);
		CS42XX_TEST_PRINTF("MuteDAC=%d", newmute);

		// Score the test
		if ((ENOERR == err) && (newmute == mute))
		{
			result = CYGNUM_TEST_PASS;
		}
		// else test failed.
	}
	// else test failed.

	return result;
}


/** Increase or decrease input gain.
 *
 * In case of error detected, test score is affected.
 *
 * @param incr   Signed value for increment/decrement
 *
 * @retval CYGNUM_TEST_PASS  Test passed
 * @retval CYGNUM_TEST_FAIL  Failure during test
 ****************************************************************************/
static Cyg_test_code
CS42XX_TestCtrl_ChangeGainIn(cyg_int8 incr)
{
	Cyg_test_code result = CYGNUM_TEST_FAIL;
	CS42XX_DEVID  devid;
	Cyg_ErrNo     err, err2;
	cyg_int8            gain[2], newgain[2], i;

	devid  = CS42XX_Test_GetDeviceId();
	gain[0] = CS42XX_GetGain_ADCA(devid, &err);
	gain[1] = CS42XX_GetGain_ADCB(devid, &err2);
	if ((ENOERR == err) && (ENOERR == err2))
	{
		// Compute increased or decreased volume levels
		for (i = 0; i < 2; i++)
		{
			if (incr < 0)
			{
				// Decrement (incr < 0)
				if (gain[i] > (-128 - incr)) {
					gain[i] += incr;  // This will decrement gain[i]...
				} else {
					gain[i] = -128;
				}
			}
			else
			{
				// Increment (incr > 0)
				if (gain[i] < 127 - incr) {
					gain[i] += incr;
				} else {
					gain[i] = 127;
				}
			}
		}

		// Apply the new gain on both channels
		err = CS42XX_SetGain_ADC(devid, gain[0], gain[1],
				CS42XX_CACHE_SYNC);

		// Get settings currently in use
		newgain[0] = CS42XX_GetGain_ADCA(devid, NULL);
		newgain[1] = CS42XX_GetGain_ADCB(devid, NULL);

		CS42XX_TEST_PRINTF("GainADC=(%d,%d)", newgain[0], newgain[1]);

		// Score the test
		switch(err)
		{
		case ENOERR:
			// Success (as told by driver)
			if ((newgain[0] == gain[0]) && (newgain[1] == gain[1]))
			{
				result = CYGNUM_TEST_PASS;
			}
			// else test failed.
			break;

		default:
			// Test failed.
			break;
		}
	}
	// else test failed.

	return result;
}


/** Increase or decrease output volume.
 *
 * In case of error detected, test score is affected.
 *
 * @param incr   Signed value for increment/decrement
 *
 * @retval CYGNUM_TEST_PASS  Test passed
 * @retval CYGNUM_TEST_FAIL  Failure during test
 ****************************************************************************/
static Cyg_test_code
CS42XX_TestCtrl_ChangeVolumeOut(cyg_int8 incr)
{
	Cyg_test_code result = CYGNUM_TEST_FAIL;
	CS42XX_DEVID  devid;
	Cyg_ErrNo     err, err2;
	cyg_uint8            vol[2], newvol[2], i;

	devid  = CS42XX_Test_GetDeviceId();
	vol[0] = CS42XX_GetVolume_DACA(devid, &err);
	vol[1] = CS42XX_GetVolume_DACB(devid, &err2);
	if ((ENOERR == err) && (ENOERR == err2))
	{
		// Compute increased or decreased volume levels
		for (i = 0; i < 2; i++)
		{
			if (incr < 0)
			{
				// Decrement (incr < 0)
				if (vol[i] > (cyg_uint8)(-incr)) {
					vol[i] += incr;  // This will decrement vol[i]...
				} else {
					vol[i] = 0;
				}
			}
			else
			{
				// Increment (incr > 0)
				if (vol[i] < (cyg_uint8)(0xFF - incr)) {
					vol[i] += incr;
				} else {
					vol[i] = 0xFF;
				}
			}
		}

		// Apply the new volume on both channels
		err = CS42XX_SetVolume_DAC(devid, vol[0], vol[1],
				CS42XX_CACHE_SYNC);

		// Get settings currently in use
		newvol[0] = CS42XX_GetVolume_DACA(devid, NULL);
		newvol[1] = CS42XX_GetVolume_DACB(devid, NULL);

		CS42XX_TEST_PRINTF("VolDAC=(%d,%d)", newvol[0], newvol[1]);

		// Score the test
		switch(err)
		{
		case ENOERR:
			// Success (as told by driver)
			if ((newvol[0] == vol[0]) && (newvol[1] == vol[1]))
			{
				result = CYGNUM_TEST_PASS;
			}
			// else test failed.
			break;

		default:
			// Test failed.
			break;
		}
	}
	// else test failed.

	return result;
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                MANAGEMENT OF TESTS ON DRIVER CONTROLS                   *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/** Thread managing control tests on the driver of CS42XX audio codec.
 *
 * According to its variable configuration, this thread starts and stops
 * control tests.
 *
 * @param data  (Not used)
 ****************************************************************************/
static void
CS42XX_TestCtrl_Thread(CYG_ADDRWORD data)
{
	Cyg_test_code  testresult;

	while(1)
	{
		switch(CS42XXTestCtrlData.testcurr)
		{

		// --- ADC ---

		case CS42XX_TEST_CTRL_ADC_SWITCH:
			testresult = CS42XX_TestCtrl_SwitchInput();
			break;

		case CS42XX_TEST_CTRL_ADC_MUTE:
			testresult = CS42XX_TestCtrl_MuteInput();
			break;

		case CS42XX_TEST_CTRL_ADC_GAINUP:
			testresult = CS42XX_TestCtrl_ChangeGainIn(+1);
			break;

		case CS42XX_TEST_CTRL_ADC_GAINDOWN:
			testresult = CS42XX_TestCtrl_ChangeGainIn(-1);
			break;

		// --- DAC ---

		case CS42XX_TEST_CTRL_DAC_MUTE:
			testresult = CS42XX_TestCtrl_MuteOutput();
			break;

		case CS42XX_TEST_CTRL_DAC_VOLUMEUP:
			testresult = CS42XX_TestCtrl_ChangeVolumeOut(+1);
			break;

		case CS42XX_TEST_CTRL_DAC_VOLUMEDOWN:
			testresult = CS42XX_TestCtrl_ChangeVolumeOut(-1);
			break;

		case CS42XX_TEST_CTRL_DAC_SWITCHAUX:
		//case CS42XX_TEST_NONE:
		default:
			break;
		}

		// No test: block until a new test is asked
		{
			void *ptr;
			ptr = cyg_mbox_get(CS42XXTestCtrlData.mailbox);

			cyg_mutex_lock(&CS42XXTestCtrlData.testmutex);
			CS42XXTestCtrlData.testcurr = *(CS42XX_TEST_ID *)ptr;
			cyg_mutex_unlock(&CS42XXTestCtrlData.testmutex);
		}

		// Post test results
		if (CYGNUM_TEST_NA != testresult)
		{
			CS42XX_TestHMI_PostResult(testresult);
		}

		testresult = CYGNUM_TEST_NA;
	}
	// Loop infinitely.
}


/** Look for pending messages in mailbox; if any, take the first message
 *  and store it as the new test Id.
 *
 * This function is not blocking.
 *
 * @retval true   A message has been found and stored as new test Id
 * @retval false  No message pending
 ****************************************************************************/
static cyg_bool
CS42XX_TestCtrl_TestMailbox(void)
{
	cyg_bool newmsg = false;
	void *ptr;

	cyg_mutex_lock(&CS42XXTestCtrlData.testmutex);
	ptr = cyg_mbox_tryget(CS42XXTestCtrlData.mailbox);
	if (NULL != ptr)
	{
		// New message available
		newmsg = true;

		// Store it as a new test Id
		CS42XXTestCtrlData.testcurr = *(CS42XX_TEST_ID *)ptr;
	}
	cyg_mutex_unlock(&CS42XXTestCtrlData.testmutex);

	return newmsg;
}


/** Select the control test to be run.
 *
 * @param[in] testid  Control test to be run
 ****************************************************************************/
 void
CS42XX_TestCtrl_Run(CS42XX_TEST_ID testid)
{
	// Post a new command to the thread
	cyg_mutex_lock(&CS42XXTestCtrlData.testmutex);
	CS42XXTestCtrlData.testreq = testid;
	cyg_mbox_put(CS42XXTestCtrlData.mailbox,
				&CS42XXTestCtrlData.testreq);
	CS42XX_TEST_PRINTF("[Post:%d]", CS42XXTestCtrlData.testreq);
	cyg_mutex_unlock(&CS42XXTestCtrlData.testmutex);
}


/** Initialize control testing.
 *
 * @warning  This part is platform-specific.
 *
 * @retval true   Initialization complete, thread launched
 ****************************************************************************/
 cyg_bool
CS42XX_TestCtrl_Init(void)
{
	// Initialize global variables
	CS42XXTestCtrlData.testcurr = CS42XX_TEST_NONE;
	CS42XXTestCtrlData.testreq  = CS42XX_TEST_NONE;
	cyg_mutex_init(&CS42XXTestCtrlData.testmutex);

	// Create a mailbox to command the control testing thread
	cyg_mbox_create(&CS42XXTestCtrlData.mailbox,
					&CS42XXTestCtrlData.mailboxobj);

	// Start the control testing thread
	cyg_thread_create(CS42XX_TESTCTRL_THRPRIORITY,
			CS42XX_TestCtrl_Thread, 0, "CS42XX control testing",
			&CS42XXTestCtrlData.stack[0], CS42XX_TESTCTRL_STACKSIZE,
			&CS42XXTestCtrlData.thread, &CS42XXTestCtrlData.threadobj);
	cyg_thread_resume(CS42XXTestCtrlData.thread);

	return true;
}


#endif // CS42XX_TEST_SUPPORT
