//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
 ****************************************************************************
 * Driver for Wolfson WM8960 stereo audio codec.
 *
 * @file   src/wm8960.c
 * @brief  Wolfson WM8960 driver.
 *
 * Code is based on the WM8960.pdf in the doc directory of this package
 *
 * This driver is using the i2c bus to configure the codec
 *
 * @author Gregoire ETIENNE
 * @date   2007-03-23
 *
 * Copyright (C) 2007, Parrot S.A. - www.parrot.biz
 ****************************************************************************
 */

#include <cyg/io/devtab.h>
#include <cyg/infra/cyg_ass.h>
#include <cyg/io/i2c.h>
#include <cyg/devs/wm8960_regs.h>                 // WM8960 Audio Codec configuration


static cyg_bool  wm8960_init     ( struct cyg_devtab_entry * );
static Cyg_ErrNo wm8960_lookup   ( struct cyg_devtab_entry **tab,
                                   struct cyg_devtab_entry *sub_tab,
                                   const char              *name );
static Cyg_ErrNo wm8960_getconfig(cyg_io_handle_t handle,
                                  cyg_uint32      key,
                                  void            *buf,
                                  cyg_uint32      *len);
static Cyg_ErrNo wm8960_setconfig(cyg_io_handle_t handle,
                                  cyg_uint32      key,
                                  const void      *buf,
                                  cyg_uint32      *len);

/** Item for the table of devices. */
DEVIO_TABLE(cyg_io_wm8960_devio, NULL, NULL, 0, wm8960_getconfig, wm8960_setconfig);
/** Operations applicable to the device. */
DEVTAB_ENTRY(wm8960,
             "/dev/wm8960",
             0,                     // Does not depend on a lower level interface
             &cyg_io_wm8960_devio,
             wm8960_init,
             wm8960_lookup,
             NULL
    );


// -------------------------------------------------------------------------
//
// Configuration of the WM8960 Audio Codec
//
CYG_I2C_DEVICE(my_codec,
                  &cyg_i2c_buses[0],
                  WM8960_ADDRESS,
                  0x00,
                  CYG_I2C_DEFAULT_DELAY);

/** \brief Full codec configuration
 *
 * Here is the description of the init configuration
 * It mixes Left inputs 1&2 ont the left channel,
 * Right inputs 1&2 ont the right channel. Then enables
 * L&R Headphone and L&R Speaker Output
 *
 */

wm8960_config_tab_t wm8960_init_tab[] =
{ { CODEC_R0_LIN_VOL       , IPVU|(LINVOL(0x30)) },
  { CODEC_R1_RIN_VOL       , IPVU|(RINVOL(0x30)) },
  { CODEC_R2_LOUT_VOL      , (OUT1VU|LOUT1VOL(0x70)) },
  { CODEC_R3_ROUT_VOL      , (OUT1VU|ROUT1VOL(0x70)) },
  { CODEC_R5_ADC_DAC_CTRL1 , 0x0000 },
  { CODEC_R9_AUDIO_ITF2    , 0x0000/*|LOOPBACK*/ },
  { CODEC_R25_POWER_MNGT1  , (VMIDSEL(2)|VREF|AINL|AINR|ADCL|ADCR|MICB) },
  { CODEC_R26_POWER_MNGT2  , (DACL|DACR|SPKL|SPKR|LOUT1|ROUT1|OUT3) },
  { CODEC_R32_ADCL_SIGPATH , (LMN1|/*LMP2|*/LMP3|LMICBOOST_20DB|LMIC2B) },
  { CODEC_R33_ADCR_SIGPATH , (RMN1|/*RMP2|*/RMP3|RMICBOOST_20DB|RMIC2B) },
  { CODEC_R34_LOUT_MIX     , (LD2LO)         }, // 0x0100
  { CODEC_R37_ROUT_MIX     , (RD2RO)         }, // 0x0100
  { CODEC_R38_MONO_OUT_MIX1, 0x0000 },
  { CODEC_R39_MONO_OUT_MIX2, 0x0000 },
  { CODEC_R40_L_SPK_VOL    , (SPKVU|SPKLVOL(0x70))},
  { CODEC_R41_R_SPK_VOL    , (SPKVU|SPKRVOL(0x70))},
  { CODEC_R43_LINBSTMIX    , 0x0000 },
  { CODEC_R44_RINBSTMIX    , 0x0000 },
  { CODEC_R45_L_BYPASS     , 0x0000 },
  { CODEC_R46_R_BYPASS     , 0x0000 },
  { CODEC_R47_POWER_MNGT3  , (LMIC|RMIC|LOMIX|ROMIX) },
  { CODEC_R49_CLASS_D_CTRL1, SPKOPEN(0x3) },
  { 0xFF, 0xFFFF }
};

/** \brief Send a command to the codec through the i2c bus
 *
 * Address and data are combined to be sent in 2 bytes
 * on the i2c bus
 *
 * @param addr Register address (coded on 7 bits)
 * @param data A pointer to the data (coded on 9 bits)
 */
void wm8960_i2c_cmd ( cyg_uint8 addr , cyg_uint16 * data)
{
  cyg_uint8 tx_byte;
  cyg_uint32 sent;

  addr = ( (addr << 1) | ((*data) >> 8 ) ) & 0xFF ;
  tx_byte = (*data) & 0xFF;

  cyg_i2c_transaction_begin(&my_codec); // lock
  sent = cyg_i2c_transaction_tx(&my_codec, true, &addr, 1, false);   // send the address: start, no stop
  CYG_ASSERT(sent == 1 , "ERROR : Sending on i2c");
  sent = cyg_i2c_transaction_tx(&my_codec, false, &tx_byte, 1, true);// send the data: start, stop
  CYG_ASSERT(sent == 1 , "ERROR : Sending on i2c");
  cyg_i2c_transaction_end(&my_codec);// unlock
}

/** \brief Initialize the WM8960 codec
 *
 * Configure the i2s bus.
 * Configure the i2c bus master clock with the value defined in the configtool
 *
 */
static cyg_bool wm8960_init( struct cyg_devtab_entry *tab )
{
   Cyg_ErrNo err = ENOERR;
   int i = 0;

//   SELECT_PIN(AC_OUT1a);
//   SELECT_PIN(AC_IN1a);
//   SELECT_PIN(AC_CLKa);
//   SELECT_PIN(AC_SYNCa);
//   SELECT_PIN(MCLKa);
//   SELECT_PIN(SCL1);
//   SELECT_PIN(SDA1);

   do{
      wm8960_i2c_cmd( wm8960_init_tab[i].addr, &(wm8960_init_tab[i].val) );
      i++;
   }while (wm8960_init_tab[i].addr != 0xFF);

   return err;
}




/** \brief Lookup interface of WM8960 codec
 *
 *
 */
static Cyg_ErrNo wm8960_lookup(struct cyg_devtab_entry **tab,
                          struct cyg_devtab_entry *sub_tab,
                          const char              *name)
{
   Cyg_ErrNo err = ENOERR;
   return err;
}



/** \brief SetConfig interface of WM8960 codec
 *
 *
 */
static Cyg_ErrNo wm8960_setconfig( cyg_io_handle_t handle,
                              cyg_uint32      key,
                              const void      *buf,
                              cyg_uint32      *len)
{



   Cyg_ErrNo err = ENOERR;
   return err;

}



/** \brief GetConfig interface of WM8960 codec
 *
 *
 */
static Cyg_ErrNo wm8960_getconfig(cyg_io_handle_t handle,
                                  cyg_uint32      key,
                                  void            *buf,
                                  cyg_uint32      *len)
{
   Cyg_ErrNo err = ENOERR;
   return err;

}


