//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
 * @file   lpo.c
 * @brief  Parrot5 LPO power management and event handling
 *
 * @author ivan.djelic@parrot.com
 * @date   2008-03-27
 *
 * This file is licensed under the eCos Public License. A copy of this license
 * can be obtained from:
 * http://ecos.sourceware.org/license-overview.html
 *
 * The Original Code is eCos - Embedded Configurable Operating System,
 *
 * Portions created by Parrot S.A. are
 * Copyright (C) 2004-2008 Parrot S.A.
 */

#include <task_priorities.h>
#include <pkgconf/hal.h>                // Platform specific configury
#include <pkgconf/system.h>
#include CYGBLD_HAL_PLATFORM_H
#include CYGHWR_HAL_CUSTOM_CONFIG_H

#include <cyg/hal/hal_io.h>             // IO macros
#include <cyg/hal/drv_api.h>
#include <cyg/hal/hal_intr.h>           // interrupt enable/disable
#include <cyg/kernel/kapi.h>

#include <cyg/hal/p5dev.h>              // platform definitions
#include <cyg/hal/lpo.h>
#include <cyg/hal/hal_platform_ints.h>
#include <cyg/hal/hal_diag.h>

#include "parrot5_lpo.h"


struct lpo_handler {
    lpo_callback_t cb;                //!< LPO event callback function
    cyg_addrword_t cb_data;           //!< LPO event callback data
};

static cyg_interrupt lpo_irq;
static int lpo_init = 0;

/* Monitoring thread resources */
static cyg_thread thread;
static char stack[4096];
static cyg_flag_t lpo_flag;
static cyg_mutex_t lpo_lock;

static struct lpo_handler rtca;
static struct lpo_handler emcy;
static struct lpo_handler wkup;

static volatile int standby_lock = 0;

static cyg_uint32 lpo_isr(cyg_vector_t vector, cyg_addrword_t data)
{
    cyg_drv_interrupt_mask(vector);
    cyg_drv_interrupt_acknowledge(vector);
    return CYG_ISR_CALL_DSR;
}

static void lpo_dsr(cyg_vector_t vector,
                    cyg_ucount32 count,
                    cyg_addrword_t data)
{
    cyg_uint32 status, ctrl, flag = 0, ack = 0;

    LPO_READ(_LPO_STATUS, status);
    LPO_READ(_LPO_CONTROL, ctrl);

    /* LPO events are level-triggered, which means we have to disable an
       event as soon as it occurs */

    if (status & LPO_STATUS_ALARM_EV) {
        ack |= LPO_ITACK_RTCA;
        ctrl &= ~LPO_CTRL_ALARM_EN;
        flag |= LPO_EVENT_RTCA;
    }
    if (status & LPO_STATUS_EMERCY_EV) {
        ack |= LPO_ITACK_EMCY;
        ctrl &= ~LPO_CTRL_EMCY_EN;
        flag |= LPO_EVENT_EMCY;
    }
    if (status & LPO_STATUS_WKUP_EV) {
        ack |= LPO_ITACK_WKUP;
        ctrl &= ~LPO_CTRL_WKUP_EN;
        flag |= LPO_EVENT_WKUP;
    }

    /* acknowledge and disable events */
    LPO_WRITE(_LPO_ITACK, ack);
    LPO_WRITE(_LPO_CONTROL, ctrl);

    if (flag) {
        /* make sure we disable standby until event is fully processed */
        parrot5_lpo_standby_lock();
        /* wake-up monitoring thread */
        cyg_flag_setbits(&lpo_flag, flag);
    }

    /* re-enable interrupt requests */
    cyg_drv_interrupt_unmask(vector);
}

//! Button monitoring function
static void lpo_monitor(cyg_addrword_t data)
{
    cyg_flag_value_t flag;

    while (1) {

        /* wait for LPO event */
        flag = cyg_flag_wait(&lpo_flag, LPO_EVENT_ALL,
                             CYG_FLAG_WAITMODE_OR|CYG_FLAG_WAITMODE_CLR);

        cyg_mutex_lock(&lpo_lock);

        /* RTC alarm event */
        if ((flag & LPO_EVENT_RTCA) && rtca.cb) {
            rtca.cb(LPO_EVENT_RTCA, rtca.cb_data);
            rtca.cb = NULL;
        }

        /* Emergency event */
        if ((flag & LPO_EVENT_EMCY) && emcy.cb) {
            emcy.cb(LPO_EVENT_EMCY, emcy.cb_data);
            emcy.cb = NULL;
        }

        /* Wake-up event */
        if ((flag & LPO_EVENT_WKUP) && wkup.cb) {
            wkup.cb(LPO_EVENT_WKUP, wkup.cb_data);
            wkup.cb = NULL;
        }

        /* all events have been processed, re-enable standby */
        parrot5_lpo_standby_unlock();

        cyg_mutex_unlock(&lpo_lock);
    }
}


//! Setup LPO interrupt handling and managing thread
static void parrot5_lpo_init(void)
{
    cyg_handle_t handle;
    cyg_handle_t lpo_irq_handle;

    rtca.cb = NULL;
    emcy.cb = NULL;
    wkup.cb = NULL;

    standby_lock = 0;

    cyg_flag_init(&lpo_flag);
    cyg_mutex_init(&lpo_lock);

    /* disable everything and ack all events */
    LPO_WRITE(_LPO_CONTROL, 0);
    LPO_WRITE(_LPO_ITACK, LPO_ITACK_RTCA|LPO_ITACK_EMCY|LPO_ITACK_WKUP);

    /* install LPO interrupt handler */
    cyg_drv_interrupt_create(CYGNUM_HAL_INTERRUPT_RTCC,
                             4,
                             (cyg_addrword_t)0,
                             lpo_isr,
                             lpo_dsr,
                             &lpo_irq_handle,
                             &lpo_irq);

    cyg_drv_interrupt_attach(lpo_irq_handle);
    cyg_drv_interrupt_unmask(CYGNUM_HAL_INTERRUPT_RTCC);

    /* launch monitoring thread */
    cyg_thread_create(PRIO_ECOS_LPO_MONITOR, lpo_monitor, (cyg_addrword_t)0,
                      "LPO Monitor", (void *)stack, 4096, &handle, &thread);
    cyg_thread_resume(handle);
}


//! Register a callback on an LPO event
Cyg_ErrNo parrot5_lpo_event_register(int event, int pin_lvl, lpo_callback_t cb,
                                     cyg_addrword_t data, cyg_uint32 rtc_alarm)
{
    cyg_uint32 ctrl;
    Cyg_ErrNo ret = ENOERR;

    if (!lpo_init) {
        lpo_init = 1;
        parrot5_lpo_init();
    }

    cyg_mutex_lock(&lpo_lock);
    cyg_drv_dsr_lock();

    LPO_READ(_LPO_CONTROL, ctrl);

    switch (event) {

    case LPO_EVENT_RTCA:
        if (rtca.cb) {
            // callback should be unregistered first
            ret = -EINVAL;
        }
        else {
            rtca.cb = cb;
            rtca.cb_data = data;
            // program RTC alarm
            LPO_WRITE(_LPO_RTCA, rtc_alarm);
            // enable RTCA
            LPO_WRITE(_LPO_CONTROL, ctrl|LPO_CTRL_ALARM_EN);
        }
        break;

    case LPO_EVENT_EMCY:
        if (emcy.cb) {
            // callback should be unregistered first
            ret = -EINVAL;
        }
        else {
            emcy.cb = cb;
            emcy.cb_data = data;
            // setup pin polarity
            ctrl &= ~LPO_CTRL_EMCY_LVL;
            ctrl |= (pin_lvl)? LPO_CTRL_EMCY_LVL : 0;
            // enable Emergency event
            LPO_WRITE(_LPO_CONTROL, ctrl|LPO_CTRL_EMCY_EN);
        }
        break;

    case LPO_EVENT_WKUP:
        if (wkup.cb) {
            // callback should be unregistered first
            ret = -EINVAL;
        }
        else {
            wkup.cb = cb;
            wkup.cb_data = data;
            // setup pin polarity
            ctrl &= ~LPO_CTRL_WKUP_LVL;
            ctrl |= (pin_lvl)? LPO_CTRL_WKUP_LVL : 0;
            // enable Wake-up event
            LPO_WRITE(_LPO_CONTROL, ctrl|LPO_CTRL_WKUP_EN);
        }
        break;

    default:
        ret = -EINVAL;
        break;
    }

    cyg_drv_dsr_unlock();
    cyg_mutex_unlock(&lpo_lock);

    return ret;
}

//! Unregister a callback
void parrot5_lpo_event_unregister(int event)
{
    cyg_uint32 ctrl;

    cyg_mutex_lock(&lpo_lock);
    cyg_drv_dsr_lock();

    LPO_READ(_LPO_CONTROL, ctrl);

    switch (event) {

    case LPO_EVENT_RTCA:
        LPO_WRITE(_LPO_CONTROL, ctrl & ~LPO_CTRL_ALARM_EN);
        rtca.cb = NULL;
        break;

    case LPO_EVENT_EMCY:
        LPO_WRITE(_LPO_CONTROL, ctrl & ~LPO_CTRL_EMCY_EN);
        emcy.cb = NULL;
        break;

    case LPO_EVENT_WKUP:
        LPO_WRITE(_LPO_CONTROL, ctrl & ~LPO_CTRL_WKUP_EN);
        wkup.cb = NULL;
        break;
    }

    cyg_drv_dsr_unlock();
    cyg_mutex_unlock(&lpo_lock);
}

//-----------------------------------------------------------------------------
// Parrot5 LPO power management. Note: Emergency (PSA) mode not supported yet.
//-----------------------------------------------------------------------------

/** Switch the Parrot5 to STOP mode.
 *
 * In STOP mode, the ARM core is put in Wait-For-Interrupt mode. Clocks other
 * than the core clock should be disabled as needed.
 */
void
parrot5_lpo_stop(void)
{

    // disable core ARM clock
    HAL_WRITE_UINT32(PARROT5_SYS+_SYS_CDIS, SYS_CDIS_ARMCLK);
    // SDRAM in self-refresh mode
    HAL_WRITE_UINT32(PARROT5_SYS+_SYS_CDIS, SYS_CDIS_SELFREQ);

    // switch core to WFI mode
    asm volatile ("mcr p15,0,r0,c7,c0,4");
}

/**
 * Switch the Parrot5 to STANDBY mode.
 *
 * In STANDBY mode, main power is shutdown, providing the lowest possible
 * consumption.
 *
 * @param rtc_alarm        Program LPO for wake-up at RTC date 'rtc_alarm'.
 *                         Passing 0 will disable RTC alarm wake-up.
 * @param wakeup_enable    Enable/disable external line wake-up.
 * @param wakeup_level     Wake-up signal polarity level.
 * @param emergency_enable Enable/disable external emergency wake-up.
 * @param emergency_level  Emergency signal polarity level.
 * @param scratch          An 8-bit value which will be saved when entering
 *                         STANDBY mode. When exiting STANDBY mode, this value
 *                         can be retrieved with @parrot5_lpo_get_scratch.
 *
 * @return                 ENOERR on success, -EBUSY if an LPO event is pending
 */
Cyg_ErrNo parrot5_lpo_standby(cyg_uint32 rtc_alarm,
                              int wakeup_enable,
                              int wakeup_level,
                              int emergency_enable,
                              int emergency_level,
                              cyg_uint8 scratch)
{
    Cyg_ErrNo ret = ENOERR;
    cyg_uint32 reg, ctrl, status;

    /* program STANDBY exit conditions */
    ctrl = 0;
    ctrl |= (wakeup_enable   )? LPO_CTRL_WKUP_EN  : 0;
    ctrl |= (wakeup_level    )? LPO_CTRL_WKUP_LVL : 0;
    ctrl |= (emergency_enable)? LPO_CTRL_EMCY_EN  : 0;
    ctrl |= (emergency_level )? LPO_CTRL_EMCY_LVL : 0;
    ctrl |= (rtc_alarm       )? LPO_CTRL_ALARM_EN : 0;

    cyg_mutex_lock(&lpo_lock);

    /* enter critical section */
    HAL_DISABLE_INTERRUPTS(reg);

    LPO_READ(_LPO_STATUS, status);

    /* abort if some event is pending or standby is locked */
    if ((standby_lock > 0) ||
        (status &
         (LPO_STATUS_ALARM_EV|LPO_STATUS_EMERCY_EV|LPO_STATUS_WKUP_EV))) {
        ret = -EBUSY;
        goto out;
    }

    /* disable everything and ack all events */
    LPO_WRITE(_LPO_CONTROL, 0);
    LPO_WRITE(_LPO_ITACK, LPO_ITACK_RTCA|LPO_ITACK_EMCY|LPO_ITACK_WKUP);

    /* program RTC alarm */
    if (rtc_alarm) {
        LPO_WRITE(_LPO_RTCA, rtc_alarm);
    }

    /* set scratch register */
    LPO_WRITE(_LPO_SCRATCH, scratch);
    LPO_WRITE(_LPO_CONTROL, ctrl);

    /* synchronize with 32 kHz block */
    do {
        HAL_READ_UINT32(PARROT5_LPO+_LPO_CONTROL, reg);
    }
    while (reg != ctrl);

#if defined(CYGHWR_HAL_ARM_ARM9_P5DEV_STANDBY_TO_SDRAM)
    // go to sleep, this function will return when we are woken up
    parrot5_standby_to_sdram();
#else
    /* these magic values are documented in Parrot5 User Manual � 5.3.3 */
    HAL_WRITE_UINT32(PARROT5_SYS+_SYS_SDBYMODE, 0x2b56924d);
    HAL_WRITE_UINT32(PARROT5_SYS+_SYS_SDBYMODE, 0x4ad365b2);
    /* now wait... */
    while (1);
#endif

 out:
    /* exit critical section */
    HAL_RESTORE_INTERRUPTS(reg);

    cyg_mutex_unlock(&lpo_lock);

    return ret;
}

//! Read LPO scratch register
cyg_uint8 parrot5_lpo_get_scratch(void)
{
    cyg_uint32 scratch;
    LPO_READ(_LPO_SCRATCH, scratch);
    return (cyg_uint8)scratch;
}

/**
 * Lock standby operation, i.e. prevent the system from going to standby mode.
 * This is useful in order to eliminate race conditions when processing LPO
 * events (e.g. wake-up).
 * Note: this is implemented as a simple atomic counter.
 */
void parrot5_lpo_standby_lock(void)
{
    cyg_uint32 reg;

    HAL_DISABLE_INTERRUPTS(reg);
    standby_lock++;
    HAL_RESTORE_INTERRUPTS(reg);
}

/**
 * Unlock standby operation, i.e. allow the system to go to standby mode.
 * This is useful in order to eliminate race conditions when processing LPO
 * events (e.g. wake-up).
 * Note: this is implemented as a simple atomic counter.
 */
void parrot5_lpo_standby_unlock(void)
{
    cyg_uint32 reg;

    HAL_DISABLE_INTERRUPTS(reg);
    standby_lock--;
    HAL_RESTORE_INTERRUPTS(reg);
}
