#include "fatfs_supp.h"
#include "fatfs_robust.h"

// -------------------------------------------------------------------------
// cluster_to_block_pos()
// Converts cluster position to blib block position.

void
cluster_to_block_pos(fatfs_disk_t *disk,
                     cyg_uint32    cluster,
                     cyg_uint32    cluster_pos,
                     cyg_uint32   *block,
                     cyg_uint32   *block_pos)
{
    cyg_uint32 block_size      = disk->disk_info.block_size;
    cyg_uint32 block_size_log2 = disk->block_size_log2;

    *block = (cluster - 2) << (disk->cluster_size_log2 - block_size_log2);

    *block_pos = disk->fat_data_pos + cluster_pos;

    if (*block_pos > block_size)
    {
        *block     += *block_pos >> block_size_log2;
        *block_pos  = *block_pos & (block_size - 1);
    }
}

// -------------------------------------------------------------------------
// disk_read()
// Reads data from disk.

int
disk_read(fatfs_disk_t *disk,
          void         *buf,
          cyg_uint32   *len,
          cyg_uint32    pos)
{
    int ret = 0;

    if (disk->bcache_mem) {
         CYG_ASSERT(disk->disk_read, "NULL pointer!");
         ret = disk->disk_read(&disk->blib, buf, len, 0, pos);
    } else {
         CYG_ASSERT(disk->disk_read, "NULL pointer!");
         ret = disk->disk_read(NULL, buf, len, 0, pos);
    }

    return ret;
}


// -------------------------------------------------------------------------
// disk_cluster_read()
// Reads data from disk at specified cluster position.

static __inline__ int
disk_cluster_read(fatfs_disk_t *disk,
                  void         *buf,
                  cyg_uint32   *len,
                  cyg_uint32    cluster,
                  cyg_uint32    cluster_pos)
{
    cyg_uint32 block, block_pos;

    cluster_to_block_pos(disk, cluster, cluster_pos, &block, &block_pos);
    CYG_ASSERT(disk->disk_read, "NULL pointer!!");

    return disk->disk_read(&disk->blib, buf, len, block, block_pos);
}

// -------------------------------------------------------------------------
// gdate_to_jdays()
// Converts gregorian date to juilan days.

static void
gdate_to_jdays(int day, int month, int year, cyg_uint32 *jd)
{
    *jd = day - 32075 + 1461*(year + 4800 + (month - 14)/12)/4 +
          367*(month - 2 - (month - 14)/12*12)/12 -
          3*((year + 4900 + (month - 14)/12)/100)/4;
}

// -------------------------------------------------------------------------
// date_dos2unix()
// Converts dos time and date to unix timestamp.

static void
date_dos2unix(cyg_uint16  dos_time,
              cyg_uint16  dos_date,
              time_t *unix_timestamp)
{
    cyg_uint32 ts;
    int        hour, min, sec;
    int        day, month, year;

    sec        = (dos_time & ((1<<5)-1)) * 2;
    dos_time >>= 5;
    min        = (dos_time & ((1<<6)-1));
    dos_time >>= 6;
    hour       = dos_time;

    day        = (dos_date & ((1<<5)-1));
    dos_date >>= 5;
    month      = (dos_date & ((1<<4)-1));
    dos_date >>= 4;
    year       = dos_date + 1980;

    gdate_to_jdays(day, month, year, &ts);

    ts -= JD_1_JAN_1970;
    ts  = (ts * 24 * 3600) + (sec + min * 60 + hour * 3600);

    *unix_timestamp = ts;

    CYG_TRACE2(TDE, "dos time=%d date=%d", dos_time, dos_date);
    CYG_TRACE7(TDE, "timestamp=%d date=%d:%d:%d %d-%d-%d",
                    ts, hour, min, sec, year, month, day);
}

//==========================================================================
// FAT boot record functions
#if TDE
// -------------------------------------------------------------------------
// print_boot_record()
// Prints FAT boot record.
static void
print_boot_record(fat_boot_record_t* fbr)
{
    diag_printf("FAT: FBR jump code:       0x%02X\n", fbr->jump);
    diag_printf("FAT: FBR oem name:       '%.8s'\n",  fbr->oem_name);
    diag_printf("FAT: FBR bytes per sec:   %u\n",     fbr->bytes_per_sec);
    diag_printf("FAT: FBR sec per cluster: %u\n",     fbr->sec_per_clu);
    diag_printf("FAT: FBR reserved sec:    %u\n",     fbr->res_sec_num);
    diag_printf("FAT: FBR fat tbls num:    %u\n",     fbr->fat_tbls_num);
    diag_printf("FAT: FBR max root dents:  %u\n",     fbr->max_root_dents);
    diag_printf("FAT: FBR sec num (32):    %u\n",     fbr->sec_num_32);
    diag_printf("FAT: FBR media desc:      0x%02X\n", fbr->media_desc);
    diag_printf("FAT: FBR sec per fat:     %u\n",     fbr->sec_per_fat);
    diag_printf("FAT: FBR sec per track:   %u\n",     fbr->sec_per_track);
    diag_printf("FAT: FBR heads num:       %u\n",     fbr->heads_num);
    diag_printf("FAT: FBR hidden sec num:  %u\n",     fbr->hsec_num);
    diag_printf("FAT: FBR sec num:         %u\n",     fbr->sec_num);

    if (0 == fbr->sec_per_fat)
    {
        diag_printf("FAT: FBR sec per fat32:   %u\n",     fbr->sec_per_fat_32);
        diag_printf("FAT: FBR ext flags:       0x%04X\n", fbr->ext_flags);
        diag_printf("FAT: FBR fs ver:          %u\n",     fbr->fs_ver);
        diag_printf("FAT: FBR root cluster:    %u\n",     fbr->root_cluster);
        diag_printf("FAT: FBR fs info sec:     %u\n",     fbr->fs_info_sec);
    }

    diag_printf("FAT: FBR drv num:         %u\n",     fbr->drv_num);
    diag_printf("FAT: FBR ext sig:         0x%02X\n", fbr->ext_sig);
    diag_printf("FAT: FBR ser num:         0x%08X\n", fbr->ser_num);
    diag_printf("FAT: FBR vol name:       '%.11s'\n", fbr->vol_name);
    diag_printf("FAT: FBR fat name:       '%.8s'\n",  fbr->fat_name);
    diag_printf("FAT: FBR exe mark:        0x%02X 0x%02X\n",
                fbr->exe_marker[0], fbr->exe_marker[1]);
}
#endif // TDE

// -------------------------------------------------------------------------
// read_boot_record()
// Reads FAT boot record from disk.

static int
read_boot_record(fatfs_disk_t *disk, fat_boot_record_t *fbr)
{
    cyg_uint8 data[0x5A];
    cyg_uint32 len;
    int err;

    len = 0x5A;
    err = disk_read(disk, (void*)data, &len, 0);
    if (err != ENOERR)
        return err;

    GET_WORD(data,  fbr->jump,           0x00);
    GET_BYTES(data, fbr->oem_name, 8,    0x03);
    GET_WORD(data,  fbr->bytes_per_sec,  0x0B);
    GET_BYTE(data,  fbr->sec_per_clu,    0x0D);
    GET_WORD(data,  fbr->res_sec_num,    0x0E);
    GET_BYTE(data,  fbr->fat_tbls_num,   0x10);
    GET_WORD(data,  fbr->max_root_dents, 0x11);
    GET_WORD(data,  fbr->sec_num_32,     0x13);
    GET_BYTE(data,  fbr->media_desc,     0x15);
    GET_WORD(data,  fbr->sec_per_fat,    0x16);
    GET_WORD(data,  fbr->sec_per_track,  0x18);
    GET_WORD(data,  fbr->heads_num,      0x1A);
    GET_DWORD(data, fbr->hsec_num,       0x1C);
    GET_DWORD(data, fbr->sec_num,        0x20);
    GET_DWORD(data, fbr->vol_id,         0x39);

    // This is a quick check for FAT12/16 or FAT32 boot record.
    // The sec_per_fat field must be 0 on FAT32, since this
    // field plays a crucial role in detection of the FAT type
    // (12,16,32) it is quite safe to make this assumption.
    if (0 == fbr->sec_per_fat)
    {
        GET_DWORD(data, fbr->sec_per_fat_32, 0x24);
        GET_WORD(data,  fbr->ext_flags,      0x28);
        GET_WORD(data,  fbr->fs_ver,         0x2A);
        GET_DWORD(data, fbr->root_cluster,   0x2C);
        GET_WORD(data,  fbr->fs_info_sec,    0x30);
        GET_WORD(data,  fbr->bk_boot_sec,    0x32);
        GET_BYTE(data,  fbr->drv_num,        0x40);
        GET_BYTE(data,  fbr->ext_sig,        0x42);
        GET_DWORD(data, fbr->ser_num,        0x43);
        GET_BYTES(data, fbr->vol_name, 11,   0x47);
        GET_BYTES(data, fbr->fat_name, 8,    0x52);
    }
    else
    {
        GET_BYTE(data,  fbr->drv_num,        0x24);
        GET_BYTE(data,  fbr->ext_sig,        0x26);
        GET_DWORD(data, fbr->ser_num,        0x27);
        GET_BYTES(data, fbr->vol_name, 11,   0x2B);
        GET_BYTES(data, fbr->fat_name, 8,    0x36);
    }

    // Read the end marker
    len = 0x02;
    err = disk_read(disk, (void*)data, &len, 0x1FE);
    if (err != ENOERR)
        return err;

    GET_BYTES(data, fbr->exe_marker, 2,  0);

    // Zero terminate strings
    fbr->oem_name[8]  = '\0';
    fbr->vol_name[11] = '\0';
    fbr->fat_name[8]  = '\0';

#if TDE
    print_boot_record(fbr);
#endif

    return ENOERR;
}

//==========================================================================
// FAT table entry functions

// -------------------------------------------------------------------------
// read_tentry()
// Reads FAT table entry from disk.

int
read_tentry(fatfs_disk_t *disk, cyg_uint32 num, cyg_uint32 *entry)
{
    cyg_uint8  data[4];
    cyg_uint32 pos, num3;
    cyg_uint32 e;
    cyg_uint32 len;
    int err;

    switch (disk->fat_type)
    {
        case FATFS_FAT12:
            num3 = num * 3;
            pos  = disk->fat_tbl_pos + (num3 >> 1);
            len  = 2;

            err = disk_read(disk, (void*)data, &len, pos);
            if (err != ENOERR)
                return err;

            GET_WORD(data, e, 0x00);

            if (0 == (num3 & 1)) *entry = e        & 0x0FFF;
            else                 *entry = (e >> 4) & 0x0FFF;

            break;

        case FATFS_FAT16:
            pos = disk->fat_tbl_pos + (num << 1);
            len = 2;

            err = disk_read(disk, (void*)data, &len, pos);
            if (err != ENOERR)
                return err;

            GET_WORD(data, e, 0x00);
            *entry = e;

            break;

        case FATFS_FAT32:
            pos = disk->fat_tbl_pos + (num << 2);
            len = 4;
            err = disk_read(disk, (void*)data, &len, pos);

            if (err != ENOERR)
                return err;

            GET_DWORD(data, e, 0x00);
            *entry = e & 0x0FFFFFFF;
            break;

        default:
            err = EINVAL;
            CYG_ASSERT(false, "Unknown FAT type");
    }
    return ENOERR;
}

/** Init the read_n_tentry function
 *
 * This function MUST be called once before to call read_n_tentry
 * @return error
 */

void
read_n_tentry_init(fatfs_disk_t *disk)
{
   disk->fatfs_tentry_first = 0;
   disk->fatfs_tentry_last  = 0;
}

/** Reads FAT table entry from disk.
 *
 * read_n_tentry_init MUST MUST be called once before to call read_n_tentry
 * @param disk  A pointer to the disk
 * @param num   the entry number to read
 * @param entry A pointer to the entry to return
 * @return error
 */
int
read_n_tentry(fatfs_disk_t *disk, cyg_uint32 num, cyg_uint32 *entry)
{
    cyg_uint8  data[4];
    cyg_uint32 pos, num3;
    cyg_uint32 e;
    cyg_uint32 len;
    int err;

    switch (disk->fat_type)
    {
        case FATFS_FAT12:
            num3 = num * 3;
            pos  = disk->fat_tbl_pos + (num3 >> 1);
            len  = 2;

            err = disk_read(disk, (void*)data, &len, pos);
            if (err != ENOERR)
                return err;

            GET_WORD(data, e, 0x00);

            if (0 == (num3 & 1)) *entry = e        & 0x0FFF;
            else                 *entry = (e >> 4) & 0x0FFF;

            break;

        case FATFS_FAT16:
            pos = disk->fat_tbl_pos + (num << 1);
            len = 2;

            err = disk_read(disk, (void*)data, &len, pos);
            if (err != ENOERR)
                return err;

            GET_WORD(data, e, 0x00);
            *entry = e;

            break;

        case FATFS_FAT32:
            if (num < disk->fatfs_tentry_last)
            {
               num3 = (num-disk->fatfs_tentry_first)<<2;
               GET_DWORD(disk->fatfs_read_n_tentry_data, e, num3) ;
               *entry = e & 0x0FFFFFFF;
               err = ENOERR;
            }
            else
            {
               pos = disk->fat_tbl_pos + (num << 2);
               len = READ_N_SIZE;
               err = disk_read(disk, (void*)disk->fatfs_read_n_tentry_data, &len, pos);
               disk->fatfs_tentry_first = num;
               disk->fatfs_tentry_last  = num+(READ_N_SIZE>>2); // divided by 4
               GET_DWORD(disk->fatfs_read_n_tentry_data, e, 0x00);
               *entry = e & 0x0FFFFFFF;

            }

            if (err != ENOERR)
                return err;

            break;

        default:
            CYG_ASSERT(false, "Unknown FAT type");
    }
    return ENOERR;
}


// -------------------------------------------------------------------------
// get_tentry_type()
// Gets the type of FAT table entry.

int
get_tentry_type(fatfs_disk_t *disk, cyg_uint32 entry)
{
    int type;

    switch (disk->fat_type)
    {
        case FATFS_FAT12:
            if (entry < 0x0FF0)
            {
                if (0x0000 == entry)  type = TENTRY_FREE;
                else                  type = TENTRY_REGULAR;
            }
            else if (entry >= 0x0FF8) type = TENTRY_LAST;
            else if (0x0FF7 == entry) type = TENTRY_BAD;
            else                      type = TENTRY_RESERVED;

            break;

        case FATFS_FAT16:
            if (entry < 0xFFF0)
            {
                if (0x0000 == entry)  type = TENTRY_FREE;
                else                  type = TENTRY_REGULAR;
            }
            else if (entry >= 0xFFF8) type = TENTRY_LAST;
            else if (0xFFF7 == entry) type = TENTRY_BAD;
            else                      type = TENTRY_RESERVED;

            break;

        case FATFS_FAT32:

            if (entry < 0x0FFFFFF0)
            {
                if (0x00000000 == entry)  type = TENTRY_FREE;
                else                      type = TENTRY_REGULAR;
            }
            else if (entry >= 0x0FFFFFF8) type = TENTRY_LAST;
            else if (0x0FFFFFF7 == entry) type = TENTRY_BAD;
            else                          type = TENTRY_RESERVED;

            break;

        default:
            CYG_ASSERT(false, "Unknown FAT type");
            type = TENTRY_BAD; // least likely to cause damage
    }
    return type;
}

// -------------------------------------------------------------------------
// get_tentry_next_cluster()
// Gets the the next file cluster number from FAT table entry.

cyg_uint32
get_tentry_next_cluster(fatfs_disk_t *disk, cyg_uint32 entry)
{
    return entry;
}

// -------------------------------------------------------------------------
// find_nth_cluster0()
// Finds nth cluster in chain (ie nth cluster of file) searching
// from given position. The result is returned by the same position
// variable.

static int
find_nth_cluster0(fatfs_disk_t     *disk,
                  fatfs_data_pos_t *pos,
                  cyg_uint32        n)
{
    cyg_uint32 cluster, cluster_snum;
    int        err = ENOERR;

    if (pos->cluster_snum == n)
        return ENOERR;

    cluster      = pos->cluster;
    cluster_snum = pos->cluster_snum;

    CYG_TRACE4(TCL, "cluster=%d snum=%d n=%d n_to_search=%d",
                    cluster, cluster_snum, n, n-cluster_snum);

    // Adjust the number of clusters that should be
    // walked according to the given position
    n -= cluster_snum;

    // Walk the cluster chain for n clusters or until last cluster
    while (n > 0)
    {
        cyg_uint32 tentry;

        err = read_tentry(disk, cluster, &tentry);
        if (err != ENOERR)
            return err;

        switch (get_tentry_type(disk, tentry))
        {
            case TENTRY_REGULAR:
                break;
            case TENTRY_LAST:
                CYG_TRACE1(TCL, "chain end at n=%d", n);
                err = EEOF; // File has less clusters than given n
                            // this err should be caught by the
                            // calling function
                goto out;
            default:
                // Inconsistant FAT table state !!!
                CYG_TRACE2(TCL, "!!! inconsistant FAT tentry=%x n=%d",
                                tentry, n);
                err = EIO;
                goto out;
        }
        cluster = get_tentry_next_cluster(disk, tentry);
        cluster_snum++;
        n--;
    }

out:
    pos->cluster      = cluster;
    pos->cluster_snum = cluster_snum;

    CYG_TRACE2(TCL, "nth cluster=%d snum=%d", cluster, cluster_snum);

    return err;
}

// -------------------------------------------------------------------------
// find_nth_cluster()
// Finds nth cluster in chain (ie nth cluster of file) searching
// from given position. The result is returned by the same position
// variable. If the chain ends one cluster before the given nth cluster
// and the CO_EXTEND is specified, than the chain is extended by one cluster.

static int
find_nth_cluster(fatfs_disk_t     *disk,
                 fatfs_data_pos_t *pos,
                 cyg_uint32        n,
                 cluster_opts_t    opts)
{
    int err;

    // Find nth cluster
    err = find_nth_cluster0(disk, pos, n);
    // EEOF meens that the cluster chain ended early
    if ((err != EEOF) || !(opts & CO_EXTEND))
        return err;

    // Check if one cluster short
    if (pos->cluster_snum == (n - 1))
    {
        // Extend the chain for one cluster
        cyg_uint32 new_cluster;

        // Append new cluster to the end of chain
        err = find_and_append_cluster(disk, pos->cluster, &new_cluster, opts);
        if (err != ENOERR)
            return err;

        // Update position
        pos->cluster       = new_cluster;
        pos->cluster_snum += 1;
        pos->cluster_pos   = 0;

        CYG_TRACE1(TCL, "appended new cluster=%d", new_cluster);
    }

    return err;
}

// -------------------------------------------------------------------------
// get_next_cluster()
// Gets next cluster in chain (ie next cluster of file).
// If CO_EXTEND is specified and the current cluster is last in the
// chain then the chain is extended by one cluster.

int
get_next_cluster(fatfs_disk_t     *disk,
                 fatfs_data_pos_t *pos,
                 cluster_opts_t    opts)
{
    int err;

    err = find_nth_cluster(disk, pos, pos->cluster_snum + 1, opts);
    if (err != ENOERR)
        return err;

    // Reset inside cluster position
    pos->cluster_pos = 0;

    return ENOERR;
}

/** Browse cluster chain and returns the number of free clusters and the first cluster free.
 *
 * @param disk            A pointer to the disk
 * @param nbfree_clust    A pointer to a counter of free cluster
 * @param firstfree_clust A pointer to the number of the first free cluster
 * @return ENOERR in any case
 */
int browse_cluster_chain( fatfs_disk_t *disk,
                          cyg_uint32   *nbfree_clust,
                          cyg_uint32   *firstfree_clust)
{
    cyg_uint32 c, tentry;
    int        err;

    c = 2;
    *nbfree_clust = 0;
    *firstfree_clust = 0;

    read_n_tentry_init(disk);

    CYG_TRACE1(TCL, "starting at cluster=%d", c);

    // Search from the starting cluster to the end of FAT and
    // from start of FAT to the starting cluster
    while (c < disk->fat_tbl_nents)
    {
        err = read_n_tentry(disk, c, &tentry);

        if (err != ENOERR)
         {
            return err;
         }

        if (TENTRY_FREE == get_tentry_type(disk, tentry))
        {
            CYG_TRACE1(TCL, "found free cluster=%d", c);
            (*nbfree_clust)++;
            if( (*firstfree_clust) == 0 )
            {
               *firstfree_clust = c;
            }
        }
        c++;
    }

    return ENOERR;
}


// -------------------------------------------------------------------------
// get_position_from_off()
// Gets position from given offset. The search is started from the
// given position and the result is returned by the same variable.
// If CO_EXTEND is specified the file is extended if one cluster too short.

static int
get_position_from_off(fatfs_disk_t     *disk,
                      cyg_uint32        first_cluster,
                      cyg_uint32        offset,
                      fatfs_data_pos_t *pos,
                      cluster_opts_t    opts)
{
    fatfs_data_pos_t new_pos;
    cyg_uint32       n;
    int              err;

    // Position inside the cluster
    new_pos.cluster_pos = offset & (disk->cluster_size - 1);

    // Cluster seq number to be searched for
    n = offset >> disk->cluster_size_log2;

    if (n < pos->cluster_snum)
    {
        // Start searching from first cluster
        new_pos.cluster      = first_cluster;
        new_pos.cluster_snum = 0;
    }
    else
    {
        // Start searching from the current position
        new_pos.cluster      = pos->cluster;
        new_pos.cluster_snum = pos->cluster_snum;
    }

    err = find_nth_cluster(disk, &new_pos, n, opts);

    // Err could be EEOF wich means that the given
    // offset if out of given file (cluster chain)

    if (EEOF == err)
        new_pos.cluster_pos = disk->cluster_size;

    *pos = new_pos;

    return err;
}

// -------------------------------------------------------------------------
// print_raw_dentry()
// Prints FAT directory entry.

#if TDE
static void
print_raw_dentry(fat_raw_dir_entry_t* dentry)
{
#ifdef CYGCFG_FS_FAT_USE_VFAT
    if (DENTRY_IS_LNDIR(dentry))
        diag_printf("FAT: LDE name:    '%.255s'\n", dentry->lname);
#endif //CYGCFG_FS_FAT_USE_VFAT

    if (DENTRY_IS_DELETED(dentry))
        diag_printf("FAT: SDE name:    '?%.7s'\n", &dentry->name[1]);
    else
        diag_printf("FAT: SDE name:    '%.8s'\n", dentry->name);


    diag_printf("FAT: SDE ext:     '%.3s'\n", dentry->ext);
    diag_printf("FAT: SDE attr:     %c%c%c%c%c%c\n",
                (DENTRY_IS_RDONLY(dentry)  ? 'R' : '-'),
                (DENTRY_IS_HIDDEN(dentry)  ? 'H' : '-'),
                (DENTRY_IS_SYSTEM(dentry)  ? 'S' : '-'),
                (DENTRY_IS_VOLUME(dentry)  ? 'V' : '-'),
                (DENTRY_IS_DIR(dentry)     ? 'D' : '-'),
                (DENTRY_IS_ARCHIVE(dentry) ? 'A' : '-'));

    diag_printf("FAT: SDE crt time: %u\n", dentry->crt_time);
    diag_printf("FAT: SDE crt date: %u\n", dentry->crt_date);
    diag_printf("FAT: SDE acc date: %u\n", dentry->acc_date);
    diag_printf("FAT: SDE wrt time: %u\n", dentry->wrt_time);
    diag_printf("FAT: SDE wrt date: %u\n", dentry->wrt_date);
    diag_printf("FAT: SDE cluster:  %u\n", dentry->cluster);
    diag_printf("FAT: SDE size:     %u\n", dentry->size);

}
#endif // TDE


/** Returns an unsigned byte checksum computed on an unsigned byte
 *  array. The array must be 11 bytes long and is assumed to contain
 *  a name stored in the format of a MS-DOS directory entry.
 *
 * @param p_data   Pointer to an unsigned byte array assumed to be 11 bytes long.
 * @return result  An 8-bit unsigned checksum of the array pointed to by pFcbName.
 */
cyg_uint8 ChkSum(char *p_data)
{
       cyg_uint8 len;
       cyg_uint8 result;
       result = 0;
       for (len=11; len!=0; len--) {
               // NOTE: The operation is an unsigned char rotate right
               result = ((result & 1) ? 0x80 : 0) + (result >> 1) + *p_data++;
       }
       return (result);
}

#ifdef CYGCFG_FS_FAT_USE_VFAT
/** Reads dir entry from disk even if they are between 2 clusters(can't be more than 2).
 *
 * @param disk   A pointer to the disk
 * @param pos    A pointer to the structure describing the current position
 * @param dentry A pointer to buffer where datas will be stored
 * @param pt_len Contains the lentgh ofthe buffer
 * @return ENOERR upon success, or an error code
 */
static int copy_raw_long_dentry( fatfs_disk_t      *disk,
                                 fatfs_data_pos_t  *pos,
                                 char              *pt_data,
                                 cyg_uint32        *pt_len,
                                 fat_raw_dir_entry_t *dentry  )
{
   cyg_uint16  next_pos;
   cyg_uint32 length_to_read;  //can't be more than 21*0x20=0x2A0 but 32bits is due to disk_cluster_read
   cyg_uint16 offset;
   fatfs_data_pos_t  pos_tmp;

   int err = ENOERR;

   pos_tmp.cluster      = pos->cluster;
   pos_tmp.cluster_snum = pos->cluster_snum;
   pos_tmp.cluster_pos  = pos->cluster_pos;

   next_pos = pos_tmp.cluster_pos + *pt_len;

    if (pos->cluster==0)
    {// we are reading the FAT12/16 root directory
        if(next_pos <= disk->fat_root_dir_size)
        {
            err = disk_read(disk, (void*)pt_data, pt_len, disk->fat_root_dir_pos + pos_tmp.cluster_pos);
        }
        else
        {
            err = EINVAL;
            return err;
        }
    }
    else //(pos->cluster!=0)
    {// we are not reading the FAT12/16 root directory
        if(next_pos <= disk->cluster_size)
        {
            err = disk_cluster_read(disk, (void*)pt_data, pt_len, pos_tmp.cluster, pos_tmp.cluster_pos);
        }
        else
        {// we are not reading the FAT12/16 root directory
            // the entries describing this file are located in several clusters
            length_to_read = disk->cluster_size - pos_tmp.cluster_pos;

            err = disk_cluster_read(disk, (void*)pt_data,  &length_to_read, pos_tmp.cluster, pos_tmp.cluster_pos);
            // Going to the next cluster of the chain
            err = get_next_cluster(disk, &pos_tmp, CO_NONE);
            if (err != ENOERR)
                return err;

            next_pos -= disk->cluster_size;
            offset    = length_to_read;
            length_to_read = next_pos;
            err = disk_cluster_read(disk, (void*) &(pt_data[offset]), &length_to_read, pos_tmp.cluster, pos_tmp.cluster_pos);
        }
    }

   return err;
}



//! UTF16 to UTF8 translation buffer
cyg_uint16 rd_utf16[LNAME_MAX_CHAR];
//! UTF8 to UTF16 translation buffer
cyg_uint16 wr_utf16[LNAME_MAX_CHAR];

/** Translate an UTF16 string into an UTF8 string.
 *
 *  An UTF8 char can be encodded on 4 bytes with 21 significant bits
 *  An UTF16 char can be encodded on 2 words with 20 significant bits
 *
 * @param tab_utf16 A pointer to the utf16 string
 * @param tab_utf8 A pointer to the utf8 string
 * @param nb_bytes_utf16 number of utf16 characters
 * @return the number of utf8 bytes or 0 upon error
 */
cyg_uint32 xlate_utf16_to_utf8( cyg_uint16 *tab_utf16,
                                char       *tab_utf8,
                                cyg_uint32 nb_bytes_utf16 )
{
    const cyg_uint16   *ip;
    cyg_uint32  nb_bytes_utf8 = 0;

    ip   = tab_utf16;

    while (*ip && nb_bytes_utf16 > 0)
    {
        if( *ip <= 0x7F )
        {
            tab_utf8[nb_bytes_utf8] = (cyg_uint8)(*ip);
            nb_bytes_utf8 += 1;
				/* check if there'is enough space in utf8 buf to write this char
				 *	and the '\0' char*/
				if(nb_bytes_utf8 > LNAME_MAX_CHAR-1){
            	nb_bytes_utf8 -= 1;
					goto out;
				}
            ip = &(ip[1]);
        }
        else if( *ip <= 0x7FF )
        {   // Then we need 2 bytes
            tab_utf8[nb_bytes_utf8]   = (cyg_uint8) ( 0xC0 | ((*ip >>  6) & 0x1F) );
            tab_utf8[nb_bytes_utf8+1] = (cyg_uint8) ( 0x80 |  (*ip        & 0x3F) );
            nb_bytes_utf8 += 2;
				/* check if there'is enough space in utf8 buf to write this char
				 *	and the '\0' char*/
				if(nb_bytes_utf8 > LNAME_MAX_CHAR-1){
            	nb_bytes_utf8 -= 2;
					goto out;
				}
            ip = &(ip[1]);
        }
        else if( (*ip <= 0xD7FF) || (*ip >= 0xE000) )
        {   // Then we need 3 bytes
            tab_utf8[nb_bytes_utf8]   = (cyg_uint8) (0xE0 | ((*ip >> 12) & 0x0F));
            tab_utf8[nb_bytes_utf8+1] = (cyg_uint8) (0x80 | ((*ip >>  6) & 0x3F));
            tab_utf8[nb_bytes_utf8+2] = (cyg_uint8) (0x80 |  (*ip        & 0x3F));
            nb_bytes_utf8 += 3;
				/* check if there'is enough space in utf8 buf to write this char
				 *	and the '\0' char*/
				if(nb_bytes_utf8 > LNAME_MAX_CHAR-1){
            	nb_bytes_utf8 -= 3;
					goto out;
				}
            ip = &(ip[1]);
        }
        else if(  ((ip[0] & UTF16_MASK_SURR_HIG ) == UTF16_MASK_SURR_HIG )
                &&((ip[1] & UTF16_MASK_SURR_LOW ) == UTF16_MASK_SURR_LOW ) )
        {   // Then we need 3 bytes
            cyg_uint32 ucs4 = (ip[1]&0x3FF) | ((ip[0]&0x3FF)<<10);
            tab_utf8[nb_bytes_utf8]   = (cyg_uint8) (0xF0 | ((ucs4 >> 18) & 0x07));
            tab_utf8[nb_bytes_utf8+1] = (cyg_uint8) (0x80 | ((ucs4 >> 12) & 0x3F));
            tab_utf8[nb_bytes_utf8+2] = (cyg_uint8) (0x80 | ((ucs4 >>  6) & 0x3F));
            tab_utf8[nb_bytes_utf8+3] = (cyg_uint8) (0x80 |  (ucs4        & 0x3F));
            nb_bytes_utf8 += 4;
				/* check if there'is enough space in utf8 buf to write this char
				 *	and the '\0' char*/
				if(nb_bytes_utf8 > LNAME_MAX_CHAR-1){
            	nb_bytes_utf8 -= 4;
					goto out;
				}
            nb_bytes_utf16 --; /*this char needed 2 utf16 chars*/
            ip = &(ip[2]);
        }
        nb_bytes_utf16 --;
    }

out:
    tab_utf8[nb_bytes_utf8] = 0;

    return nb_bytes_utf8;
}

#endif //CYGCFG_FS_FAT_USE_VFAT

/** Reads dir entry from disk.
 *
 * @param disk A pointer to the disk
 * @param pos ---
 * @param dentry A pointer to long dir name structure
 * @return ENOERR upon success, or an error code
 */
int
read_raw_dentry( fatfs_disk_t        *disk,
                 fatfs_data_pos_t    *pos,
                 fat_raw_dir_entry_t *dentry)
{
    cyg_uint32 len;
    int err;

    //-----------------------------------------------------
    // At first, we're gonna read the 32 bytes to determine
    // wether it's a long entry or not
    //-----------------------------------------------------
    len = DENTRY_SIZE;

   // Check if we are reading the FAT12/16 root directory
   if (0 == pos->cluster)
      err = disk_read(disk, (void*)disk->rd_data, &len,
                      disk->fat_root_dir_pos + pos->cluster_pos);
   else
     err = disk_cluster_read(disk, (void*)disk->rd_data, &len,
                             pos->cluster, pos->cluster_pos);


    if (err != ENOERR)
        return err;

#ifdef CYGCFG_FS_FAT_USE_VFAT
   // Get the attribute in order to know if it's a short or long dentry
   GET_BYTE(disk->rd_data,  dentry->lndattr, 0x0B);

   if(   ( DENTRY_IS_LNDIR(dentry) )// we deals with a long name entry
      && (disk->rd_data[0]!= 0xE5)           // entry is deleted
      && (disk->rd_data[0]!= 0x00))          // entry is zeroed
   {
      //---------------------
      // Reads the long entry
      //---------------------
      cyg_uint8  i_le;           // index of long entries
      cyg_uint8  ldir_chksum;    // short entry chksum
      cyg_uint8  ldir_chksum_tmp;// short entry chksum to compare
      cyg_uint8  ldir_type;      // used to verify LDIR_type
      char       *pt_entryblock; // pointer to a single long entry
      cyg_uint16 name_offset;
      cyg_uint32 nb_bytes_utf8;

      // Read the number of long entries
      GET_BYTE(disk->rd_data, dentry->nb_entries, 0x00 );
      dentry->nb_entries = NBR_OF_LONG_ENTRIES(dentry->nb_entries) + 1; // +1 to include the short entry

     if(dentry->nb_entries > NUMBER_OF_ENTRY_MAX)
        return EINVAL;
      //-----------------------------------------------------
      // Now, we know it's a long dir entry, we need to use
      // a larger memory buffer depending on the nbr_of_le
      //-----------------------------------------------------
      len = (dentry->nb_entries)*DENTRY_SIZE;
      memset(disk->rd_data, 0, len);
      err = copy_raw_long_dentry( disk, pos, disk->rd_data, &len, dentry );

      if (err != ENOERR)
         return err;
      // computation of the short entry checksum
      ldir_chksum = ChkSum( &(disk->rd_data[(dentry->nb_entries-1)*DENTRY_SIZE ]) );

      memset(dentry->lname, 0, LNAME_MAX_CHAR);

      name_offset = 0;

      for(i_le = 1; i_le <= (dentry->nb_entries-1); i_le++)
      {
         // avoiding useless computation
         pt_entryblock = &(disk->rd_data[ ((dentry->nb_entries-1)-i_le)*DENTRY_SIZE ]);
         name_offset   = (i_le-1)*CHAR_PER_LE;

         GET_WORD(pt_entryblock, rd_utf16[0+name_offset],  0x01 );
         GET_WORD(pt_entryblock, rd_utf16[1+name_offset],  0x03 );
         GET_WORD(pt_entryblock, rd_utf16[2+name_offset],  0x05 );
         GET_WORD(pt_entryblock, rd_utf16[3+name_offset],  0x07 );
         GET_WORD(pt_entryblock, rd_utf16[4+name_offset],  0x09 );
         GET_WORD(pt_entryblock, rd_utf16[5+name_offset],  0x0E );
         GET_WORD(pt_entryblock, rd_utf16[6+name_offset],  0x10 );
         GET_WORD(pt_entryblock, rd_utf16[7+name_offset],  0x12 );
         GET_WORD(pt_entryblock, rd_utf16[8+name_offset],  0x14 );
         GET_WORD(pt_entryblock, rd_utf16[9+name_offset],  0x16 );
         GET_WORD(pt_entryblock, rd_utf16[10+name_offset], 0x18 );
         GET_WORD(pt_entryblock, rd_utf16[11+name_offset], 0x1C );
         GET_WORD(pt_entryblock, rd_utf16[12+name_offset], 0x1E );

         // verify checksum short entry on each long entry
         GET_BYTE(pt_entryblock, ldir_chksum_tmp,  0x0D );

         if( ldir_chksum != ldir_chksum_tmp )
         {
            FAT_WARNING("Error : checksum is wrong \n");
         }

         // verifies LDIR_type == 0x0000
         GET_WORD(pt_entryblock, ldir_type,  0x1A );

         if( ldir_type != 0x0 )
         {
            FAT_WARNING("Error : LDIR_type is wrong \n");
         }

      } // end for i_le loop

      nb_bytes_utf8 = xlate_utf16_to_utf8(rd_utf16, dentry->lname, (name_offset+CHAR_PER_LE) );
      if(nb_bytes_utf8 == 0 )
        return EINVAL;

      // Read the short entry
      pt_entryblock = &(disk->rd_data[ (dentry->nb_entries-1)*DENTRY_SIZE ]);
      GET_BYTES(pt_entryblock, dentry->name,     8, 0x00);
      GET_BYTES(pt_entryblock, dentry->ext,      3, 0x08);
      GET_BYTE(pt_entryblock,  dentry->attr,        0x0B);
      GET_BYTE(pt_entryblock,  dentry->nt_reserved, 0x0C);
      GET_BYTE(pt_entryblock,  dentry->crt_sec_100, 0x0D);
      GET_WORD(pt_entryblock,  dentry->crt_time,    0x0E);
      GET_WORD(pt_entryblock,  dentry->crt_date,    0x10);
      GET_WORD(pt_entryblock,  dentry->acc_date,    0x12);
      GET_WORD(pt_entryblock,  dentry->cluster_HI,  0x14);
      GET_WORD(pt_entryblock,  dentry->wrt_time,    0x16);
      GET_WORD(pt_entryblock,  dentry->wrt_date,    0x18);
      GET_WORD(pt_entryblock,  dentry->cluster,     0x1A);
      GET_DWORD(pt_entryblock, dentry->size,        0x1C);

   } // end if Long Dir Entry

   else
#endif //CYGCFG_FS_FAT_USE_VFAT
   {
      //----------------------
      // Reads the short entry
      //----------------------
#ifdef CYGCFG_FS_FAT_USE_VFAT
      dentry->lndattr = 0;
#endif //CYGCFG_FS_FAT_USE_VFAT
      dentry->nb_entries = 1; // a short entry is only 32 bytes long

      // Read the short entry
      GET_BYTES(disk->rd_data, dentry->name,     8, 0x00);
      GET_BYTES(disk->rd_data, dentry->ext,      3, 0x08);
      GET_BYTE(disk->rd_data,  dentry->attr,        0x0B);
      GET_BYTE(disk->rd_data,  dentry->nt_reserved, 0x0C);
      GET_BYTE(disk->rd_data,  dentry->crt_sec_100, 0x0D);
      GET_WORD(disk->rd_data,  dentry->crt_time,    0x0E);
      GET_WORD(disk->rd_data,  dentry->crt_date,    0x10);
      GET_WORD(disk->rd_data,  dentry->acc_date,    0x12);
      GET_WORD(disk->rd_data,  dentry->cluster_HI,  0x14);
      GET_WORD(disk->rd_data,  dentry->wrt_time,    0x16);
      GET_WORD(disk->rd_data,  dentry->wrt_date,    0x18);
      GET_WORD(disk->rd_data,  dentry->cluster,     0x1A);
      GET_DWORD(disk->rd_data, dentry->size,        0x1C);

    }

    if (disk->fat_type == FATFS_FAT16 || disk->fat_type == FATFS_FAT12)
        dentry->cluster_HI = 0;

    // Zero terminate strings
    dentry->name[8] = '\0';
    dentry->ext[3]  = '\0';

    if( DENTRY_IS_DELETED(dentry) )
    {
    FAT_DEBUG("entry info cl=0x%04x-%04x, name=DELETED, size=%d\tpos(0x%08x,0x%08x,0x%08x)\n",
                dentry->cluster_HI, dentry->cluster, 
                dentry->size,
                pos->cluster, pos->cluster_snum, pos->cluster_pos);
    }else if(DENTRY_IS_ZERO(dentry)){
    FAT_DEBUG("entry info cl=0x%04x-%04x, name=ZERO, size=%d\tpos(0x%08x,0x%08x,0x%08x)\n",
                dentry->cluster_HI, dentry->cluster, 
                dentry->size,
                pos->cluster, pos->cluster_snum, pos->cluster_pos);
    }else{
    FAT_DEBUG("entry info cl=0x%04x-%04x, name=%s,\text=%s size=%d\tpos(0x%08x,0x%08x,0x%08x)\n",
                dentry->cluster_HI, dentry->cluster, 
                dentry->name, dentry->ext, dentry->size,
                pos->cluster, pos->cluster_snum, pos->cluster_pos);
    }


#if TDE
    print_raw_dentry(dentry);
#endif

    return ENOERR;
}

// -------------------------------------------------------------------------
// get_raw_dentry_filename()
// Gets the filename from given dir entry.

static void
get_raw_dentry_filename(fat_raw_dir_entry_t *dentry, char *name)
{

    int   i     = 0;
    char *cname = name;

#ifdef CYGCFG_FS_FAT_USE_VFAT
   if( DENTRY_IS_LNDIR(dentry) )
   { // it's a VFAT entry
      char *cptr  = dentry->lname;

      while (*cptr != '\0' && i < 256)
      {
        *cname++ = *cptr++; i++;
      }
      *cname = '\0';

      FAT_DEBUG("vfat name='%s'\n", name);
   }
   else
#endif //CYGCFG_FS_FAT_USE_VFAT
   { // it's a short entry
    char *cptr  = dentry->name;

    while (*cptr != ' ' && i < 8)
    {
#ifdef CYGCFG_FS_FAT_USE_VFAT
		//  in VFAT Windows is only using a single short entry if
		// strlen(name) <= 11 and if letters are all between low or high
		// in this case, the nt reserved flag is used to differentiate
		// lower cases from upper ones
		if (dentry->nt_reserved & SNAME_LCASE)
		{
			*cname++ = tolower(*cptr++); i++;
		}
		else
		{
			*cname++ = *cptr++; i++;
		}
#else
        *cname++ = *cptr++; i++;
#endif //CYGCFG_FS_FAT_USE_VFAT
    }
    cptr = dentry->ext;

    if (*cptr != ' ')
    {
        *cname++ = '.'; i = 0;
        while (*cptr != ' ' && i < 3)
        {
#ifdef CYGCFG_FS_FAT_USE_VFAT
			//  in VFAT Windows is only using a single short entry if
			// strlen(name) <= 11 and if letters are all between low or high
			// in this case, the nt reserved flag is used to differentiate
			// lower cases from upper ones
			if (dentry->nt_reserved & SEXT_LCASE)
			{
				*cname++ = tolower(*cptr++); i++;
			}
			else
			{
				*cname++ = *cptr++; i++;
			}
#else
            *cname++ = *cptr++; i++;
#endif //CYGCFG_FS_FAT_USE_VFAT
        }
    }
    *cname = '\0';

    FAT_DEBUG("dos name='%s' dos ext='%s' filename='%s'",
                    dentry->name, dentry->ext, name);
   }
}


#ifdef CYGCFG_FS_FAT_USE_VFAT
/** Permits to jump to the next entry.
 *
 * @param disk   A pointer to the disk
 * @param pos    A pointer to the structure describing the current position and returning the next one
 * @param dentry A pointer to the current raw entry
 * @param opts   Option of the next cluster to get
 * @return number of free consecutive entries from pos
 */
int go_to_next_raw_entry( fatfs_disk_t        *disk,
                          fatfs_data_pos_t    *pos,
                          fat_raw_dir_entry_t *dentry,
                          cluster_opts_t      opts)
{
    int err = ENOERR;

    if( (pos->cluster != 0)&&((pos->cluster_pos + dentry->nb_entries*DENTRY_SIZE) > disk->cluster_size))
    {
        int cluster_offset;
        cluster_offset = (pos->cluster_pos + dentry->nb_entries*DENTRY_SIZE) - disk->cluster_size;
        err = get_next_cluster(disk, pos, opts);
        if(err != ENOERR)
            return err;
        pos->cluster_pos =  cluster_offset;
    }
    else if( (pos->cluster == 0)&&((pos->cluster_pos + dentry->nb_entries*DENTRY_SIZE) > disk->fat_root_dir_size))
    {
        int cluster_offset;
        cluster_offset = (pos->cluster_pos + dentry->nb_entries*DENTRY_SIZE) - disk->fat_root_dir_size;
        err = get_next_cluster(disk, pos, opts);
        if(err != ENOERR)
            return err;
        pos->cluster_pos =  cluster_offset;
    }
    else
        pos->cluster_pos =  (pos->cluster_pos + dentry->nb_entries*DENTRY_SIZE);

    return err;
}
#endif

// -------------------------------------------------------------------------
// read_next_raw_dentry()
// Gets next dir entry searching from given position to the end.
// If EEOF is returned there are no more entries in given dir.

int
read_next_raw_dentry(fatfs_disk_t        *disk,
                     fatfs_data_pos_t    *pos,
                     fat_raw_dir_entry_t *dentry)
{
    int err = ENOERR;

    // If we are reading the root dir on FAT32 we have
    // to correct the position to the root dir cluster
    if (FATFS_FAT32 == disk->fat_type && 0 == pos->cluster)
        pos->cluster = disk->fat_root_dir_cluster;

    while(true){
        // FAT12/16 root dir check
        if (0 == pos->cluster)
        {
            if (pos->cluster_pos >= disk->fat_root_dir_size)
                err = EEOF;
        }

        else
        {
            // Change cluster if needed
            if (pos->cluster_pos >= disk->cluster_size)
                err = get_next_cluster(disk, pos, CO_NONE);
        }

        if (err != ENOERR)
            break;

        err = read_raw_dentry(disk, pos, dentry);
        if (err == EEOF)
        {   // in this case, it means that the entry is located on several clusters
            // but this cluster is the last of the chain -> it's not a valid entry
            // it could for exemple happen when a disk is formatted
            break;
        }


        if (err != ENOERR)
            return err;

        if (DENTRY_IS_ZERO(dentry))
        {
            // If we get a ZERO dir entry, we assume that
            // there are no more entries in current dir
            CYG_TRACE0(TDE, "end of dir");
            err = EEOF;
            break;
        }
        else if (!DENTRY_IS_DELETED(dentry) && !DENTRY_IS_VOLUME(dentry))
        {
            // Dir entry found
            CYG_TRACE3(TDE, "found new dentry at cluster=%d snum=%d pos=%d",
                            pos->cluster, pos->cluster_snum, pos->cluster_pos);
            break;
        }
#ifdef CYGCFG_FS_FAT_USE_VFAT
        if ( DENTRY_IS_DELETED(dentry) )
            pos->cluster_pos += DENTRY_SIZE;
        else
         {
            err = go_to_next_raw_entry(disk, pos, dentry, CO_NONE);
            if (err != ENOERR)
               return err;
         }
#else
         pos->cluster_pos += DENTRY_SIZE;
#endif
    }

    // EEOF could be returned if there are no more entries in this
    // dir - this should be cought by the calling function
    return err;
}


/** Browse root dir to find the volume id.
 *
 * @param disk   A pointer to the disk
 * @param buf    A pointer to the volume id name buffer

 * @return EEOF if no volume id is set
 */
int fatfs_get_id(   fatfs_disk_t   *disk,
                     char           *buf)
{
    fat_raw_dir_entry_t raw_cdentry;
    fatfs_data_pos_t    pos;
    int                 err = ENOERR;

    err = fatfs_find_label(disk, &raw_cdentry, &pos);

    if(err == ENOERR)
    {
#ifdef CYGCFG_FS_FAT_USE_VFAT
        if ( DENTRY_IS_LNDIR(&raw_cdentry) ){
            sprintf( buf, "%s",raw_cdentry.lname);
		    FAT_INFO("Volume label found : %s\n", buf);
        }
        else
#endif
        {
            int i_char, len;
            sprintf( buf, "%s%s",raw_cdentry.name,raw_cdentry.ext);
            len = strlen(buf);
            for(i_char=0;i_char<len;i_char++){
                if(buf[i_char] > 0x7F) buf[i_char] = 0x2A; // 0x2A = '*'
            }
            FAT_INFO("Volume label found : %s\n", buf);
        }
    }
    return err;
}

/** Count the number of free consecutive entries from a position.
 *
 * @param disk   A pointer to the disk
 * @param pos    A pointer to the structure describing the current position
 * @return number of free consecutive entries from pos
 */
int nb_consecutive_free_entries( fatfs_disk_t     *disk,
                                 fatfs_data_pos_t *pos)
{

   char         data[DENTRY_SIZE];
   int          err                = ENOERR;
   int          nb_free_entries    = 0;
   cyg_uint32   entry_len          = DENTRY_SIZE;

   do{
      // Check if we are reading the FAT12/16 root directory
      if (0 == pos->cluster)
         err = disk_read(disk, (void*)data, &entry_len, disk->fat_root_dir_pos + pos->cluster_pos);
      else
         err = disk_cluster_read(disk, (void*)data,  &entry_len, pos->cluster, pos->cluster_pos);

      if( data[0] == 0xE5 )
      {
         nb_free_entries++;
         pos->cluster_pos += DENTRY_SIZE;
         // Change cluster if needed
         if ((pos->cluster != 0)&&(pos->cluster_pos >= disk->cluster_size))
         {
            err = get_next_cluster(disk, pos, CO_NONE);
         }
         else if((pos->cluster == 0)&&(pos->cluster_pos >= disk->fat_root_dir_pos))
         {  //reached the end of the root dir
            return nb_free_entries;
         }
      }
      else if( data[0] == 0x00 )
      {
         // this free entry is also the last of this dir
         // any long entry can replace it
         nb_free_entries =  NUMBER_OF_ENTRY_MAX;
         break;
      }

   }while( (data[0]==0xE5) && (err==ENOERR) );

   return nb_free_entries;
}

// -------------------------------------------------------------------------
// raw_to_dentry()
// Converts raw FAT dir entry to dir entry.

void raw_to_dentry(fat_raw_dir_entry_t *raw_dentry,
                   fatfs_data_pos_t    *raw_dentry_pos,
                   fatfs_dir_entry_t   *dentry)
{
    get_raw_dentry_filename(raw_dentry, dentry->filename);

    if (DENTRY_IS_DIR(raw_dentry))
        dentry->mode = __stat_mode_DIR;
    else
        dentry->mode = __stat_mode_REG;

#ifdef CYGCFG_FS_FAT_USE_ATTRIBUTES
    dentry->attrib = raw_dentry->attr;
#endif // CYGCFG_FS_FAT_USE_ATTRIBUTES

    date_dos2unix(raw_dentry->crt_time, raw_dentry->crt_date, &dentry->ctime);
    date_dos2unix(0,                    raw_dentry->acc_date, &dentry->atime);
    date_dos2unix(raw_dentry->wrt_time, raw_dentry->wrt_date, &dentry->mtime);

    dentry->size       = raw_dentry->size;
    dentry->priv_data  = raw_dentry->nt_reserved;
    dentry->cluster    = raw_dentry->cluster | (raw_dentry->cluster_HI << 16);
    dentry->disk_pos   = *raw_dentry_pos;
}

// -------------------------------------------------------------------------
// read_data()
// Reads data from given position.

static int
read_data(fatfs_disk_t     *disk,
          void             *data,
          cyg_uint32       *len,
          fatfs_data_pos_t *pos)
{
    cyg_uint8   *buf   = (cyg_uint8 *) data;
    cyg_uint32   size  = *len;
    int          err   = ENOERR;

    FAT_DEBUG("len=%d cluster=%d snum=%d pos=%d\n",
               *len, pos->cluster, pos->cluster_snum,
               pos->cluster_pos);

    while (size > 0)
    {
        cyg_uint32 csize;

        // Check if we are still inside current cluster
        if ((pos->cluster!=0)&&(pos->cluster_pos >= disk->cluster_size     ))
        {
            // Get next cluster of file, if at the last
            // cluster try to extend the cluster chain
            err = get_next_cluster(disk, pos, CO_EXTEND);
            if (err != ENOERR)
                goto out;
        }
        else if ((pos->cluster==0)&&(pos->cluster_pos >= disk->fat_root_dir_size))
        {
            err = ENOSPC;
            goto out;
        }

        // Adjust the data chunk size to be read to the cluster boundary
        if (size > (disk->cluster_size - pos->cluster_pos))
            csize = disk->cluster_size - pos->cluster_pos;
        else
            csize = size;

        FAT_DEBUG("-- len=%d cluster=%d snum=%d pos=%d\n",
                        csize, pos->cluster, pos->cluster_snum,
                        pos->cluster_pos);

        err = disk_cluster_read(disk, (void*)buf, &csize,
                                pos->cluster, pos->cluster_pos);
        if (err != ENOERR)
            goto out;

        // Adjust running variables
        buf              += csize;
        pos->cluster_pos += csize;
        size             -= csize;
    }

out:
    *len -= size;

    CYG_TRACE1(TDO, "total len=%d", *len);

    return err;
}


cyg_bool fatfs_read_ClnShutBit(fatfs_disk_t *disk)
{
    int     err = EINVAL;
    int     val = 0;
    char    *bufsector;
    cyg_uint32 n_sector, lensec;

    //allocate a buffer to work on a whole sector
    bufsector = (char*)malloc(disk->sector_size);
    if(bufsector == 0)
    {
        FAT_ERROR("can't allocate buffer\n");
        return EINVAL;
    }  

    /*-----------------------------------------------------------------------*/
    /* Reading FAT0 CRC                                                      */
    /*-----------------------------------------------------------------------*/
    n_sector = disk->fat_tbl_pos / disk->sector_size;
    lensec   = 1;
    err = cyg_io_bread(disk->dev_h, (void*)bufsector, &lensec, n_sector);
    if (err != ENOERR)
    {// reading in this sector failed
        FAT_ERROR("cyg_io_bread ERROR sector= %d\n", n_sector);
    }
    else
    {
        if(FATFS_FAT32 == disk->fat_type){        
            GET_DWORD(bufsector, val, 4);
            val = val >> 31;
        }else{
            GET_WORD(bufsector, val, 2);
            val = val >> 15;
        }
    }

    return (cyg_bool)val;
}



// -------------------------------------------------------------------------
// fatfs_init()
// Gets disk info.

int
fatfs_init(fatfs_disk_t *disk)
{
    cyg_uint32        sec_num, sec_per_fat, root_dir_sec_num;
    cyg_uint32        data_sec_num, data_clu_num;
    fat_boot_record_t boot_rec;
    int               err;

    CYG_CHECK_DATA_PTRC(disk);

    err = read_boot_record(disk, &boot_rec);
    if (err != ENOERR)
        return err;

    // Check some known boot record values
    if (0x29 != boot_rec.ext_sig       ||
        0x55 != boot_rec.exe_marker[0] ||
        0xAA != boot_rec.exe_marker[1])
        return EINVAL;

    // Sector and cluster sizes
    disk->sector_size       = boot_rec.bytes_per_sec;
    disk->sector_size_log2  = get_val_log2(disk->sector_size);
    disk->cluster_size      = boot_rec.bytes_per_sec * boot_rec.sec_per_clu;
    disk->cluster_size_log2 = get_val_log2(disk->cluster_size);
    disk->vol_id            = boot_rec.vol_id;

    // Sector and cluster size should always be a power of 2
    if (0 == disk->sector_size_log2 || 0 == disk->cluster_size_log2)
        return EINVAL;

    // Determine number of sectors
    if (boot_rec.sec_num_32 != 0)
        sec_num = boot_rec.sec_num_32;
    else
        sec_num = boot_rec.sec_num;

    // Determine number of sectors per fat
    if (boot_rec.sec_per_fat != 0)
        sec_per_fat = boot_rec.sec_per_fat;
    else
        sec_per_fat = boot_rec.sec_per_fat_32;

    // Number of sectors used by root directory
    root_dir_sec_num = ((boot_rec.max_root_dents * DENTRY_SIZE) +
                        (boot_rec.bytes_per_sec - 1)) / boot_rec.bytes_per_sec;

    // Number of data sectors
    data_sec_num = sec_num - (boot_rec.res_sec_num +
        (boot_rec.fat_tbls_num * sec_per_fat) + root_dir_sec_num);

    // Number of data clusters
    data_clu_num = data_sec_num / boot_rec.sec_per_clu;

    // FAT table size and position
    disk->fat_tbl_pos   = boot_rec.bytes_per_sec * boot_rec.res_sec_num;
    disk->fat_tbl_size  = boot_rec.bytes_per_sec * sec_per_fat;
    disk->fat_tbl_nents = data_clu_num + 2;
    disk->fat_tbls_num  = boot_rec.fat_tbls_num;

#ifdef CYGCFG_FS_FAT_ROBUST
    if( disk->fat_tbls_num >= 2 )
    {   // if FAT table backup exists 
        disk->fat_tbl_pos_backup = disk->fat_tbl_pos + disk->fat_tbl_size;
    }
#endif

    // Determine the type of FAT based on number of data clusters
    if (data_clu_num < 4085)
        disk->fat_type  = FATFS_FAT12;
    else if (data_clu_num < 65525)
        disk->fat_type  = FATFS_FAT16;
    else
        disk->fat_type  = FATFS_FAT32;

    // Determine root dir and data positions
    if (FATFS_FAT32 != disk->fat_type)
    {
        disk->fat_root_dir_pos     = disk->fat_tbl_pos +
                                     disk->fat_tbls_num * disk->fat_tbl_size;
        disk->fat_root_dir_size    = boot_rec.max_root_dents * DENTRY_SIZE;
        disk->fat_root_dir_nents   = boot_rec.max_root_dents;
        disk->fat_root_dir_cluster = 0;
        disk->fat_data_pos         = disk->fat_root_dir_pos +
                                     disk->fat_root_dir_size;

    }
    else
    {
        disk->fat_root_dir_pos     = 0;
        disk->fat_root_dir_size    = 0;
        disk->fat_root_dir_nents   = 0;
        disk->fat_root_dir_cluster = boot_rec.root_cluster;
        disk->fat_data_pos         = disk->fat_tbl_pos +
                                     disk->fat_tbls_num * disk->fat_tbl_size;
    }

    // fsinfo is not considered as valid at first time
    disk->read_fsinfo = false;
    return ENOERR;
}

// -------------------------------------------------------------------------
// fatfs_get_root_dir_entry()
// Gets root dir entry.

void
fatfs_get_root_dir_entry(fatfs_disk_t *disk, fatfs_dir_entry_t *dentry)
{
    CYG_CHECK_DATA_PTRC(disk);
    CYG_CHECK_DATA_PTRC(dentry);

    dentry->mode           = __stat_mode_DIR;
#ifdef CYGCFG_FS_FAT_USE_ATTRIBUTES
    dentry->attrib         = S_FATFS_DIR;
#endif // CYGCFG_FS_FAT_USE_ATTRIBUTES
    dentry->size           = disk->fat_root_dir_size;
    dentry->ctime          = 0;
    dentry->atime          = 0;
    dentry->mtime          = 0;
    dentry->filename[0]    = '\0';
    dentry->cluster        = 0;
    dentry->parent_cluster = 0;

    dentry->disk_pos.cluster      = 0;
    dentry->disk_pos.cluster_snum = 0;
    dentry->disk_pos.cluster_pos  = 0;
}

// -------------------------------------------------------------------------
// fatfs_get_disk_usage()
// Gets disk space.

int
fatfs_get_disk_usage(fatfs_disk_t *disk,
                     cyg_uint32   *total_clusters,
                     cyg_uint32   *free_clusters)
{
    cyg_uint32 c, nfree, tentry;
    int        err;

    read_n_tentry_init(disk);

    nfree = 0;
    for (c = 2; c < disk->fat_tbl_nents; c++)
    {
        err = read_n_tentry(disk, c, &tentry);

        if (err != ENOERR)
            return err;

        if (TENTRY_FREE == get_tentry_type(disk, tentry))
            nfree++;
    }

    *total_clusters = disk->fat_tbl_nents - 2;
    *free_clusters  = nfree;

    return ENOERR;
}


// -------------------------------------------------------------------------
// fatfs_read_dir_entry()
// Reads dir entry at given position.
// If there is no dir entry at given position the next closest is returned
// and the position is updated. If EEOF error is returned, then there are
// no more dir entries in given dir.

int
fatfs_read_dir_entry(fatfs_disk_t      *disk,
                     fatfs_dir_entry_t *dir,
                     fatfs_data_pos_t  *pos,
                     fatfs_dir_entry_t *dentry)
{
    fat_raw_dir_entry_t raw_dentry;
    int                 err;

    CYG_CHECK_DATA_PTRC(disk);
    CYG_CHECK_DATA_PTRC(dir);
    CYG_CHECK_DATA_PTRC(pos);
    CYG_CHECK_DATA_PTRC(dentry);

    err = read_next_raw_dentry(disk, pos, &raw_dentry);
    if (err != ENOERR)
       return err;

    raw_to_dentry(&raw_dentry, pos, dentry);
    dentry->parent_cluster = dir->cluster;

    // Increment position for next call
#ifdef CYGCFG_FS_FAT_USE_VFAT
    if ( DENTRY_IS_DELETED(&raw_dentry) || DENTRY_IS_ZERO(&raw_dentry))
            pos->cluster_pos += DENTRY_SIZE;
    else
    {
      err = go_to_next_raw_entry(disk, pos, &raw_dentry, CO_NONE);
      if (err != ENOERR)
         return err;
    }
#else
   pos->cluster_pos += DENTRY_SIZE;
#endif


    return ENOERR;
}



// GE- patch fred
// -------------------------------------------------------------------------
// fatfs_read_next_dir_entry()
// Reads dir entry next to the given position.
// If there is no dir entry at given position the next closest is returned
// and the position is updated. If EEOF error is returned, then there are
// no more dir entries in given dir.

int
fatfs_read_next_dir_entry(fatfs_disk_t      *disk,
                     fatfs_dir_entry_t *dir,
                     fatfs_data_pos_t  *pos,
                     fatfs_dir_entry_t *dentry)
{
    fat_raw_dir_entry_t raw_dentry;
    int                 err;

    CYG_CHECK_DATA_PTRC(disk);
    CYG_CHECK_DATA_PTRC(dir);
    CYG_CHECK_DATA_PTRC(pos);
    CYG_CHECK_DATA_PTRC(dentry);

    err = read_next_raw_dentry(disk, pos, &raw_dentry);
    if (err != ENOERR)
       return err;

    // Increment position
#ifdef CYGCFG_FS_FAT_USE_VFAT
    err = go_to_next_raw_entry(disk, pos, &raw_dentry, CO_NONE);
    if (err != ENOERR)
      return err;
#else
   pos->cluster_pos += DENTRY_SIZE;
#endif

    err = read_next_raw_dentry(disk, pos, &raw_dentry);
    if (err != ENOERR)
       return err;

   if ( DENTRY_IS_DELETED(&raw_dentry) || DENTRY_IS_ZERO(&raw_dentry))
       return ENOENT;

    raw_to_dentry(&raw_dentry, pos, dentry);
    dentry->parent_cluster = dir->cluster;

    return ENOERR;
}


//GE- patch fred


// -------------------------------------------------------------------------
// fatfs_setpos()
// Sets the file position from offset.

int
fatfs_setpos(fatfs_disk_t      *disk,
             fatfs_dir_entry_t *file,
             fatfs_data_pos_t  *pos,
             cyg_uint32         offset)
{
    int err;

    CYG_CHECK_DATA_PTRC(disk);
    CYG_CHECK_DATA_PTRC(file);
    CYG_CHECK_DATA_PTRC(pos);

    err = get_position_from_off(disk, file->cluster, offset, pos, CO_NONE);

    if (EEOF == err && offset == file->size)
        return ENOERR;
    else
        return err;
}

// -------------------------------------------------------------------------
// fatfs_getpos()
// Gets the file offset from position.

cyg_uint32
fatfs_getpos(fatfs_disk_t      *disk,
             fatfs_dir_entry_t *file,
             fatfs_data_pos_t  *pos)
{
    CYG_CHECK_DATA_PTRC(disk);
    CYG_CHECK_DATA_PTRC(file);
    CYG_CHECK_DATA_PTRC(pos);

    return (pos->cluster_snum << disk->cluster_size_log2) + pos->cluster_pos;
}

// -------------------------------------------------------------------------
// fatfs_read_data()
// Reads data from disk.

int
fatfs_read_data(fatfs_disk_t      *disk,
                fatfs_dir_entry_t *file,
                fatfs_data_pos_t  *pos,
                void              *data,
                cyg_uint32        *len)
{
    CYG_CHECK_DATA_PTRC(disk);
    CYG_CHECK_DATA_PTRC(file);
    CYG_CHECK_DATA_PTRC(data);
    CYG_CHECK_DATA_PTRC(len);
    CYG_CHECK_DATA_PTRC(pos);

    return read_data(disk, data, len, pos);
}

// -------------------------------------------------------------------------
// Support routines
// These enable the definition of local versions of certain routines
// if the given packages are not present.

#ifndef CYGPKG_LIBC_I18N

__externC int
toupper( int c )
{
    return (('a' <= c) && (c <= 'z')) ? c - 'a' + 'A' : c ;
}

#ifdef CYGCFG_FS_FAT_USE_VFAT
__externC int
tolower( int c )
{
    return (('A' <= c) && (c <= 'Z')) ? c - 'A' + 'a' : c ;
}
#endif //CYGCFG_FS_FAT_USE_VFAT

#endif

#ifndef CYGFUN_LIBC_STRING_BSD_FUNCS

__externC int
strcasecmp( const char *s1, const char *s2 )
{
    int ret;
    CYG_REPORT_FUNCNAMETYPE( "strcasecmp", "returning %d" );
    CYG_REPORT_FUNCARG2( "s1=%08x, s2=%08x", s1, s2 );

    CYG_CHECK_DATA_PTR( s1, "s1 is not a valid pointer!" );
    CYG_CHECK_DATA_PTR( s2, "s2 is not a valid pointer!" );

    while (*s1 != '\0' && toupper(*s1) == toupper(*s2))
    {
        s1++;
        s2++;
    }

    ret = toupper(*(unsigned char *) s1) - toupper(*(unsigned char *) s2);
    CYG_REPORT_RETVAL( ret );
    return ret;
}

__externC int
strncasecmp( const char *s1, const char *s2, size_t n )
{
    int ret;
    CYG_REPORT_FUNCNAMETYPE( "strncasecmp", "returning %d" );
    CYG_REPORT_FUNCARG3( "s1=%08x, s2=%08x, n=%d", s1, s2, n );

    if (n == 0)
    {
        CYG_REPORT_RETVAL(0);
        return 0;
    }

    CYG_CHECK_DATA_PTR( s1, "s1 is not a valid pointer!" );
    CYG_CHECK_DATA_PTR( s2, "s2 is not a valid pointer!" );

    while (n-- != 0 && toupper(*s1) == toupper(*s2))
    {
        if (n == 0 || *s1 == '\0' || *s2 == '\0')
            break;
        s1++;
        s2++;
    }

    ret = toupper(*(unsigned char *) s1) - toupper(*(unsigned char *) s2);
    CYG_REPORT_RETVAL( ret );
    return ret;
}

#endif
