/**
 * ============================================================================
 * @file  dlmallocapi.cxx
 * @brief C API for user mode dlmalloc
 *
 * Memory protection memory allocation.
 *
 * Copyright (C) 2004,2005 Parrot S.A.
 *
 * @author     ivan.djelic@parrot.fr
 * @date       2004-07-02
 * ============================================================================
 */

#include <pkgconf/system.h>
#ifdef CYGPKG_KERNEL
# include <pkgconf/kernel.h>
#endif

#include <cyg/infra/cyg_type.h>        // types
#include <cyg/infra/cyg_ass.h>         // assertion macros
#include <cyg/infra/cyg_trac.h>        // tracing macros
#include <cyg/infra/diag.h>
#include <cyg/kernel/kapi.h>           // C API

#include "dlmallocuser.hxx"
#include "dlmallocuserd.h"

// MACROS

#ifdef CYGDBG_USE_ASSERTS

#define CYG_ASSERT_SIZES(cstruct, cxxstruct)                      \
CYG_MACRO_START                                                   \
    char *msg = "Size of C struct " #cstruct                      \
                       " != size of C++ struct " #cxxstruct ;     \
    CYG_ASSERT( sizeof(cstruct) == sizeof(cxxstruct) , msg );     \
CYG_MACRO_END

#else

#define CYG_ASSERT_SIZES(cstruct, cxxstruct)

#endif

//! Magic new function
inline void *operator new( size_t size, void *ptr ) {
  CYG_CHECK_DATA_PTR( ptr, "Bad pointer" );
  return ptr;
}

/** Create a variable size memory pool
 *
 * This function creates a dlmalloc memory pool.
 * @param base Base of memory to use for pool
 * @param size Size of memory in bytes
 * @param handle Handle of pool
 * @param dl Space to put pool in
 * @param sbrk_enabled If '1' allow heap growing/shrinking with sbrk().
 * @param domain ARM domain
*/
externC void
cyg_mempool_dlmalloc_create( void *base,
			     cyg_int32 size,
			     cyg_handle_t *handle,
			     cyg_mempool_dlmalloc *dl,
			     int sbrk_enabled,
			     int domain ) {
  CYG_ASSERT_SIZES( cyg_mempool_dlmalloc,
		    Cyg_Mempool_dlmallocuser_Implementation);
  Cyg_Mempool_dlmallocuser_Implementation *t = new((void *)dl)
    Cyg_Mempool_dlmallocuser_Implementation ((cyg_uint8 *)base, size,
					     (CYG_ADDRWORD)sbrk_enabled,
					     (CYG_ADDRWORD)domain);
  t=t;
  CYG_CHECK_DATA_PTR( handle, "Bad handle pointer" );
  *handle = (cyg_handle_t)dl;
}

//! Delete variable size memory pool
externC void
cyg_mempool_dlmalloc_delete( cyg_handle_t dlpool ) {
  ((Cyg_Mempool_dlmallocuser_Implementation *)dlpool)
    ->~Cyg_Mempool_dlmallocuser_Implementation();
}

/** Allocates a block of length size.
 *
 * @return NULL is returned if no memory is available.
 */
externC void *
cyg_mempool_dlmalloc_try_alloc( cyg_handle_t dlpool, cyg_int32 size ) {
  return ((Cyg_Mempool_dlmallocuser_Implementation *)dlpool)->try_alloc(size);
}

//! Frees memory back into variable size pool
externC void
cyg_mempool_dlmalloc_free( cyg_handle_t dlpool, void *p ) {
  cyg_bool b;
  b = ((Cyg_Mempool_dlmallocuser_Implementation *)dlpool)->free((cyg_uint8 *)p,
								0);
  CYG_ASSERT( b, "Bad free");
}

//! Resize allocated block
externC void *
cyg_mempool_dlmalloc_resize_alloc( cyg_handle_t dlpool,
				   cyg_uint8 *ptr, cyg_int32 size,
				   cyg_int32 *oldsize) {
  return ((Cyg_Mempool_dlmallocuser_Implementation *)dlpool)
    ->resize_alloc(ptr,size,oldsize);
}

