//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
 * ============================================================================
 * @file  memalloc.c
 * @brief User mode memory allocation.
 *
 * Memory protection implementation of malloc/calloc/free etc.
 *
 * Copyright (C) 2004,2005 Parrot S.A.
 *
 * @author     ivan.djelic@parrot.fr
 * @date       2004-06-26
 * ============================================================================
 */

#include <cyg/hal/hal_arch.h>
#include <cyg/hal/hal_cache.h>
#include <cyg/infra/diag.h>
#include <cyg/infra/cyg_ass.h>
#include <cyg/infra/cyg_trac.h>        // tracing macros
#include <cyg/hal/hal_mmu.h>
#include <cyg/kernel/kapi.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include "dlmallocuser.h"
#include "mmu.h"
#include "thread.h"
#include "vp_layout.h"
#include "diag.h"

extern void __swi_cyg_mutex_lock(cyg_mutex_t *lock);
extern void __swi_cyg_mutex_unlock(cyg_mutex_t *lock);
extern int __swi_cyg_memprot_get_current_domain(cyg_uint32 addr);
extern cygmem_vp_layout_t cygmem_vp_layouts[16];

//! user mutexes
static cyg_mutex_t mempool_mutex[16];

//cyg_mempool_dlmalloc cyg_memprot_heaps[15];
//! user mempools
static cyg_handle_t cyg_memprot_mempools[16];
static cyg_mutex_t sbrk_mutex;

#define PAGE_CEIL(_addr)  (((_addr)+MMU_PAGE_SIZE-1)&~(MMU_PAGE_SIZE-1))

#define heap_lock(_dom) \
 __swi_cyg_mutex_lock(&mempool_mutex[cyg_memprot_get_current_domain(_dom)]);
#define heap_unlock(_dom) \
 __swi_cyg_mutex_unlock(&mempool_mutex[cyg_memprot_get_current_domain(_dom)]);

#undef heap_lock
#define heap_lock(_dom)
#undef heap_unlock
#define heap_unlock(_dom)

/** Initialise memory pools.
 *
 * This function is called at start-up to create and initialise heaps and
 * memory pools.
 */
void cyg_memprot_create_mempools(void) {
  int i, ret, ap;
  cygmem_vp_layout_t *layout;
  cyg_mempool_dlmalloc *dlpool;

  // this is the kernel mempool - NOT USED
  //cyg_memprot_mempools[0] = cyg_mempool_dlmalloc_get_heap();

  // create a memory pool for each (non-empty) process
  for ( i = 1; i < 16; i++ ) {
    layout = &cygmem_vp_layouts[i];
    if ( layout->ro_start < layout->ro_end ) {
      // allocate and map a page at beginning of heap
      ap = AP_RW_RW;
      // use only 4 KB pages to allow partial deallocation
      ret = memprot_mmu_map( layout->heap_start, 0, MMU_PAGE_SIZE, ap, 1, i,
			 MMU_ALLOCATE, 1 );
      CYG_ASSERT( ret == ENOERR, "Unable to allocate heap mempool !\n" );
      dlpool = (cyg_mempool_dlmalloc *) layout->heap_start;
      layout->heap_start += sizeof(cyg_mempool_dlmalloc);
      layout->heap_top = layout->heap_start;
      CYG_ASSERT( layout->heap_top < layout->heap_end,
		  "Default heap size is too small !\n" );
      // create mempool
      cyg_mempool_dlmalloc_create((void *)layout->heap_start,
				  layout->heap_end-layout->heap_start,
				  &cyg_memprot_mempools[i], dlpool, 1, i );
      // init mutex
      cyg_mutex_init(&mempool_mutex[i]);
    }
  }
  // init sbrk mutex
  cyg_mutex_init(&sbrk_mutex);
}

/** Heap resizing.
 *
 * This function is called from DL malloc routines to automatically
 * extend/trim the heap area. It can result in memory pages being mapped
 * in/out.
 *
 * @param incr How much memory should be added on top of heap (can be negative)
 * @return Previous heap top address
 */
CYG_ADDRESS cyg_memprot_sbrk( ptrdiff_t incr, int domain ) {
  int ret, ap;
  cygmem_vp_layout_t *layout;
  CYG_ADDRESS new_heap_top, real_new_heap_top, old_heap_top, page;

  cyg_mutex_lock(&sbrk_mutex);

  // access permission
  ap = ( domain > 0 )? AP_RW_RW : AP_RW_RO;
  // get memory layout
  layout = &cygmem_vp_layouts[domain];
  new_heap_top = layout->heap_top + incr;
  real_new_heap_top = PAGE_CEIL(new_heap_top);
  if (( new_heap_top < layout->heap_start )||
      ( real_new_heap_top > layout->heap_end )) {
    DEBUG("cyg_memprot_sbrk() failed with incr=%d\n", incr );
    cyg_mutex_unlock(&sbrk_mutex);
    return (CYG_ADDRESS)-1;
  }
  if ( real_new_heap_top < PAGE_CEIL(layout->heap_top) ) {
    // free physical ram if required
    ret = memprot_mmu_unmap( real_new_heap_top,
			 PAGE_CEIL(layout->heap_top) - real_new_heap_top, 1 );
    CYG_ASSERT( ret == ENOERR, "Unable to unmap page !\n");
  }
  /*
  for ( page = real_new_heap_top;
	page < PAGE_CEIL(layout->heap_top);
	page += MMU_PAGE_SIZE ) {
    ret = memprot_mmu_unmap( page, MMU_PAGE_SIZE, 1 );
    CYG_ASSERT( ret == ENOERR, "Unable to unmap page !\n");
  }
  */
  if ( PAGE_CEIL(layout->heap_top) < real_new_heap_top ) {
    // allocate physical ram if required
    ret = memprot_mmu_map( PAGE_CEIL(layout->heap_top), 0,
		       real_new_heap_top - PAGE_CEIL(layout->heap_top),
		       ap, 1, domain, MMU_ALLOCATE, 0 );
    if ( ret == ENOMEM ) {
      INFO("out of ram in cyg_memprot_sbrk() with incr=%d (process %d)\n",
	   incr, domain );
      // return already allocated pages
      real_new_heap_top = page;
      cyg_mutex_unlock(&sbrk_mutex);
      return (CYG_ADDRESS)-1;
    }
    CYG_ASSERT(( ret != EINVAL ),
	       "memprot_mmu_map() failed with EINVAL in cyg_memprot_sbrk !\n");
  }
  /*
  for ( page = PAGE_CEIL(layout->heap_top);
	page < real_new_heap_top;
	page += MMU_PAGE_SIZE ) {
    ret = memprot_mmu_map( page, 0, MMU_PAGE_SIZE, ap, 1, domain, MMU_ALLOCATE,1 );
    if ( ret == ENOMEM ) {
      INFO("out of ram in cyg_memprot_sbrk() with incr=%d (process %d)\n",
	   incr, domain );
      // return already allocated pages
      real_new_heap_top = page;
      for ( page = PAGE_CEIL(layout->heap_top);
	    page < real_new_heap_top;
	    page += MMU_PAGE_SIZE )
	memprot_mmu_unmap( page, MMU_PAGE_SIZE, 1 );
      cyg_mutex_unlock(&sbrk_mutex);
      return (CYG_ADDRESS)-1;
    }
    CYG_ASSERT(( ret != EINVAL ),
	       "memprot_mmu_map() failed with EINVAL in cyg_memprot_sbrk !\n");
  }
  */
  old_heap_top = layout->heap_top;
  layout->heap_top = new_heap_top;
  cyg_mutex_unlock(&sbrk_mutex);
  DEBUG("cyg_memprot_sbrk(): heap_top[%d] changed from 0x%08x to 0x%08x\n",
	domain, old_heap_top, new_heap_top );
  return old_heap_top;
}

// the following code was taken and modified from malloc.cxx

//! User mode malloc
void *cyg_memprot_malloc( size_t size, cyg_uint32 _lr ) {
  void *data_ptr;
  cyg_handle_t pool;
  int domain = __swi_cyg_memprot_get_current_domain(_lr);

  if ( domain == 0 ) {
#if CYGINT_ISO_MALLOC
    return malloc(size);
#else
    INFO("no kernel memory allocation available !\n");
    return NULL;
#endif
  }

#ifdef CYGSEM_MEMALLOC_MALLOC_ZERO_RETURNS_NULL
  // first check if size wanted is 0
  if ( 0 == size ) {
    return NULL;
  } // if
#endif
  pool = cyg_memprot_mempools[domain];
  heap_lock(domain);
  data_ptr = cyg_mempool_dlmalloc_try_alloc( pool, size );
  heap_unlock(domain);
  return data_ptr;
}

//! User mode free
void cyg_memprot_free( void *ptr, cyg_uint32 _lr ) {
  cyg_handle_t pool;
  int domain = __swi_cyg_memprot_get_current_domain(_lr);

  if( domain == 0 ) {
#if CYGINT_ISO_MALLOC
    free(ptr);
#else
    INFO("no kernel memory allocation available !\n");
#endif
    return;
  }
  // if null pointer, do nothing as per spec
  if ( NULL == ptr )
    return;

  pool = cyg_memprot_mempools[domain];
  heap_lock(domain);
  // get pool to free it
  cyg_mempool_dlmalloc_free( pool, (cyg_uint8 *) ptr );
  heap_unlock(domain);
}

//! User mode calloc
void *cyg_memprot_calloc( size_t nmemb, size_t size, cyg_uint32 _lr ) {
  void *data_ptr;
  cyg_ucount32 realsize;
  int domain = __swi_cyg_memprot_get_current_domain(_lr);

  if( domain == 0 ) {
#if CYGINT_ISO_MALLOC
    return calloc( nmemb, size );
#else
    INFO("no kernel memory allocation available !\n");
    return NULL;
#endif
  }
  realsize = nmemb * size;
  data_ptr = cyg_memprot_malloc( realsize, _lr );
  // Fill with 0's if non-NULL
  if ( data_ptr != NULL )
    memset( data_ptr, 0, realsize );
  return data_ptr;
}

//! User mode realloc
void *cyg_memprot_realloc( void *ptr, size_t size, cyg_uint32 _lr ) {
  cyg_int32 oldsize;
  void *newptr;
  cyg_handle_t pool;
  int domain = __swi_cyg_memprot_get_current_domain(_lr);

  if( domain == 0 ) {
#if CYGINT_ISO_MALLOC
    return realloc( ptr, size );
#else
    INFO("no kernel memory allocation available !\n");
    return NULL;
#endif
  }
  // if pointer is NULL, we must malloc it
  if ( ptr == NULL ) {
    ptr = cyg_memprot_malloc( size, _lr );
    return ptr;
  } // if
    // if size is 0, we must free it
  if ( size == 0 ) {
    cyg_memprot_free(ptr, _lr);
    return NULL;
  } // if

  pool = cyg_memprot_mempools[domain];
  heap_lock(domain);
  // otherwise try to resize allocation
  newptr = cyg_mempool_dlmalloc_resize_alloc( pool, (cyg_uint8 *)ptr,
					      size, &oldsize);
  heap_unlock(domain);
  if ( NULL == newptr ) {
    newptr = cyg_memprot_malloc( size, _lr );
    if ( NULL != newptr ) {
      memcpy( newptr, ptr, size < (size_t) oldsize ? size : (size_t) oldsize );
      cyg_memprot_free( ptr, _lr );
    }
  }
  return newptr;
}
