//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
 * ============================================================================
 * @file  mmu.c
 * @brief MMU programming functions
 *
 * Memory protection MMU handling routines. This is a minimal set of functions
 * used to program the ARM v4 MMU. It is tailored to meet the needs of the
 * memprot package, and may not be suitable for another use. Specifically, tiny
 * pages (1 KB) and generic access permission setting are not supported.
 *
 * Copyright (C) 2004,2005 Parrot S.A.
 *
 * @author     ivan.djelic@parrot.fr
 * @date       2004-05-20
 * ============================================================================
 */

#include <cyg/kernel/kapi.h>
#include <cyg/hal/hal_mmu.h>
#include <cyg/infra/cyg_type.h>
#include <cyg/infra/diag.h>
#include <cyg/infra/cyg_trac.h>
#include <errno.h>
#include <string.h>
#include <assert.h>
#include <pkgconf/system.h>
#include <pkgconf/memprot.h>

#include "mmu.h"
#include "ram_alloc.h"
#include "diag.h"

// Platform memory layout
#include <pkgconf/system.h>
#include CYGHWR_MEMORY_LAYOUT_H

//! Default translation table base address
#define DEFAULT_TTB               CYGMEM_DEFAULT_TTB

//@{
/** Physical allocator defaults to RAM allocator.
 *
 * We distinguish the regular physical allocator from the MMU page table
 * allocator. This may be useful at startup if we need to allocate pages in
 * contiguous chunks, while allowing page table allocation at the same time.
 */
#define DEFAULT_PHYS_ALLOCATE     ram_allocate
#define DEFAULT_PHYS_FREE         ram_free
#define DEFAULT_PHYS_ALLOCATE_PT  ram_allocate
#define DEFAULT_PHYS_FREE_PT      ram_free
//@}

// ===========================================================================
// MMU specific types and macros
// ===========================================================================

//! The first level descriptor
typedef union {
  unsigned int word;
  struct {
    unsigned id     :  2;
    unsigned sbz    : 30;
  } fault;
  struct {
    unsigned id     :  2;
    unsigned imp    :  3;
    unsigned domain :  4;
    unsigned sbz    :  1;
    unsigned base   : 22;
  } page_table;
  struct {
    unsigned id     :  2;
    unsigned b      :  1;
    unsigned c      :  1;
    unsigned imp    :  1;
    unsigned domain :  4;
    unsigned sbz    :  1;
    unsigned ap     :  2;
    unsigned sbz1   :  8;
    unsigned base   : 12;
  } section;
  struct {
    unsigned id     :  2;
    unsigned imp    :  3;
    unsigned domain :  4;
    unsigned sbz    :  3;
    unsigned base   : 20;
  } fine_page_table;
} arm_mmu_l1_desc_t;

//! First level descriptor types
typedef enum {
  ARM_MMU_L1_FAULT = 0,
  ARM_MMU_L1_PAGE_TABLE,
  ARM_MMU_L1_SECTION,
  ARM_MMU_L1_FINE_PAGE_TABLE,
  ARM_MMU_L1_RESERVED = ARM_MMU_L1_FINE_PAGE_TABLE
} ARM_MMU_L1_DESC_ID;

//! The second level descriptor
typedef union {
  unsigned int word;
  struct {
    unsigned id     :  2;
    unsigned sbz    : 30;
  } fault;
  struct {
    unsigned id     :  2;
    unsigned b      :  1;
    unsigned c      :  1;
    unsigned ap0    :  2;
    unsigned ap1    :  2;
    unsigned ap2    :  2;
    unsigned ap3    :  2;
    unsigned sbz    :  4;
    unsigned base   : 16;
  } large_page;
  struct {
    unsigned id     :  2;
    unsigned b      :  1;
    unsigned c      :  1;
    unsigned ap0    :  2;
    unsigned ap1    :  2;
    unsigned ap2    :  2;
    unsigned ap3    :  2;
    unsigned base   : 20;
  } small_page;
  struct {
    unsigned id     :  2;
    unsigned b      :  1;
    unsigned c      :  1;
    unsigned ap     :  2;
    unsigned sbz    :  4;
    unsigned base   : 22;
  } tiny_page;
} arm_mmu_l2_desc_t;

/* Second level descriptor types */
typedef enum {
  ARM_MMU_L2_FAULT = 0,
  ARM_MMU_L2_LARGE_PAGE,
  ARM_MMU_L2_SMALL_PAGE,
  ARM_MMU_L2_TINY_PAGE,
  ARM_MMU_L2_RESERVED = ARM_MMU_L2_TINY_PAGE
} ARM_MMU_L2_DESC_ID;

//! Extract the page table physical address from a level one descriptor
#define ARM_MMU_PAGE_TABLE_BASE(d)	((arm_mmu_l2_desc_t *)\
					((d).word & 0xFFFFFC00))
//! Extract the section physical base from a level one descriptor
#define ARM_MMU_SECTION_BASE(d)		((d).word & 0xFFF00000)
//! Extract the large page physical base address from a level two descriptor
#define ARM_MMU_LARGE_PAGE_BASE(d)	((d).word & 0xFFFF0000)
//! Extract the small page physical base address from a level two descriptor
#define ARM_MMU_SMALL_PAGE_BASE(d)	((d).word & 0xFFFFF000)

#define ARM_MMU_SECTION_AP(_a_)	((_a_) << 10)
#define ARM_MMU_AP_ALL(_a)  (((_a) << 4)|((_a) << 6)|((_a) << 8)|((_a) << 10))

#define ARM_MMU_DOMAIN(_p)	(((_p)<16 ? (_p) : 15)<<5)

// Cache/Buffer parameters
#define ARM_MMU_BUFFERABLE   0x04
#define ARM_MMU_CACHEABLE    0x08
#define ARM_MMU_MEMTYPE(_c)  ((_c)? (ARM_MMU_BUFFERABLE|ARM_MMU_CACHEABLE) : 0)

// ============================================================================

//! returns the TTB start address (requires SYSTEM mode)
#define GET_TTB_BASE()	(arm_mmu_l1_desc_t *)( {	        \
	register unsigned long _ttb_base;			\
	asm volatile (						\
	    "mrc p15, 0, %0, c2, c0, 0;"			\
	    "mov %0, %0, lsr #14;"				\
	    "mov %0, %0, lsl #14;"				\
	    : "=r"(_ttb_base) );				\
	_ttb_base;						\
    } )

//! returns the DACR register (requires SYSTEM mode)
#define GET_DACR() ({   					\
	register unsigned long _dacr;				\
	asm volatile ( "mrc p15,0,%0,c3,c0,0" : "=r"(_dacr) );	\
	_dacr;							\
    })

//! sets the DACR register (requires SYSTEM mode)
#define SET_DACR(x) {						\
	register unsigned long _dacr = (x);			\
	asm volatile ( "mcr p15,0,%0,c3,c0,0" : : "r"(_dacr));  \
    }

//! returns the DFSR register (requires SYSTEM mode)
#define GET_DFSR()	({					\
	register unsigned long _dfsr;				\
	asm volatile ( "mrc p15,0,%0,c5,c0,0" : "=r"(_dfsr));   \
	_dfsr;							\
    })

//! returns the FAR register (requires SYSTEM mode)
#define GET_FAR()	({					\
	register unsigned long _far;				\
	asm volatile ( "mrc p15,0,%0,c6,c0,0" : "=r"(_far));    \
	_far;							\
    })

//! Invalidates a single entry from the D-TLB
#define INVALIDATE_TLB_ENTRY(a)	{				\
        register unsigned long _addr = (a);			\
	asm volatile ( "mcr p15,0,%0,c8,c6,1" : : "r"(_addr) ); \
    }

//! Invalidates the whole TLB
#define INVALIDATE_TLB()  asm volatile ( "mcr p15,0,r0,c8,c7,0" : : : "r0" )

/** Returns true if _P is aligned to a _S boundary.
 *
 * @param _P address
 * @param _S alignment size
 */
#define IS_ALIGNED(_P,_S)	  (((_P)&((_S)-1)) == 0 )

//@{
//! Some extra MMU defines
#define MMU_ABORT_TERMINAL_EXCEPTION            (0x2)
#define MMU_ABORT_VECTOR_EXCEPTION              (0x0)
#define MMU_ABORT_ALIGNMENT                     (0x1)
#define MMU_ABORT_ALIGNMENT2                    (0x3)
#define MMU_ABORT_EXT_TRANSLATION_L1            (0xC)
#define MMU_ABORT_EXT_TRANSLATION_L2            (0xE)
#define MMU_ABORT_TRANSLATION_SECTION           (0x5)
#define MMU_ABORT_TRANSLATION_PAGE              (0x7)
#define MMU_ABORT_DOMAIN_SECTION                (0x9)
#define MMU_ABORT_DOMAIN_PAGE                   (0xB)
#define MMU_ABORT_PERMISSION_SECTION            (0xD)
#define MMU_ABORT_PERMISSION_PAGE               (0xF)
#define MMU_ABORT_EXT_LINEFETCH_SECTION         (0x4)
#define MMU_ABORT_EXT_LINEFETCH_PAGE            (0x6)
#define MMU_ABORT_EXT_NON_LINEFETCH_SECTION     (0x8)
#define MMU_ABORT_EXT_NON_LINEFETCH_PAGE        (0xA)
//@}

static const char *mmu_abort_type[16] = {
  [MMU_ABORT_TERMINAL_EXCEPTION] = "terminal exception (irrecoverable)",
  [MMU_ABORT_VECTOR_EXCEPTION]  = "vector exception",
  [MMU_ABORT_ALIGNMENT] = "alignment fault",
  [MMU_ABORT_ALIGNMENT2] = "alignment fault",
  [MMU_ABORT_EXT_TRANSLATION_L1] = "external translation abort (L1)",
  [MMU_ABORT_EXT_TRANSLATION_L2] = "external translation abort (L2)",
  [MMU_ABORT_TRANSLATION_SECTION] = "section translation fault",
  [MMU_ABORT_TRANSLATION_PAGE] = "page translation fault",
  [MMU_ABORT_DOMAIN_SECTION] = "section domain fault",
  [MMU_ABORT_DOMAIN_PAGE] = "page domain fault",
  [MMU_ABORT_PERMISSION_SECTION] = "section permission fault",
  [MMU_ABORT_PERMISSION_PAGE] = "page permission fault",
  [MMU_ABORT_EXT_LINEFETCH_SECTION] = "section linefetch external abort",
  [MMU_ABORT_EXT_LINEFETCH_PAGE] = "page linefetch external abort",
  [MMU_ABORT_EXT_NON_LINEFETCH_SECTION]="section non-linefetch external abort",
  [MMU_ABORT_EXT_NON_LINEFETCH_PAGE] = "page non-linefetch external abort",
};

// ===========================================================================

//@{
//! Local configuration
static arm_mmu_l1_desc_t *ttb = (arm_mmu_l1_desc_t *)CYGMEM_DEFAULT_TTB;
static arm_mmu_phys_allocate_t phys_allocate = DEFAULT_PHYS_ALLOCATE;
static arm_mmu_phys_free_t phys_free = DEFAULT_PHYS_FREE;
static arm_mmu_phys_allocate_t phys_allocate_pt = DEFAULT_PHYS_ALLOCATE_PT;
static arm_mmu_phys_free_t phys_free_pt = DEFAULT_PHYS_FREE_PT;
//@}

//! a mutex to allow exclusive access to MMU table
static cyg_mutex_t pgtab_mutex;

#define lock_pt()   cyg_mutex_lock(&pgtab_mutex)   //!< locks MMU tables
#define unlock_pt() cyg_mutex_unlock(&pgtab_mutex) //!< unlocks MMU tables

// ===========================================================================
// Page table allocation routines
// ===========================================================================

static cyg_ucount32 pt_end = 0;
static cyg_ucount32 pt_free = 0;
// enough to manage 1024 x 4 x 1 MB = 4 GB of address space
static CYG_ADDRESS pt_table[1024];

/** A very basic page table allocation system.
 *
 * Each entry in a static array of pointers is used to manage 4 contiguous
 * 1k pages. The pointer itself points to the whole 4k area, while its lower
 * 4 bits indicate which 1k sub-page is in use and which is free.
 *
 * @return A pointer to the allocated table, or 0 upon failure.
 */
static arm_mmu_l2_desc_t *
alloc_page_table(void) {
  int j, slot, slots;
  CYG_ADDRESS new_pt = 0;
  if ( pt_free < pt_end ) {
    // we can use a free slot in some already used 4k ram pages
    new_pt = pt_table[pt_free] & 0xFFFFF000;
    // use 4 lower bits of pointer to record 1k pages in use
    slots = pt_table[pt_free] & 0xF;
    slot = -1;
    for ( j = 0; j < 4; j++ )
      if ( !( slots & ( 0x1 << j )) ) {
        slot = j;
        break;
      }
    CYG_ASSERT( slot >= 0, "bad slots in pt_table" );
    // address of 1k page
    new_pt += ( slot << 10 );
    // update slots in use
    pt_table[pt_free] |= ( 0x1 << slot );
    if (( pt_table[pt_free] & 0xF ) == 0xF )
      pt_free++;
  }
  else {
    // we need a fresh 4k page
    new_pt = phys_allocate_pt(4096);
    if ( !new_pt )
      return NULL;
    pt_table[pt_end++] = new_pt | 0x1;
    CYG_ASSERT( pt_end < 1024, "pt_end out of bounds" );
  }
  CYG_ASSERT((( new_pt & 0xFFFFF000 ) != 0 ),"unable to allocate page table" );
  memset( (void*)new_pt, 0, 1024 );
  DEBUG("alloc_page_table: 0x%08x\n", new_pt );
  return (arm_mmu_l2_desc_t *)new_pt;
}

/** A very basic page table allocation system.
 *
 * This function is used to free a page table allocated with
 * @ref alloc_page_table.
 *
 * @param pt pointer to the page table
 */
static void
free_page_table(arm_mmu_l2_desc_t *pt) {
  int i, pt_index, slot;
  CYG_ADDRESS pt_base = ((unsigned int)pt) & 0xFFFFF000;
  // search pt_table
  pt_index = -1;
  for ( i = 0; i < pt_end; i++ )
    if (( pt_table[i] & 0xFFFFF000 ) == pt_base ) {
      pt_index = i;
      break;
    }
  CYG_ASSERT( pt_index >= 0, "page table not in pt_table" );
  slot = (((unsigned int)pt) & 0xFFF ) >> 10;
  // update slots in use
  pt_table[pt_index] &= ~( 0x1 << slot );
  if (( pt_table[pt_index] & 0xF ) == 0 ) {
    // all slots are free, free ram page
    phys_free_pt(pt_base, 4096);
    // pt_free <= pt_index < pt_end
    // remove hole in table
    pt_table[pt_index] = pt_table[pt_end-1];
    pt_end--;
  }
  else {
    if ( pt_index < pt_free ) {
      // transfer pt_index entry to [pt_free;pt_end[
      pt_base = pt_table[pt_index];
      pt_table[pt_index] = pt_table[pt_free-1];
      pt_free--;
      pt_table[pt_free] = pt_base;
    }
  }
  DEBUG("free_page_table: 0x%08x\n", pt );
}

// ===========================================================================
// MMU routines
// ===========================================================================

/** Configure working TTB and physical RAM allocator/deallocator.
 *
 * This function allows you to change working translation table and physical
 * allocation routines from their default values.
 *
 * @param ttb Translation table base. Passing NULL restores the default value.
 * @param allocate Physical RAM allocator. Passing NULL restores the default
 * value.
 * @param free Physical RAM deallocator. Passing NULL restores the default
 * value.
 * @param allocate Physical page table allocator. Passing NULL restores the
 * default value.
 * @param free Physical page table deallocator. Passing NULL restores the
 * default value.
 */
void
memprot_mmu_configure( unsigned int *_ttb,
                       CYG_ADDRESS (*_allocate)(size_t),
                       void (*_free)(CYG_ADDRESS, size_t),
                       CYG_ADDRESS (*_allocate_pt)(size_t),
                       void (*_free_pt)(CYG_ADDRESS, size_t)) {
  ttb = ( _ttb )? (arm_mmu_l1_desc_t *)_ttb : (arm_mmu_l1_desc_t *)DEFAULT_TTB;
  phys_allocate = ( _allocate )? _allocate : DEFAULT_PHYS_ALLOCATE;
  phys_free = ( _free )? _free : DEFAULT_PHYS_FREE;
  phys_allocate_pt = ( _allocate_pt )? _allocate_pt : DEFAULT_PHYS_ALLOCATE_PT;
  phys_free_pt = ( _free_pt )? _free_pt : DEFAULT_PHYS_FREE_PT;
}

/** Returns the page size to which a virtual address is mapped.
 *
 * This internal function returns the page size to which a virtual address is
 * mapped, or 0 if the address is unmapped.
 *
 * @param virtual The virtual address
 * @return The page size (0, 4 KB, 64 KB or 1MB)
 */
static size_t
memprot_mmu_page_size( CYG_ADDRESS virtual ) {
  arm_mmu_l1_desc_t section;
  arm_mmu_l2_desc_t *page_table;
  arm_mmu_l2_desc_t page;
  size_t size;

  section = ttb[ virtual >> 20 ];
  switch ( section.fault.id ) {
  case ARM_MMU_L1_SECTION :
    size = MMU_SECTION_SIZE;
    break;
  case ARM_MMU_L1_PAGE_TABLE :
    page_table = ARM_MMU_PAGE_TABLE_BASE( section );
    page = page_table[ (virtual >> 12) & 0xff ];
    switch ( page.fault.id ) {
    case ARM_MMU_L2_SMALL_PAGE :
      size = MMU_PAGE_SIZE;
      break;
    case ARM_MMU_L2_LARGE_PAGE :
      size = MMU_LARGE_PAGE_SIZE;
      break;
    default :
      size = 0;
      break;
    }
    break;
  default :
    size = 0;
    break;
  }
  return size;
}

/** Returns the physical address to which a virtual address is mapped.
 *
 * @param vaddr The virtual address
 * @return The physical address, or 0 is vaddr is unmapped
 */
CYG_ADDRESS
memprot_mmu_physical_address( CYG_ADDRESS vaddr ) {
  arm_mmu_l1_desc_t section;
  arm_mmu_l2_desc_t *table;
  arm_mmu_l2_desc_t ltd;
  CYG_ADDRESS pa;
  int index;

  lock_pt();
  section = ttb[ vaddr >> 20 ];

  switch ( section.fault.id ) {

  case ARM_MMU_L1_SECTION :
    pa = ARM_MMU_SECTION_BASE(section) + (vaddr & 0x000FFFFF);
    break;

  case ARM_MMU_L1_PAGE_TABLE :
    table = ARM_MMU_PAGE_TABLE_BASE(section);
    index = ( vaddr >> 12 ) & 0xFF;
    ltd = table[index];
    switch ( ltd.fault.id ) {
    case ARM_MMU_L2_SMALL_PAGE :
      pa = ARM_MMU_SMALL_PAGE_BASE(ltd) + (vaddr & 0xFFF);
      break;
    case ARM_MMU_L2_LARGE_PAGE :
      pa = ARM_MMU_LARGE_PAGE_BASE(ltd) + (vaddr & 0xFFFF);
      break;
    default :
      pa = 0;
      break;
    }
    break;

  default:
    pa = 0;
    break;
  }
  unlock_pt();
  return pa;
}

/** Get a page table for a given virtual address.
 *
 * This internal function is used to get a page table in order to map large or
 * small pages. If there is no page table describing vaddr then it is created.
 *
 * @param vaddr The virtual address for which we need a page table
 * @param domain The domain to which the virtual address should belong
 * @return Returns a pointer to the page table, or 0 upon failure.
 */
static arm_mmu_l2_desc_t *
memprot_mmu_get_page_table(CYG_ADDRESS vaddr, int domain) {
  arm_mmu_l2_desc_t *table;
  arm_mmu_l1_desc_t l1d = ttb[vaddr >> 20];

  switch ( l1d.fault.id ) {
  case ARM_MMU_L1_FAULT :
    // no page table, let's allocate one
    if ( !( table = alloc_page_table() ))
      return 0;
    ttb[vaddr >> 20].word = 0x10 |
      ((CYG_ADDRESS)table) |
      ARM_MMU_DOMAIN(domain) |
      ARM_MMU_L1_PAGE_TABLE;
    INVALIDATE_TLB_ENTRY(vaddr);
    break;
  case ARM_MMU_L1_PAGE_TABLE :
    // there is already a page table
    table = ARM_MMU_PAGE_TABLE_BASE(l1d);
    break;
  default :
    // descriptor already in use !
    return 0;
  }
  return table;
}

/** Map an address range.
 *
 * This function maps an address range of virtual memory to physical memory.
 *
 * @param vaddr The virtual address
 * @param paddr The physical address
 * @param size The address range
 * @param ap The access permission
 * @param cache The caching behaviour
 * @param domain The page domain (0-15)
 * @param allocate A boolean value indicating if physical ram should be
 * allocated or not (1 = allocate)
 * @param use_small_pages_only Forces use of small pages for the whole range.
 * Passing 1 maps the address range using only small 4 KB pages.
 * @return ENOERR when successful, ENOMEM if out of memory or if vaddr is
 * already mapped
 */
int
memprot_mmu_map( CYG_ADDRESS vaddr, CYG_ADDRESS paddr, size_t size, int ap,
                 int cache, int domain, int allocate,
                 int use_small_pages_only ) {
  int i, status  = ENOERR;
  arm_mmu_l1_desc_t l1d;

  // There should be no need to flush cache here if it is done in unmap()

  // a small sanity check first
  if ( !IS_ALIGNED( vaddr, MMU_PAGE_SIZE ) ||
       ( !allocate && !IS_ALIGNED( paddr, MMU_PAGE_SIZE )) ||
       !IS_ALIGNED( size, MMU_PAGE_SIZE ))
    return ENOMEM;

  lock_pt();

  while ( size ) {
    l1d = ttb[vaddr >> 20];

    if ( !use_small_pages_only ) {
      // First try to map a section
      if (( size >= MMU_SECTION_SIZE )&&
          ( IS_ALIGNED( vaddr, MMU_SECTION_SIZE ))&&
          (( !allocate && IS_ALIGNED( paddr, MMU_SECTION_SIZE ))||
           ( allocate && ( paddr = phys_allocate(MMU_SECTION_SIZE))))) {
        // vaddr and paddr are properly aligned, now map section
        // check that section is not already mapped
        if ( l1d.fault.id != ARM_MMU_L1_FAULT ) {
          status = ENOMEM;
          break;
        }
        DEBUG("memprot_mmu_map: SECTION 0x%08x -> 0x%08x\n", vaddr, paddr );
        ttb[vaddr >> 20].word = 0x10 |
          paddr |
          ARM_MMU_MEMTYPE(cache) |
          ARM_MMU_DOMAIN(domain) |
          ARM_MMU_SECTION_AP(ap) |
          ARM_MMU_L1_SECTION;
        vaddr += MMU_SECTION_SIZE;
        paddr += MMU_SECTION_SIZE;
        size -= MMU_SECTION_SIZE;
        INVALIDATE_TLB_ENTRY(vaddr);
        continue;
      }

      // Section mapping failed, try to map a large page
      if (( size >= MMU_LARGE_PAGE_SIZE )&&
          ( IS_ALIGNED( vaddr, MMU_LARGE_PAGE_SIZE ))&&
          (( !allocate && IS_ALIGNED( paddr, MMU_LARGE_PAGE_SIZE ))||
           ( allocate && ( paddr = phys_allocate(MMU_LARGE_PAGE_SIZE))))) {
        int index;
        arm_mmu_l2_desc_t *table, l2d;
        table = memprot_mmu_get_page_table(vaddr, domain);
        if ( !table ) {
          status = ENOMEM;
          break;
        }
        // is table entry already in use ?
        index = (vaddr & 0x000FF000) >> 12;
        for ( i = index; i < index+16; i++ )
          if ( table[i].fault.id != ARM_MMU_L2_FAULT ) {
            status = EINVAL;
            break;
          }
        l2d.word =
          paddr |
          ARM_MMU_MEMTYPE(cache) |
          ARM_MMU_AP_ALL(ap) |
          ARM_MMU_L2_LARGE_PAGE;
        DEBUG("memprot_mmu_map: LARGE_PAGE 0x%08x -> 0x%08x\n", vaddr, paddr );
        // duplicate level two descriptor
        for ( i = index; i < index+16; i++ )
          table[i] = l2d;

        // we need to invalidate all four subpages separately
        // (cf. ARM926-EJS Tech. Manual page 3-19)
        INVALIDATE_TLB_ENTRY(vaddr + 0*MMU_PAGE_SIZE);
        INVALIDATE_TLB_ENTRY(vaddr + 4*MMU_PAGE_SIZE);
        INVALIDATE_TLB_ENTRY(vaddr + 8*MMU_PAGE_SIZE);
        INVALIDATE_TLB_ENTRY(vaddr + 12*MMU_PAGE_SIZE);

        vaddr += MMU_LARGE_PAGE_SIZE;
        paddr += MMU_LARGE_PAGE_SIZE;
        size -= MMU_LARGE_PAGE_SIZE;
        continue;
      }
    }

    // Large page mapping failed, try to map a small page
    if (( size >= MMU_PAGE_SIZE )&&
        ( IS_ALIGNED( vaddr, MMU_PAGE_SIZE ))&&
        (( !allocate && IS_ALIGNED( paddr, MMU_PAGE_SIZE ))||
         ( allocate && ( paddr = phys_allocate(MMU_PAGE_SIZE))))) {
      int index;
      arm_mmu_l2_desc_t *table, l2d;
      table = memprot_mmu_get_page_table(vaddr, domain);
      if ( !table ) {
        status = ENOMEM;
        break;
      }
      // is table entry already in use ?
      index = (vaddr & 0x000FF000) >> 12;
      if ( table[index].fault.id != ARM_MMU_L2_FAULT ) {
        status = EINVAL;
        break;
      }
      l2d.word =
        paddr |
        ARM_MMU_MEMTYPE(cache) |
        ARM_MMU_AP_ALL(ap) |
        ARM_MMU_L2_SMALL_PAGE;
      table[index] = l2d;
      DEBUG("memprot_mmu_map: SMALL_PAGE 0x%08x -> 0x%08x\n", vaddr, paddr );
      INVALIDATE_TLB_ENTRY(vaddr);
      vaddr += MMU_PAGE_SIZE;
      paddr += MMU_PAGE_SIZE;
      size -= MMU_PAGE_SIZE;
      continue;
    }

    // mapping failed
    status = ENOMEM;
    break;
  }
  unlock_pt();
  return status;
}

/** Split a section/large page into smaller pages.
 *
 * This internal function is called from \ref memprot_mmu_unmap whenever a
 * mapped range needs to be partially unmapped.
 */
static int
memprot_mmu_split( CYG_ADDRESS vaddr ) {
  arm_mmu_l1_desc_t section;
  arm_mmu_l2_desc_t *table;
  arm_mmu_l2_desc_t page;
  int i, domain, ap, cache, index;
  CYG_ADDRESS paddr;

  section = ttb[vaddr >> 20];
  switch ( section.fault.id ) {
  case ARM_MMU_L1_SECTION :
    domain = section.section.domain;
    paddr = ARM_MMU_SECTION_BASE(section);
    DEBUG("splitting SECTION: 0x%08x -> 0x%08x\n", vaddr, paddr );
    ap = section.section.ap;
    cache = section.word & 0xC;
    // we need a page table
    if ( !( table = alloc_page_table() ))
      return -1;
    ttb[vaddr >> 20].word = 0x10|
      ((CYG_ADDRESS)table) |
      ARM_MMU_DOMAIN(domain) |
      ARM_MMU_L1_PAGE_TABLE;
    // fill in page table with large pages
    for ( i = 0; i < 256; i += 16 ) {
      int j;
      arm_mmu_l2_desc_t l2d;
      l2d.word =
        (paddr + i*MMU_PAGE_SIZE) |
        cache |
        ARM_MMU_AP_ALL(ap) |
        ARM_MMU_L2_LARGE_PAGE;
      // duplicate level two descriptor
      for ( j = i; j < i+16; j++ )
        table[j] = l2d;
    }
    INVALIDATE_TLB_ENTRY(vaddr);
    return 0;
  case ARM_MMU_L1_PAGE_TABLE :
    table = ARM_MMU_PAGE_TABLE_BASE( section );
    index = (vaddr >> 12) & 0xff;
    page = table[index];
    switch ( page.fault.id ) {
    case ARM_MMU_L2_SMALL_PAGE :
      CYG_FAIL("Trying to split a small page !");
    case ARM_MMU_L2_LARGE_PAGE :
      // split large page into small pages
      vaddr &= 0xFFFF0000;
      index &= 0xf0;
      page.word = ( table[index].word & ~0x3 ) | ARM_MMU_L2_SMALL_PAGE;
      DEBUG( "splitting LARGE PAGE: 0x%08x -> 0x%08x\n", vaddr,
             ARM_MMU_LARGE_PAGE_BASE(page));
      for ( i = 0; i < 16; i++ ) {
        table[index+i].word = page.word|(i << 12);
      }
      // we need to invalidate all four subpages separately
      // (cf. ARM926-EJS Tech. Manual page 3-19)
      INVALIDATE_TLB_ENTRY(vaddr);
      INVALIDATE_TLB_ENTRY(vaddr + 4*MMU_PAGE_SIZE);
      INVALIDATE_TLB_ENTRY(vaddr + 8*MMU_PAGE_SIZE);
      INVALIDATE_TLB_ENTRY(vaddr + 12*MMU_PAGE_SIZE);
      return 0;
    }
  }
  CYG_FAIL("Trying to split an unmapped area !");
  return -1;
}

/** Unmap an address range.
 *
 * This function unmaps a virtual memory range. It is able to free underlying
 * physical ram (deallocate = 1) if it was previously allocated with
 * \ref memprot_mmu_map.
 *
 * @param vaddr The virtual address
 * @param size The address range
 * @param deallocate A boolean value indicating if physical ram should be freed
 * @return ENOERR when successful, EINVAL if bad parameters were passed
 */

int
memprot_mmu_unmap( CYG_ADDRESS vaddr, size_t size, int deallocate ) {
  int i, index, status = ENOERR;
  CYG_ADDRESS paddr;
  size_t psize;
  arm_mmu_l1_desc_t l1d;
  arm_mmu_l2_desc_t *table;

  // a small sanity check first
  if ( !IS_ALIGNED(size, MMU_PAGE_SIZE)|| !IS_ALIGNED(vaddr, MMU_PAGE_SIZE))
    return EINVAL;

  FLUSH_CACHE();
  lock_pt();

  while ( size ) {
    l1d = ttb[vaddr >> 20];
    table = 0;
    psize = memprot_mmu_page_size(vaddr);
    // if vaddr is already unmapped we signal an error
    CYG_ASSERT( psize, "Unmapping an already unmapped range !" );

    if (( psize > size )||( !IS_ALIGNED( vaddr, psize ))) {
      // split section or large page into smaller pages
      if ( memprot_mmu_split(vaddr) < 0 ) {
        status = ENOMEM;
        break;
      }
      // try again
      continue;
    }
    // unmap a 1 MB/64 KB/4 KB range, we assume vaddr is aligned with psize
    switch ( psize ) {

    case MMU_SECTION_SIZE:
      paddr = ARM_MMU_SECTION_BASE(l1d);
      ttb[vaddr >> 20].word = ARM_MMU_L1_FAULT;
      DEBUG("memprot_mmu_unmap: SECTION 0x%08x\n", vaddr );
      INVALIDATE_TLB_ENTRY(vaddr);
      break;

    case MMU_LARGE_PAGE_SIZE:
      table = ARM_MMU_PAGE_TABLE_BASE(l1d);
      index = (vaddr & 0x000FF000) >> 12;
      paddr = ARM_MMU_LARGE_PAGE_BASE(table[index]);
      DEBUG("memprot_mmu_unmap: LARGE PAGE 0x%08x\n", vaddr );
      for ( i = index; i < index+16; i++ )
        table[i].word = ARM_MMU_L2_FAULT;
      // we need to invalidate all four subpages separately
      // (cf. ARM926-EJS Tech. Manual page 3-19)
      INVALIDATE_TLB_ENTRY(vaddr + 0*MMU_PAGE_SIZE);
      INVALIDATE_TLB_ENTRY(vaddr + 4*MMU_PAGE_SIZE);
      INVALIDATE_TLB_ENTRY(vaddr + 8*MMU_PAGE_SIZE);
      INVALIDATE_TLB_ENTRY(vaddr + 12*MMU_PAGE_SIZE);
      break;

    case MMU_PAGE_SIZE:
      table = ARM_MMU_PAGE_TABLE_BASE(l1d);
      index = (vaddr & 0x000FF000) >> 12;
      paddr = ARM_MMU_SMALL_PAGE_BASE(table[index]);
      table[index].word = ARM_MMU_L2_FAULT;
      DEBUG("memprot_mmu_unmap: SMALL PAGE 0x%08x\n", vaddr );
      // we need to invalidate all four subpages separately
      // (cf. ARM926-EJS Tech. Manual page 3-19)
      INVALIDATE_TLB_ENTRY(vaddr);
      INVALIDATE_TLB_ENTRY(vaddr + 0x400);
      INVALIDATE_TLB_ENTRY(vaddr + 0x800);
      INVALIDATE_TLB_ENTRY(vaddr + 0xc00);
      break;

    default:
      paddr = 0;
      CYG_FAIL("Invalid page size");
    }
    // check if page table is empty
    if ( table ) {
      int page_used = 0;
      for ( i = 0; (( i < 256 )&&( !page_used )); i++ )
        page_used |= ( table[i].fault.id != ARM_MMU_L2_FAULT );
      if ( !page_used ) {
        free_page_table(table);
        // unmap page table
        ttb[vaddr >> 20].word = ARM_MMU_L1_FAULT;
        INVALIDATE_TLB_ENTRY(vaddr & 0xFFF00000); // is this right ?
      }
    }
    // free ram if required
    if ( deallocate )
      phys_free(paddr, psize);

    size -= psize;
    vaddr += psize;
  }
  unlock_pt();
  return status;
}

// ===========================================================================

/** Setup MMU TTB.
 *
 * This function is called in SYSTEM mode to set up the MMU TTB.
 */
/*
  void memprot_mmu_init(unsigned int *_ttb) {
  // setup ttb
  DEBUG("memprot_mmu_init: ttb = 0x%08x\n", ttb );
  asm volatile ("mcr p15,0,%0,c2,c0,0" : : "r"(ttb) );
  }
*/
/** Returns the domain to which a virtual address is mapped.
 *
 * This function is called in SYSTEM mode to get the page domain to which a
 * virtual address is mapped.
 *
 * @param vaddr The virtual address
 * @return The domain (0-15) or -1 if the address is unmapped
 */
int
memprot_mmu_get_domain( CYG_ADDRESS vaddr ) {
  arm_mmu_l1_desc_t l1d = ttb[ vaddr >> 20 ];
  if ( l1d.fault.id == ARM_MMU_L1_FAULT )
    return -1;
  return l1d.section.domain;
}

/** Returns the access rights of a virtual address.
 *
 * This function is called in SYSTEM mode to get the access rights of a virtual
 * address.
 * @param vaddr The virtual address
 * @return The access rights (AP_RO_RO, AP_RW_NA, AP_RW_RO or AP_RW_RW or -1
 * if the address is unmapped)
 */
int
memprot_mmu_get_ap( CYG_ADDRESS vaddr ) {
  arm_mmu_l1_desc_t section;
  arm_mmu_l2_desc_t *page_table;
  arm_mmu_l2_desc_t page;
  int ap = -1, subpage;

  lock_pt();
  section = ttb[ vaddr >> 20 ];
  switch ( section.fault.id ) {
  case ARM_MMU_L1_SECTION :
    ap = section.section.ap;
    break;
  case ARM_MMU_L1_PAGE_TABLE :
    page_table = ARM_MMU_PAGE_TABLE_BASE( section );
    page = page_table[ (vaddr>>12) & 0xff ];
    switch ( page.fault.id ) {
    case ARM_MMU_L2_SMALL_PAGE :
    case ARM_MMU_L2_LARGE_PAGE :
      if ( page.fault.id == ARM_MMU_L2_SMALL_PAGE )
        subpage = (( vaddr & 0xfff ) >> 10 );
      else
        subpage = (( vaddr & 0xffff ) >> 14 );
      switch ( subpage ) {
      case 0:
        ap = page.small_page.ap0;
        break;
      case 1:
        ap = page.small_page.ap1;
        break;
      case 2:
        ap = page.small_page.ap2;
        break;
      case 3:
        ap = page.small_page.ap3;
        break;
      }
      break;
    default :
      break;
    }
    break;
  default :
    break;
  }
  unlock_pt();
  return ap;
}

/** Sets the access rights of a virtual address.
 *
 * This function sets the access rights of a virtual address (at the smallest
 * subpage level).
 *
 * @param vaddr The virtual address
 * @param ap The access rights
 * @return 1 if successful, 0 upon failure
 */
int
memprot_mmu_set_subpage_ap( CYG_ADDRESS vaddr, int ap ) {
  arm_mmu_l1_desc_t section, *psection;
  arm_mmu_l2_desc_t *page_table;
  arm_mmu_l2_desc_t *page;
  int subpage, ret = 1;

  lock_pt();
  section = ttb[ vaddr >> 20 ];
  switch ( section.fault.id ) {

  case ARM_MMU_L1_SECTION :
    psection = &ttb[ vaddr >> 20 ];
    psection->section.ap = ap;
    break;

  case ARM_MMU_L1_PAGE_TABLE :
    page_table = ARM_MMU_PAGE_TABLE_BASE( section );
    page = &page_table[ (vaddr>>12) & 0xff ];

    switch ( page->fault.id ) {
    case ARM_MMU_L2_SMALL_PAGE :
    case ARM_MMU_L2_LARGE_PAGE :
      if ( page->fault.id == ARM_MMU_L2_SMALL_PAGE )
        subpage = (( vaddr & 0xfff ) >> 10 );
      else
        subpage = (( vaddr & 0xffff ) >> 14 );
      switch ( subpage ) {
      case 0:
        page->small_page.ap0 = ap;
        break;
      case 1:
        page->small_page.ap1 = ap;
        break;
      case 2:
        page->small_page.ap2 = ap;
        break;
      case 3:
        page->small_page.ap3 = ap;
        break;
      }
      break;
    default :
      ret = 0;
      break;
    }
    break;
  default :
    ret = 0;
  }
  unlock_pt();
  // invalidate page entry
  if ( ret )
    INVALIDATE_TLB_ENTRY(vaddr);
  return ret;
}

//! Setup DACR and enable MMU
void memprot_mmu_enable(unsigned int *_ttb) {
  register unsigned long _cr1;
  // set TTB
  asm volatile ("mcr p15,0,%0,c2,c0,0" : : "r"(_ttb));
  // flush TLB
  INVALIDATE_TLB();
  // set dacr to client for all domains
  SET_DACR(0x55555555);
  // read control register cr1
  asm volatile ("mrc p15,0,%0,c1,c0,0" : "=r" (_cr1));
  // enable MMU, ROM protection and cache
  _cr1 |= MMU_Control_M | MMU_Control_R | MMU_Control_C | MMU_Control_I;
  // disable system protection
  _cr1 &= ~MMU_Control_S;
  // and write register back
  asm volatile ("mcr p15,0,%0,c1,c0,0" : : "r"(_cr1));
}

//! Disable MMU
void memprot_mmu_disable(void) {
  register unsigned long _cr1;
  // read control register cr1
  asm volatile ("mrc p15,0,%0,c1,c0,0" : "=r" (_cr1));
  // disable MMU and cache
  _cr1 &= ~( MMU_Control_M |
             MMU_Control_C |
             MMU_Control_I |
             MMU_Control_R |
             MMU_Control_S );
  // and write register back
  asm volatile ("mcr p15,0,%0,c1,c0,0\n\tnop\n\tnop\n\tnop" : : "r"(_cr1));
  // flush caches and write buffer
  FLUSH_CACHE();
}

cyg_uint32 memprot_mmu_get_DACR(void) {
  return GET_DACR();
}

/** Translation table switch.
 *
 * This function allows you to perform a 'hot' switch between two translation
 * tables, i.e. without disabling the MMU. Of course, this function should
 * better be mapped at the same address in both translation tables...
 */
void memprot_mmu_switch_ttb( unsigned int *_ttb ) {
  cyg_uint32 old;
  HAL_DISABLE_INTERRUPTS(old);
  // set TTB
  asm volatile ("mcr p15,0,%0,c2,c0,0" : : "r"(_ttb));
  INVALIDATE_TLB();
  HAL_RESTORE_INTERRUPTS(old);
}


/** Modify DACR using a read-modify-write sequence.
 *
 * This function is used to atomically change the contents of the DACR register
 *
 * @param dacr The new DACR value
 * @param mask The write mask, s.t.
 * new_DACR = (old_DACR & ~mask)|( dacr & mask)
 */
void memprot_mmu_set_DACR(cyg_uint32 dacr, cyg_uint32 mask) {
  cyg_uint32 old, olddacr;
  HAL_DISABLE_INTERRUPTS(old);
  olddacr = GET_DACR();
  SET_DACR((olddacr & ~mask)|(dacr & mask));
  HAL_RESTORE_INTERRUPTS(old);
}

//! Dumps the entire memory map, useful for debugging
void memprot_mmu_dump_memory_map(void) {
  static const char * ap2str[4] = { "RO_RO", "RW_NA", "RW_RO", "RW_RW" };
  int i, ap, domain, repeat = 0;
  CYG_ADDRESS va, pa;
  arm_mmu_l1_desc_t l1d, l1d_old;
  l1d_old.word = 0xffffffff;

  for ( i = 0; i < 4096; i++ ) {
    l1d = ttb[i];
    if (( !repeat )&&( l1d.word == l1d_old.word )) {
      repeat = 1;
      INFO("...\n");
    }
    else if ( l1d.word != l1d_old.word ) {
      int j, repeat2 = 0;
      arm_mmu_l2_desc_t ltd, ltd_old, *table;
      repeat = 0;
      ltd_old.word = 0xffffffff;
      // print section map
      switch ( l1d.fault.id ) {
      case ARM_MMU_L1_FAULT :
        INFO( "section 0x%08x -> unmapped\n", i << 20 );
        break;

      case ARM_MMU_L1_SECTION :
        va = i << 20;
        pa = ARM_MMU_SECTION_BASE(l1d);
        domain = l1d.section.domain;
        ap = l1d.section.ap;
        INFO( "section 0x%08x -> 0x%08x dom=%02d ap=%s C=%d B=%d\n",
              va, pa, domain, ap2str[ap], l1d.section.c,l1d.section.b);
        break;

      case ARM_MMU_L1_PAGE_TABLE :
        domain = l1d.page_table.domain;
        INFO( "section 0x%08x -> page table dom=%02d\n", i << 20, domain );
        table = ARM_MMU_PAGE_TABLE_BASE(l1d);
        for ( j = 0; j < 256; j++ ) {
          va = ( i << 20 ) + ( j << 12 );
          ltd = table[j];

          if (( !repeat2 )&&( ltd.word == ltd_old.word )) {
            repeat2 = 1;
            INFO("    ...\n");
          }
          else if ( ltd.word != ltd_old.word ) {

            // print page map
            switch ( ltd.fault.id ) {
            case ARM_MMU_L2_FAULT :
              INFO( "    small page 0x%08x -> unmapped\n", va );
              break;
            case ARM_MMU_L2_SMALL_PAGE :
              pa = ARM_MMU_SMALL_PAGE_BASE(ltd);
              ap = ltd.small_page.ap3;
              INFO( "    small page 0x%08x -> 0x%08x ap=%s C=%d B=%d\n",
                    va, pa, ap2str[ap], ltd.small_page.c, ltd.small_page.b);
              break;
            case ARM_MMU_L2_LARGE_PAGE :
              pa = ARM_MMU_LARGE_PAGE_BASE(ltd);
              ap = ltd.large_page.ap3;
              INFO( "    large page 0x%08x -> 0x%08x ap=%s C=%d B=%d\n",
                    va, pa, ap2str[ap], ltd.large_page.c, ltd.large_page.b);
              break;
            default :
              INFO( "    page 0x%08x -> ????????\n", va );
              break;
            }
          }
          ltd_old.word = ltd.word;
        }

        break;
      default :
        INFO( "section 0x%08x -> ????????\n", i << 20 );
        break;
      }
    }
    l1d_old.word = l1d.word;
  }
}

/** Get information about a DATA abort exception.
 *
 * This function is called from the exception handler
 * (@ref cyg_memprot_exception_handler) to get information about a DATA abort
 * exception.
 */
void memprot_mmu_get_data_abort_info( CYG_ADDRESS *vaddr, int *domain,
                                      char ** msg ) {
  unsigned long fsr;
  int type;
  fsr = GET_DFSR();
  type = fsr & 0xF;
  *vaddr = GET_FAR();
  *msg = (char *) mmu_abort_type[type];
  if (( type != MMU_ABORT_TRANSLATION_SECTION )&&
      ( type != MMU_ABORT_EXT_TRANSLATION_L1 )&&
      ( type != MMU_ABORT_ALIGNMENT )&&
      ( type != MMU_ABORT_ALIGNMENT2 )&&
      ( type != MMU_ABORT_VECTOR_EXCEPTION )&&
      ( type != MMU_ABORT_TERMINAL_EXCEPTION ))
    *domain = (fsr & 0xF0) >> 4;
  else
    // no domain for vaddr
    *domain = -1;
}
