//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
 * ============================================================================
 * @file  syscall.c
 * @brief A few functions used to check system calls arguments.
 *
 * Memory protection system call handling.
 *
 * Copyright (C) 2004,2005 Parrot S.A.
 *
 * @author     ivan.djelic@parrot.fr
 * @date       2004-04-04
 * ============================================================================
 */

#include "mmu.h"
#include "thread.h"
#include "syscall.h"

#define USER_READ     0
#define USER_WRITE    1
#define SYSTEM_READ   2
#define SYSTEM_WRITE  3

//! System access = read-only, User access = read-only (assuming S=0, R=1)
#define AP_RO_RO	0x0
//! System access = read/write, User access = no access
#define AP_RW_NA	0x1
//! System access = read/write, User access = read-only
#define AP_RW_RO	0x2
//! System access = read/write, User access = read/write
#define AP_RW_RW	0x3

/** Check access to a memory range.
 *
 * This function is used in several macros (see syscall.h) to check if a
 * given access to a given memory area is valid. Note that this check is only
 * partial since only the starting and ending addresses are verified. This
 * should be enough for sizes less than the MMU page size (4KB).
 *
 * @param addr The memory range start address
 * @param size The memory range size
 * @param access One of USER_READ, USER_WRITE, SYSTEM_READ, SYSTEM_WRITE
 * @return '1' if access is authorized, '0' otherwise.
 */
int __check_access( CYG_ADDRESS addr, cyg_uint32 size, int access ) {
  int context = cyg_memprot_get_current_domain(0);
  int dom1 = memprot_mmu_get_domain(addr);
  int dom2 = memprot_mmu_get_domain(addr+size);
  int ap1 = memprot_mmu_get_ap(addr);
  int ap2 = memprot_mmu_get_ap(addr+size);

  if ( context == 0 ) {
    // kernel context
    switch ( access ) {
    case USER_READ:
    case USER_WRITE:
    case SYSTEM_READ:
      return (( ap1 >= AP_RO_RO )&&( ap2 >= AP_RO_RO ));
    case SYSTEM_WRITE:
      return (( ap1 >= AP_RW_NA )&&( ap2 >= AP_RW_NA ));
    default:
      return 0;
    }
  }
  // user context
  switch ( access ) {
  case USER_READ:
    return ((( dom1 == 0 )&&( dom2 == 0 )&&
	     ((( ap1 == AP_RO_RO )&&( ap2 == AP_RO_RO ))||
	      (( ap1 == AP_RW_RO )&&( ap2 == AP_RW_RO ))))||
	    (( dom1 == context )&&( dom2 == context )&&
	     (( ap1 >= AP_RO_RO )&&( ap2 >= AP_RO_RO )&&
	      ( ap1 != AP_RW_NA )&&( ap2 != AP_RW_NA ))));

  case USER_WRITE:
    return (( dom1 == context )&&( dom2 == context )&&
	    ( ap1 == AP_RW_RW )&&( ap2 == AP_RW_RW ));

  case SYSTEM_READ:
    return (((( dom1 == 0 )&&( dom2 == 0 ))||
	     (( dom1 == context )&&( dom2 == context )))&&
	    ( ap1 >= AP_RO_RO )&&( ap2 >= AP_RO_RO ));

  case SYSTEM_WRITE:
    return (((( dom1 == 0 )&&( dom2 == 0 ))||
	     (( dom1 == context )&&( dom2 == context )))&&
	    ( ap1 >= AP_RW_NA )&&( ap2 >= AP_RW_NA ));
  default:
    return 0;
  }
  return 0;
}

/** Check ownership of a memory range.
 *
 * This function is used to check if the current execution context matches
 * the domain of a given memory range.
 *
 * @param addr The memory range start address
 * @param size The memory range size
 * @return '1' if we are in system context, or if we are in user context and
 * we own the memory range, and '0' otherwise.
 */
int __check_owner_domain( CYG_ADDRESS addr, cyg_uint32 size ) {
  int context = cyg_memprot_get_current_domain(0);
  int dom1 = memprot_mmu_get_domain(addr);
  int dom2 = memprot_mmu_get_domain(addr+size);
  return (( context == 0 )||(( context == dom1 )&&( context == dom2 )));
}
