//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
*******************************************************************************
* @file  thread.c
* @brief Thread handling functions
*
* Memory protection thread handling.
*
* Copyright (C) 2004,2005,2006 Parrot S.A.
*
* @author     ivan.djelic@parrot.fr
* @date       2004-05-04
*******************************************************************************
*/

#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <cyg/kernel/kapi.h>
#include <cyg/hal/hal_intr.h>
#include <cyg/hal/hal_arch.h>
#include <cyg/infra/diag.h>
#include <cyg/infra/cyg_ass.h>
#include <cyg/hal/hal_mmu.h>

#include <pkgconf/hal.h>
#include <pkgconf/hal_arm.h>

#include <pkgconf/memprot.h>

#include "mmu.h"
#include "diag.h"
#include "vp_layout.h"
#include "syscall.h"

extern cygmem_vp_layout_t cygmem_vp_layouts[16];
extern int memprot_data_index;

//! This structure is stored at the bottom of the thread stack.
typedef struct _cyg_memprot_thread_data_t {
  //! A small private stack to handle system calls
  CYG_ADDRESS swi_stack[16];
  //! Redirection of thread entry
  cyg_thread_entry_t *entry;
  //! System-allocated stack address is stored here (for later deallocation)
  CYG_ADDRESS stack_base;
  cyg_uint32 stack_size;
} cyg_memprot_thread_data_t;

//! This macro computes the value of the DACR register for a given user context
#define USER_DACR(_domain) (( 0x1 << (2*(_domain))) | 0x1 )

/** This macro computes the address of a userland thread private data.
 * Note: here we assume sizeof(cyg_thread) is 32-bit aligned...
 */
#define USER_THREAD(_thread) \
  ((cyg_memprot_thread_data_t *) \
  (((cyg_thread *)(_thread))->thread_data[memprot_data_index]))

//! Switch to USER mode (abandoning privileges)
#define HAL_ENTER_USER_MODE()		\
 asm volatile ( "mrs r0, cpsr;"		\
 		"bic r0, r0, #0x1f;"	\
 		"orr r0, r0, #0x10;"	\
 		"msr cpsr, r0;"		\
 		"nop" : : : "r0", "memory" );

/** Get current thread execution context/domain.
 *
 * This function return the current thread execution context. This context is
 * determined using the thread stack base domain, or the address parameter
 * domain if it is not NULL.
 *
 * @param addr
 * @return The thread current context/domain (0 <= domain <= 15 )
 */
int cyg_memprot_get_current_domain(cyg_uint32 addr) {
  int domain;
  if ( !addr ) {
    cyg_thread *self = (cyg_thread *)cyg_thread_self();
    addr = self->stack_base;
  }
  // derive current thread domain from address
  // ( this should also work for the idle thread )
  domain = memprot_mmu_get_domain((CYG_ADDRESS)addr);
  CYG_ASSERT( domain >= 0, "Invalid thread domain !\n");
  return domain;
}

//! Redirection of cyg_thread_exit()
void cyg_memprot_thread_exit(void) {
  cyg_thread_exit();
}

//! User mode assertion failure (displays a message and suspends thread)
void cyg_memprot_thread_fail(const char *file, int line, const char *msg) {
  fprintf( stderr, "%s: %d: ASSERT FAILED (", file, line );
  fputs( msg , stderr );
  fputs( ")\n", stderr );
  cyg_thread_suspend(cyg_thread_self());
}

extern void __swi_cyg_thread_exit(void);

/** Thread entry replacement.
 *
 * This function is used as replacement for the user-provided entry of a
 * thread. It performs a few privilege-related setups and calls the real entry
 * point.
 */
void cyg_memprot_thread_entry(cyg_addrword_t entry_data) {
  int domain;
  cyg_memprot_thread_data_t *data = USER_THREAD(cyg_thread_self());
  // find out execution domain
  domain = memprot_mmu_get_domain((CYG_ADDRESS)data);
  CYG_ASSERT(( domain > 0 ), "Invalid execution domain !\n" );
  // setup DACR
  memprot_mmu_set_DACR(USER_DACR(domain), CYGOPT_MEMPROT_DACR_MASK);
  // abandon system privileges
  HAL_ENTER_USER_MODE();
  // and enter real thread code
  data->entry(entry_data);
  // return to system mode and exit
  __swi_cyg_thread_exit();
}

/** Thread stack allocator.
 *
 * This function is used to allocate a thread stack in virtual memory. It uses
 * a basic slot based allocation. The allocated stack cannot exceed 60 kB, and
 * has a special property (used for memory protection of kernel data): its
 * lowest 1 kB is mapped read-only.
 *
 * @param domain The domain in which the stack should be allocated
 * @param stack_size The required stack size. This will not necessarily be the
 * allocated stack size, depending on page alignment and size requirements
 * @param stack_base The allocated stack base address
 * @param real_size The allocated stack size
 * @return '0' if allocation was successful, '-1' otherwise
 */
static int
cyg_memprot_allocate_stack( int domain, cyg_uint32 stack_size,
                            CYG_ADDRESS *stack_base, cyg_uint32 *real_size ) {
  static cyg_uint32 thread_counter = 0;
  cyg_uint32 slot, slot_size, nb_slots, i, required_size;
  CYG_ADDRESS base = 0;
  cygmem_vp_layout_t *layout = &cygmem_vp_layouts[domain];

  // maximum stack size is 64kB (actually 60 kB to ensure 4 kB mappings)
  slot_size = 16*MMU_PAGE_SIZE;
  nb_slots = ( layout->stacks_end - layout->stacks_start ) / slot_size;
  slot = thread_counter++;

  if ( thread_counter >= nb_slots )
    thread_counter = 0;
  for ( i = slot; i < slot+nb_slots; i++ ) {
    base = layout->stacks_start + slot_size*(i % nb_slots);
    if ( memprot_mmu_physical_address(base) == 0 )
      break;
  }
  if ( memprot_mmu_physical_address(base) != 0 ) {
    INFO("cyg_thread_create: too many threads (>= %d) in process !\n",
         nb_slots);
    return -1;
  }
  // ok we have an empty slot, try to allocate required size (at least 4 KB)
  required_size = ( stack_size )? ( stack_size + 0xfff ) & 0xfffff000 : 0x1000;
  if ( required_size > slot_size ) {
    INFO("cyg_thread_create: required stack is too big (%d bytes)\n",
         stack_size);
    return -1;
  }
  // map stack as cacheable, bufferable
  if ( memprot_mmu_map( base, (CYG_ADDRESS)0, required_size, AP_RW_RW, 1,
                        domain, MMU_ALLOCATE, 1 ) != ENOERR ) {
    INFO("cyg_thread_create: unable to allocate thread stack\n");
    return -1;
  }
  // map lower 1k read-only for kernel data
  if ( !memprot_mmu_set_subpage_ap( base, AP_RW_RO ) ) {
    INFO("cyg_thread_create: unable to protect thread stack\n");
    return -1;
  }
  *stack_base = base;
  *real_size = required_size;
  return 0;
}

//! Free an allocated stack
static
void cyg_memprot_free_stack( CYG_ADDRESS base, cyg_uint32 size ) {
  memprot_mmu_unmap( base, size, 1 );
}

/** Thread creation function replacement.
 *
 * This function is used a replacement for cyg_thread_create(). It performs
 * several checks before creating a user mode thread.
 */
void
cyg_memprot_thread_create( cyg_addrword_t      sched_info,
                           cyg_thread_entry_t  *entry,
                           cyg_addrword_t      entry_data,
                           char                *name,
                           void                *stack_base,
                           cyg_ucount32        stack_size,
                           cyg_handle_t        *handle,
                           cyg_thread          *thread ) {
  int pdomain, domain;
  cyg_memprot_thread_data_t *data;
  CYG_ADDRESS base, s_base;
  cyg_uint32 size;
  cyg_thread *mythread;

  domain = memprot_mmu_get_domain((CYG_ADDRESS)entry);
  pdomain = cyg_memprot_get_current_domain(0);
  if (( domain < 0 )||(( domain != pdomain )&&( pdomain > 0 ))) {
    // parent & child thread must be in the same domain
    // (except if we are currently in kernel context)
    INFO("cyg_thread_create: permission denied\n");
    return;
  }
  if ( domain == 0 ) {
    // kernel/privileged thread
    cyg_thread_create( sched_info, entry, entry_data,
                       name, stack_base, stack_size, handle, thread );
    return;
  }

  // allocate stack in virtual memory
  if ( cyg_memprot_allocate_stack( domain, stack_size, &base, &size ) < 0 )
    return;

  s_base = base;
  // allocate cyg_thread struct in lower 1k
  mythread = (cyg_thread *)base;
  base += sizeof(cyg_thread);
  // allocate thread-specific data in lower 1k
  data = (cyg_memprot_thread_data_t *)base;
  data->stack_base = s_base;
  data->stack_size = size;
  data->entry = entry;
  // create thread with a redirected entry point
  cyg_thread_create( sched_info, cyg_memprot_thread_entry, entry_data,
                     name, (void *)(s_base + 0x400), size-0x400, handle,
                     mythread );
  // hook user thread data to thread object
  mythread->thread_data[memprot_data_index] = (CYG_ADDRWORD)data;
}

/** Thread delete function replacement.
 *
 * This function is used a replacement for cyg_thread_delete(). It performs
 * stack deallocation if necessary.
 */
void
cyg_memprot_thread_delete(cyg_handle_t thread) {
  cyg_memprot_thread_data_t *data;
  int domain;
  cyg_thread_delete(thread);
  domain = memprot_mmu_get_domain((CYG_ADDRESS)
                                  ((cyg_thread *)thread)->stack_base);
  if ( domain > 0 ) {
    data = USER_THREAD(thread);
    if ( data->stack_base != 0 )
      cyg_memprot_free_stack( data->stack_base, data->stack_size );
  }
}

// POSIX pthread interface redefinitions

#if defined(CYGPKG_POSIX) && defined(CYGPKG_POSIX_PTHREAD)
#include <pthread.h>
// we also need a private header
#include "../../../compat/posix/current/src/pprivate.h"

typedef void *pthread_entry_t(void *);
extern void __swi_pthread_exit(void *);

/** POSIX thread entry replacement
 *
 * This function is used as replacement for the user-provided entry of a
 * thread. It performs a few privilege-related setups and calls the real entry
 * point.
 */
void *cyg_memprot_pthread_entry(void *arg) {
  int domain;
  void *retval;
  cyg_memprot_thread_data_t *data = USER_THREAD(cyg_thread_self());
  // find out execution domain
  domain = memprot_mmu_get_domain((CYG_ADDRESS)data);
  CYG_ASSERT(( domain > 0 ), "Invalid execution domain !\n" );
  // setup DACR
  memprot_mmu_set_DACR(USER_DACR(domain), CYGOPT_MEMPROT_DACR_MASK );
  // abandon system privileges
  HAL_ENTER_USER_MODE();
  // and enter real pthread code
  retval = ((pthread_entry_t *)data->entry)(arg);
  __swi_pthread_exit(retval);
  return NULL; // dummy return
  // when do we free the allocated stack ?
}

/** POSIX thread creation function replacement.
 *
 * This function is used a replacement for pthread_create(). It performs
 * several checks before creating a user mode POSIX thread. Note that the only
 * attribute parameter taken into account is stack size, all other
 * attributes are ignored (including stack base).
 */
int
cyg_memprot_pthread_create( pthread_t *thread,
                            const pthread_attr_t *attr,
                            void *(*start_routine)(void *),
                            void *arg ) {
  int ret, pdomain, domain;
  cyg_memprot_thread_data_t *data;
  size_t stack_size;
  pthread_attr_t nattr;
  struct sched_param schedparam;
  CYG_ADDRESS base;
  cyg_uint32 size;
  pthread_info *nthread;

  domain = memprot_mmu_get_domain((CYG_ADDRESS)start_routine);
  pdomain = cyg_memprot_get_current_domain(0);

  if (( domain < 0 )||(( domain != pdomain )&&( pdomain > 0 ))) {
    // parent & child thread must be in the same domain
    // (except if we are currently in kernel context)
    INFO("pthread_create: permission denied\n");
    return -1;
  }
  if ( domain == 0 )
    // kernel/privileged pthread
    return pthread_create( thread, attr, start_routine, arg );

  // allocate our own stack
  if (( attr == NULL )||( pthread_attr_getstacksize(attr, &stack_size) ))
    // stack size defaults to 4 kB
    stack_size = 0x1000;

  schedparam.sched_priority = CYGNUM_POSIX_MAIN_DEFAULT_PRIORITY;
  pthread_attr_init(&nattr);
  pthread_attr_setinheritsched( &nattr, PTHREAD_EXPLICIT_SCHED );
  pthread_attr_setschedpolicy( &nattr, SCHED_RR );
  pthread_attr_setschedparam( &nattr, &schedparam );

  if ( cyg_memprot_allocate_stack( domain, stack_size, &base, &size ) < 0 )
    return 0;

  // allocate thread-specific data in lower 1k
  data = (cyg_memprot_thread_data_t *)base;
  data->stack_base = base;
  data->stack_size = size;
  data->entry = (cyg_thread_entry_t *)start_routine;

  // reserve 256 bytes, which leaves 768 bytes for pthread stuff in RO area
  pthread_attr_setstacksize( &nattr, (size_t)(size - 0x100));
  pthread_attr_setstackaddr( &nattr, (void *)(base + size));

  // redirect thread entry point
  ret = pthread_create( thread, &nattr, cyg_memprot_pthread_entry, arg );

  // hook user thread data to thread object
  nthread = (pthread_info *)(base + 0x100);
  nthread->thread->thread_data[memprot_data_index] = (CYG_ADDRWORD)data;

  return ret;
}

#endif

// ============================================================================
// ParrotOS replacement calls
// ============================================================================

#ifdef CYGPKG_PARROTOS

/** This function is used to regain system privileges. */
void cyg_memprot_system(void) {
  CYG_ADDRESS _lr;
  cyg_memprot_thread_data_t *data = USER_THREAD(cyg_thread_self());
  // skip usual return from system call (and avoid loosing privileges)
  _lr = data->swi_stack[6];
  // this ugly hack relies on gcc calling conventions
  asm volatile ( "str %0,[fp,#-4]" : : "r"(_lr) );
}

extern void __swi_memprot_system(void);

#include <cyg/hal/drv_api.h>
#include <stdlib.h>

// include ParrotOS headers
#include <parrotos/parrotos.h>

// these are not public headers but we still need them !
#include "../../../parrotos/current/src/supervis.h"
#include "../../../parrotos/current/src/uart.h"

extern UART_INFO UART_infos[UARTS_COUNT];
extern const char cUART_Names[][10];

/** Global mutex used to synchronize Parrot4 threads. */
extern cyg_mutex_t sup_parrot4_mutex;

/** Table of threads. */
extern SUP_TASKINFO sup_tasks[SUP_NB_MOD];

/** Number of used slots in the task table */
extern int sup_tasks_count;

/** Semaphore to put messages in the queue. */
extern cyg_sem_t sup_msg_queue;


/** Header for Parrot4 tasks (memprot replacement).
 * This header is used to wrap a Parrot4 task into a regular eCos task.
 * All such tasks are synchronized using a single mutex.
 * It is possible to cleanly terminate the task by calling cyg_thread_release()
 * on the task.
 */
void
cyg_memprot_sup_task_header(cyg_addrword_t taskIdx) {
  int domain;
  SUP_TASKINFO * taskinfo = sup_tasks + taskIdx;
  SUP_MESSAGE * msg;

  CYG_ASSERT(taskIdx < SUP_NB_MOD, "Invalid task ID !");
  CYG_ASSERT(taskinfo->tinfo != NULL, "Null taskinfo->tinfo !" );

  while ( cyg_semaphore_wait(&taskinfo->tinfo->msgsem) ) {
    // Parrot4 compat: some P4 tasks are synchronized using a single mutex
    if ( taskinfo->mode == MODULE_MODE_P4_MUTEX )
      sup_p4mutex_lock();
    // Get the message
    cyg_drv_dsr_lock();
    if ( taskinfo->tinfo->msgCount ) {
      msg = &taskinfo->tinfo->msg[taskinfo->tinfo->msgRead++];
      taskinfo->tinfo->msgRead %= SUP_MAX_MES;
      taskinfo->tinfo->msgCount--;
    }
    else
      msg = NULL;
    cyg_drv_dsr_unlock();

    if (msg) {
      // if proc entry is null it is a container and we need to forward the
      // message to the right task
      if ( taskinfo->entry.proc == NULL ) {
        int i;
        // Handle message
        // deliver the message to the right task
        for ( i = 0; i < sup_tasks_count; i++ ) {
          if (( sup_tasks[i].container == taskinfo->taskid )&&
              ( sup_tasks[i].taskid == msg->task)) {

            // get container execution domain
            domain = cyg_memprot_get_current_domain(0);
            if ( domain > 0 ) {
              PROCEDURE proc = sup_tasks[i].entry.proc;
              // check procedure domain
              CYG_ASSERT( domain == memprot_mmu_get_domain((CYG_ADDRESS)proc),
                          "Container and module in different domains !" );
              // setup DACR
              memprot_mmu_set_DACR(USER_DACR(domain),CYGOPT_MEMPROT_DACR_MASK);
              // abandon system privileges
              HAL_ENTER_USER_MODE();
              // and execute procedure
              proc( msg->message, msg->param );
              // regain system privileges
              __swi_memprot_system();
            }
            else
              // regular procedure call
              (sup_tasks[i].entry.proc)(msg->message, msg->param);
            break;
          }
        }
      }
      else {
        // get current execution domain
        domain = cyg_memprot_get_current_domain(0);
        if ( domain > 0 ) {
          // setup DACR
          memprot_mmu_set_DACR(USER_DACR(domain), CYGOPT_MEMPROT_DACR_MASK);
          // abandon system privileges
          HAL_ENTER_USER_MODE();
          // and execute procedure
          taskinfo->entry.proc(msg->message, msg->param);
          // regain system privileges
          __swi_memprot_system();
        }
        else
          // Handle the message
          taskinfo->entry.proc(msg->message, msg->param);
      }

      //Release the message slot
      cyg_semaphore_post(&sup_msg_queue);
    }
    else {
      RT_POST_WARNING("No more message");
      RT_POST_INT( taskIdx );
    }
    //Release Parrot4 semaphore
    if (taskinfo->mode == MODULE_MODE_P4_MUTEX)
      sup_p4mutex_unlock();
  }
}

/** Thread creation function replacement.
 *
 * This function is used as a replacement for sup_start_module.
 */
Cyg_ErrNo cyg_memprot_sup_start_module_stack( PRIORITY_OR_MODULE pom,
                                        MODULE_ENTRY entry,
                                        eMODULE_MODE mode,
                                        MODULE* taskid,
                                        char* name,
										void* stack, 
										unsigned int stack_size) {
  int ret, pdomain, domain;
  cyg_memprot_thread_data_t *data;
  CYG_ADDRESS base, s_base;
  cyg_uint32 size;
  cyg_thread *mythread;
  SUP_TASKINFO * taskinfo;

  // find out exection domain
  if ( (CYG_ADDRESS)entry.thread )
    domain = memprot_mmu_get_domain((CYG_ADDRESS)entry.thread);
  else
    /* thread entry is NULL, so this is a container thread...
     * we have no way of knowing in advance which modules will be added to
     * this container, yet we must choose an execution domain now !
     * for now we get a default value from CDL configuration...
     */
    domain = CYGOPT_MEMPROT_PARROTOS_CONTAINER_DOMAIN;

  // permission checks
  pdomain = cyg_memprot_get_current_domain(0);
  if (( domain < 0 )||(( domain != pdomain )&&( pdomain > 0 ))) {
    // parent & child thread must be in the same domain
    // (except if we are currently in kernel context)
    INFO("cyg_memprot_sup_start_module(): permission denied\n");
    return -1;
  }

  if ( domain == 0 )
    // we want a kernel thread, use regular call
    return sup_start_module_stack( pom, entry, mode, taskid, name, stack, stack_size);

  // create a userland thread
  CYG_ASSERT( sup_tasks_count < SUP_NB_MOD, "Too many modules" );
  taskinfo = sup_tasks + sup_tasks_count;

  // Allocate memory ressources
  taskinfo->name = name;

  if ( mode == MODULE_MODE_SHARED ) {
    SUP_TASKINFO *cont_ti = &sup_tasks[*(pom.task)];
    // check domain consistency
    CYG_ASSERT ( domain == CYGOPT_MEMPROT_PARROTOS_CONTAINER_DOMAIN,
                 "Module and container are not in the same domain !" );
    diag_printf( "Starting module %s in thread \"%s\" [in domain %d]\n",
                 name, cont_ti->name, domain );
  }
  else {
    diag_printf( "Starting thread %d [in domain %d]: %s!\n",
                 sup_tasks_count, domain, name? name : "<no name>" );
  }

  taskinfo->taskid = *taskid = sup_tasks_count;
  taskinfo->entry = entry;
  taskinfo->mode = mode;
  taskinfo->tinfo = NULL;

  if ( mode != MODULE_MODE_SHARED ) {
    /* case where the task is a thread */
    void (*fun)(cyg_addrword_t) = cyg_memprot_sup_task_header;

    // allocate stack in virtual memory
    ret = cyg_memprot_allocate_stack( domain, SUP_STACKSIZE, &base, &size );
    CYG_ASSERT( !ret, "Out of memory in cyg_memprot_sup_start_module !" );
    s_base = base;

    // allocate thread-specific data in lower 1k
    data = (cyg_memprot_thread_data_t *)base;
    data->stack_base = s_base;
    data->stack_size = size;
    data->entry = entry.thread;

    taskinfo->prio = pom.priority;
    taskinfo->tinfo = malloc(sizeof(SUP_THREADED_TASK_INFO));
    CYG_ASSERT( taskinfo->tinfo,
                "Out of memory in cyg_memprot_sup_start_module !" );
    memset(taskinfo->tinfo, 0, sizeof(SUP_THREADED_TASK_INFO));
    // do not allocate cyg_thread struct in lower 1k, use parrotos taskinfo
    mythread = &taskinfo->tinfo->thread_obj;

    // Create mailbox
    cyg_semaphore_init(&taskinfo->tinfo->msgsem, 0);
    taskinfo->tinfo->msgCount = 0;
    taskinfo->tinfo->msgRead = 0;

    // Create the thread
    if ( mode == MODULE_MODE_P5 )
      fun = cyg_memprot_thread_entry;

    cyg_thread_create( taskinfo->prio, fun, (cyg_addrword_t)*taskid, name,
                       (void *)(s_base + 0x400), size-0x400,
                       &taskinfo->tinfo->thread, mythread );
    // hook user thread data to thread object
    mythread->thread_data[memprot_data_index] = (CYG_ADDRWORD)data;
    cyg_thread_resume(taskinfo->tinfo->thread);
  }
  else {
    /* add a task to a running thread */
    MODULE *pcont = pom.task;
    taskinfo->prio = 0; // not used but better clear it
    CYG_ASSERT( *pcont < SUP_NB_MOD, "Invalid container" );
    sup_p4mutex_lock();
    taskinfo->container = *pcont;
    sup_p4mutex_unlock();
  }

  sup_tasks_count++;
  return ENOERR;
}

/** sup_init_modules() replacement. */
Cyg_ErrNo cyg_memprot_sup_init_modules( const MODULE_STRUCT ms[],
                                        unsigned int nb_mod) {
  int i;
  Cyg_ErrNo res;
  for ( i = 0; i < nb_mod; i++ ) {
    res = cyg_memprot_sup_start_module_stack( ms[i].pom, ms[i].entry, ms[i].mode,
											  ms[i].taskid, ms[i].task_name, NULL, 0);
    if ( res != ENOERR )
      return res;
  }
  return ENOERR;
}

// UART API replacement

// Internal Function Prototypes
void UART_task(cyg_addrword_t uart_id);

/** ParrotOS UART callback common stub. */
void
cyg_memprot_uart_callback(int uart_id, U8 *buffer, unsigned int length ) {
  int domain;
  UART_CALLBACK callback;
  cyg_memprot_thread_data_t *data = USER_THREAD(cyg_thread_self());
  callback = (UART_CALLBACK)data->entry;
  // find out execution domain
  domain = memprot_mmu_get_domain((CYG_ADDRESS)data);
  CYG_ASSERT(( domain > 0 ), "Invalid execution domain !\n" );
  // setup DACR
  memprot_mmu_set_DACR(USER_DACR(domain), CYGOPT_MEMPROT_DACR_MASK);
  // abandon system privileges
  HAL_ENTER_USER_MODE();
  // and execute procedure
  callback(uart_id, buffer, length);
  // regain system privileges
  __swi_memprot_system();
}

/**
 * Start receiving data on a uart (memprot replacement).
 * This function is used to configure a
 * uart--previously opened with UART_Init--
 * so that a callback is called whenever some data is available.
 *
 * @param uard_id: UART number
 * @param priority: priority of the reception task
 * @param buffer: reception buffer
 * @param length: reception buffer length
 * @param callback: reception callback.
 */
void cyg_memprot_UART_Create_Task( int uart_id, int priority, U8 *buffer,
                                   unsigned int length,
                                   UART_CALLBACK callback ) {
  int ret, pdomain, domain;
  cyg_memprot_thread_data_t *data;
  CYG_ADDRESS base, s_base;
  cyg_uint32 size;
  cyg_thread *mythread;
  UART_INFO * info = UART_infos + uart_id;

  // permission checks
  domain = memprot_mmu_get_domain((CYG_ADDRESS)callback);
  pdomain = cyg_memprot_get_current_domain(0);
  if (( domain < 0 )||(( domain != pdomain )&&( pdomain > 0 ))) {
    // parent & child thread must be in the same domain
    // (except if we are currently in kernel context)
    INFO("cyg_memprot_UART_Create_Task(): permission denied\n");
    return;
  }

  if ( domain == 0 ) {
    // we want the callback to be called in kernel context, use regular call
    UART_Create_Task( uart_id, priority, buffer, length, callback );
    return;
  }

  CYG_ASSERT( uart_id >= 0 && uart_id < UARTS_COUNT, "Bad UART id");
  CYG_ASSERT( info->handle != 0, "Bad UART handle" );
  CYG_ASSERT( info->thread == 0, "Bad UART thread" );

  //Initialize variables
  info->buffer = buffer;
  info->length = length;
  info->callback = cyg_memprot_uart_callback;

  // allocate stack in virtual memory
  ret = cyg_memprot_allocate_stack( domain, SUP_STACKSIZE, &base, &size );
  CYG_ASSERT( !ret, "Out of memory in cyg_memprot_UART_Create_Task() !" );
  s_base = base;

  // allocate thread-specific data in lower 1k
  data = (cyg_memprot_thread_data_t *)base;
  data->stack_base = s_base;
  data->stack_size = size;
  data->entry = (cyg_thread_entry_t *)callback;

  // do not allocate cyg_thread struct in lower 1k, use info
  mythread = &info->thread_obj;

  // create thread
  cyg_thread_create( priority, UART_task, uart_id,(char *)cUART_Names[uart_id],
                     (void *)(s_base+0x400), size-0x400,
                     &info->thread, mythread );
  // hook user thread data to thread object
  mythread->thread_data[memprot_data_index] = (CYG_ADDRWORD)data;
}


#endif
