//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
*******************************************************************************
* @file  exception.c
* @brief Exception handlers
*
* Dynamo exception handlers.
*
* Copyright (C) 2007 Parrot S.A.
*
* @author     ivan.djelic@parrot.fr
* @date       2007-03-01
*******************************************************************************
*/

#include <cyg/kernel/kapi.h>
#include <cyg/infra/diag.h>
#include <cyg/infra/cyg_ass.h>
#include <cyg/infra/cyg_type.h>
#include <cyg/hal/hal_arch.h>

#include "common.h"

//! declaration of the swi_handler routine from swi_handler.S
cyg_VSR_t swi_handler;

/** The exception handler.
 *
 * This function is called in SYSTEM mode when an exception occurs.
 * It performs exception decoding and prints an adequate error message.
 * If the exception occured in
 * user mode, the current thread is terminated; otherwise the system is halted.
 *
 * @param data This parameter is simply ignored
 * @param exception A code indicating which type of exception occured
 * @param info A pointer to a frame with all saved pre-exception registers
 */
void
dynamo_exception_handler( cyg_addrword_t data,
                          cyg_code_t exception,
                          cyg_addrword_t info ) {

    HAL_SavedRegisters *frame = (HAL_SavedRegisters *)info;
    CYG_ADDRESS vaddr;
    char *msg;
    int mode, domain;

    switch ( exception ) {

    case CYGNUM_HAL_EXCEPTION_ILLEGAL_INSTRUCTION:
        DIAG("undefined instruction @ pc = 0x%08x\n", frame->pc );
        break;

    case CYGNUM_HAL_EXCEPTION_INTERRUPT:
        DIAG("software interrupt @ pc = 0x%08x\n", frame->pc );
        break;

    case CYGNUM_HAL_EXCEPTION_DATA_ACCESS:

        arm_mmu_get_data_abort_info( &vaddr, &domain, &msg );
        DIAG("data abort @ pc = 0x%08x (%s)\n", frame->pc-4, msg );

        if ( domain >= 0 ) {
            DIAG("[trying to access 0x%08x in domain %d]\n", vaddr, domain );
        }
        else {
            DIAG("[trying to access 0x%08x]\n", vaddr );
        }
        break;

    case CYGNUM_HAL_EXCEPTION_CODE_ACCESS:
        DIAG("prefetch abort @ pc = 0x%08x\n", frame->pc );
        break;

    default:
        DIAG("unknown exception (%d) @ pc = 0x%08x\n", exception, frame->pc );
        break;
    }

    // in which mode did the exception occur ?
    mode = frame->cpsr & CPSR_MODE_BITS;

    if ( mode == CPSR_USER_MODE ) {
        dynamo_thread_t *thread;
        cyg_thread * handle = (cyg_thread *) cyg_thread_self();

        // get extended thread object
        thread = (dynamo_thread_t *)handle;

        // switch to this thread kernel stack
        frame->sp = ((CYG_ADDRESS)(thread->data.kstack)) +
            CYGOPT_DYNAMO_KSTACK_SIZE;

        // switch the faulting thread to system mode before calling exit()
        frame->cpsr &= ~CPSR_MODE_BITS;
        frame->cpsr |= CPSR_SYSTEM_MODE;

        // force thread to call exit()
        frame->pc = (cyg_uint32)cyg_thread_exit;

        if ( exception != CYGNUM_HAL_EXCEPTION_CODE_ACCESS ) {
            // fake a jump to the forced pc value above
            // this makes debugging easier
            frame->lr = frame->pc;
        }
        DIAG( "exiting thread \"%s\" (handle = 0x%08x, domain = %d)\n",
              handle->name, (cyg_uint32)handle, thread->data.process->domain );

        // @todo send kill signal to all process threads
    }
    else {
        CYG_FAIL("\nException occured in a privileged mode ! Halting...");
    }
}

/** Exception handler installer.
 *
 * This function installs a custom exception handler and redirects SWI
 * interrupt to swi_handler() in swi_handler.S.
 */
void
dynamo_install_exception_handlers(void) {

    DEBUG1("installing exception handlers\n");

    cyg_interrupt_set_vsr( CYGNUM_HAL_VECTOR_SOFTWARE_INTERRUPT, swi_handler );
    cyg_exception_set_handler( CYGNUM_HAL_EXCEPTION_DATA_ACCESS,
                               dynamo_exception_handler, 0, NULL, NULL );
}
