//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
*******************************************************************************
* @file  ram_alloc.c
* @brief RAM memory allocation.
*
* Memory protection physical RAM allocator. This RAM allocator is designed
* primarily as an ARM MMU physical allocator. It is able to allocate
* contiguous chunks of memory of only 3 possible sizes : 1 MB (section), 64 KB
* (large page) and 4 KB (small page). All allocated chunks are aligned with
* respect to their size. Unlike traditional memory allocators, it is also able
* to partially free an allocated block, provided the requested range to free
* has one of the 3 valid sizes.
*
* Note: This allocator can handle only one bank of RAM memory. It is not
* thread-safe, as it is supposed to be used in a thread-safe context (in MMU
* routines).
*
* Copyright (C) 2004,2005 Parrot S.A.
*
* @author     ivan.djelic@parrot.fr
* @date       2004-06-26
*******************************************************************************
*/

#include <cyg/infra/cyg_type.h>
#include <cyg/infra/cyg_ass.h>
#include <string.h>

//! A page is 2^12 = 4096 bytes
#define PAGE_SIZE_BITS     (12)
#define PAGE_SIZE          (1 << PAGE_SIZE_BITS)
#define LARGE_PAGE_SIZE    (PAGE_SIZE << 4)
#define SECTION_SIZE       (PAGE_SIZE << 8)

// Page information is stored in a 32-bit word

#define PINFO_NEXT_MASK    0xFFFFF000   //!< Next chunk address mask
#define PINFO_CHUNK        0x00000800   //!< Set if this is a valid chunk
#define PINFO_ALLOCATED    0x00000400   //!< Set if chunk is allocated
#define PINFO_SIZE_MASK    0x0000000F   //!< Chunk size mask
#define PINFO_SECTION      0x00000002   //!< Section (1 MB = 256 pages)
#define PINFO_LARGE_PAGE   0x00000001   //!< Large page (64 KB = 16 pages)
#define PINFO_SMALL_PAGE   0x00000000   //!< Small page (4 KB = 1 page)

#define IN_RAM(_addr)      (((_addr) >= ram_base )&&((_addr) < ram_top ))
#define PINFO(_addr)       table[((_addr)-ram_base) >> PAGE_SIZE_BITS ]
#define PSIZE(_log)        ( PAGE_SIZE << ((_log) << 2))
#define IS_ALIGNED(_P,_S)  (((_P)&((_S)-1)) == 0 )
#define ALIGNED(_P,_S)     ((_P)&~((_S)-1))

static unsigned int *table;
static CYG_ADDRESS ram_base;
static CYG_ADDRESS ram_top;
static CYG_ADDRESS free_chunk[3];
static cyg_uint32 ram_used;

/** Allocator initialisation.
 *
 * This function is called to setup the RAM allocator.
 *
 * @param base Physical RAM base address must be page-aligned
 * @param table A array of size @ref nb_pages
 * @param nb_pages The number of available physical pages (1 page = 4 kB)
 * @return 0 upon success, -1 otherwise
 */
int
ram_init( CYG_ADDRESS base, unsigned int *ptable, int nb_pages ) {
    int i, ret = -1;
    unsigned int size;
    CYG_ADDRESS addr;

    ram_base = ALIGNED( base, PAGE_SIZE );
    ram_top = ram_base + nb_pages* PAGE_SIZE;
    table = ptable;

    if (( ram_base == base )&&
        ( nb_pages > 0 )) {

        // clear page information
        memset( table, 0, sizeof(unsigned int)*nb_pages );

        for ( i = 0; i < 3; i++ ) {
            free_chunk[i] = 0;
        }
        // build up free chunk lists
        size = ram_top - ram_base;
        addr = ram_base;

        while ( size > 0 ) {
            int psize, registered = 0;
            // try all chunk sizes, starting with the largest
            for ( i = 2; i >= 0; i-- ) {
                psize = PSIZE(i);
                if (( size >= psize )&&( IS_ALIGNED( addr, psize ))) {
                    PINFO(addr) = (free_chunk[i] & PINFO_NEXT_MASK)|
                        PINFO_CHUNK|i;
                    free_chunk[i] = addr;
                    size -= psize;
                    addr += psize;
                    registered = 1;
                    break;
                }
            }
            CYG_ASSERT( registered, "Cannot register page !" );
        }
        ret = 0;
        ram_used = 0;
    }
    return ret;
}

/** RAM Allocator.
 *
 * This allocator only accepts 3 possible sizes: 4 KB, 64 KB or 1 MB. It
 * guarantees that the returned address is aligned w.r.t the required size.
 *
 * @param size The required size: 4 KB, 64 KB or 1 MB
 * @return An aligned address if allocation was successful, 0 otherwise.
 */
CYG_ADDRESS
ram_allocate(size_t size) {
    CYG_ADDRESS addr;
    int i, log;
    unsigned int psize;

    // check size
    CYG_ASSERT(( size == SECTION_SIZE )||
               ( size == LARGE_PAGE_SIZE )||
               ( size == PAGE_SIZE ), "Bad allocation size !" );

    // try to find a free chunk, the smaller the better
    log = -1;
    for ( i = 0; i < 3; i++ ) {
        if (( size <= PSIZE(i) )&&( free_chunk[i] )) {
            log = i;
            break;
        }
    }
    if ( log < 0 ) {
        // no chunks of this size
        return 0;
    }
    // split chunk until we get the right size
    while ( PSIZE(log) > size ) {
        addr = free_chunk[log];
        // remove big chunk from free list
        free_chunk[log] = PINFO(addr) & PINFO_NEXT_MASK;
        log--;
        psize = PSIZE(log);
        // and add 16 new small chunks
        for ( i = 0; i < 16; i++ ) {
            PINFO(addr + psize*i) = free_chunk[log]|PINFO_CHUNK|log;
            free_chunk[log] = addr + psize*i;
        }
    }
    // remove chunk from free list
    addr = free_chunk[log];
    CYG_ASSERT( addr, "Null chunk address !" );

    PINFO(addr) |= PINFO_ALLOCATED;
    free_chunk[log] = PINFO(addr) & PINFO_NEXT_MASK;

    ram_used += size;
    return addr;
}

/** Coalesce buddy chunks.
 *
 * This internal function is used to merge free chunks into bigger chunks.
 */
static void
ram_coalesce(CYG_ADDRESS addr) {
    int i, log;
    unsigned int psize, bsize, info;
    CYG_ADDRESS naddr, paddr, caddr;

    log = PINFO(addr) & PINFO_SIZE_MASK;
    psize = PSIZE(log);
    bsize = psize << 4;
    naddr = ALIGNED( addr, bsize );

    // do not try to coalesce sections
    if (( log >= 2 )|| !IN_RAM(naddr) || !IN_RAM(naddr+bsize-1) ) {
        return;
    }
    // check that all buddy chunks are free
    for ( i = 0; i < 16; i++ ) {
        info = PINFO(naddr + i*psize);
        CYG_ASSERT( info & PINFO_CHUNK, "Invalid chunk !" );
        if (( info & PINFO_ALLOCATED )||(( info & PINFO_SIZE_MASK ) != log)) {
            return;
        }
    }
    // merge all chunks
    for ( i = 0; i < 16; i++ ) {
        paddr = 0;
        caddr = naddr + i*psize;
        // remove chunk from free list
        addr = free_chunk[log];
        while ( addr && ( caddr != addr )) {
            paddr = addr;
            addr = PINFO(addr) & PINFO_NEXT_MASK;
        }
        CYG_ASSERT( addr, "Free chunk missing in free list !" );
        if ( paddr ) {
            PINFO(paddr) &= ~PINFO_NEXT_MASK;
            PINFO(paddr) |= ( PINFO(caddr) & PINFO_NEXT_MASK );
        }
        else {
            free_chunk[log] = PINFO(caddr) & PINFO_NEXT_MASK;
        }
        PINFO(caddr) = 0;
    }
    // add merged chunk to free list
    PINFO(naddr) = free_chunk[log+1]|PINFO_CHUNK|(log+1);
    free_chunk[log+1] = naddr;
    // try to coalesce this chunk
    ram_coalesce(naddr);
}

/** RAM deallocator.
 *
 * @param addr The chunk address to free
 * @param size If this parameter is not zero, the deallocator will try to free
 * only 'size' bytes, instead of the size of the chunk that was allocated at
 * 'addr'.
 */
void
ram_free( CYG_ADDRESS addr, size_t size ) {
    int i, log;
    unsigned int psize;
    CYG_ADDRESS naddr;

    CYG_ASSERT( IN_RAM(addr), "Trying to free address out of ram !" );
    CYG_ASSERT( IS_ALIGNED( addr, PAGE_SIZE ), "Unaligned free address !" );

    log = -1;
    // find smallest chunk containing addr
    for ( i = 0; i < 3; i++ ) {
        naddr = ALIGNED( addr, PSIZE(i));
        if ( IN_RAM(naddr) &&( PINFO(naddr) & PINFO_CHUNK )) {
            log = PINFO(naddr) & PINFO_SIZE_MASK;
            break;
        }
    }
    psize = PSIZE(log);

    CYG_ASSERT( log >= 0, "Cannot find container chunk !" );
    CYG_ASSERT( PINFO(naddr) &PINFO_ALLOCATED, "Freeing unallocated memory !");

    size = ( size )? size : psize;

    CYG_ASSERT(( size == PAGE_SIZE )||
               ( size == LARGE_PAGE_SIZE )||
               ( size == SECTION_SIZE ), "Invalid free size !" );
    CYG_ASSERT( naddr+psize >= addr+size, "Free size is too large !");

    if ( psize > size ) {
        psize = PSIZE(log-1);
        // partial deallocation : split this chunk into smaller ones
        for ( i = 0; i < 16; i++ ) {
            PINFO(naddr + psize*i) = PINFO_CHUNK|PINFO_ALLOCATED|(log-1);
        }
        // try again
        ram_free( addr, size );
    }
    else {
        // free chunk
        PINFO(addr) = free_chunk[log]|PINFO_CHUNK|log;
        free_chunk[log] = addr;

        ram_used -= psize;
        // coalesce chunk
        ram_coalesce(addr);
    }
}

//! Get used RAM
cyg_uint32
ram_get_used_mem(void) {
    return ram_used;
}
