/**
*******************************************************************************
* @file  memprof.h
* @brief A simple memory profiling package for eCos
*
* Copyright (C) 2007 Parrot S.A.
*
* @author     ivan.djelic@parrot.com
* @date       2007-11-19
*******************************************************************************
*/
#ifndef CYGONCE_MEMPROF_H
#define CYGONCE_MEMPROF_H

#include <cyg/infra/cyg_type.h>

#define MEMPROF_TAG_SAMPLES (8)

struct memprof_tag {
    cyg_uint32         magic;
    const char         *func;
    cyg_uint32         line;
    cyg_uint32         calls;
    cyg_uint32         calls_total;
    cyg_uint32         alloc;
    cyg_uint64         alloc_total;
    cyg_uint32         sample_index;
    cyg_uint32         sample_size;
    cyg_uint32         sample_count;
    cyg_uint32         sample[MEMPROF_TAG_SAMPLES];
    struct memprof_tag *next;
};

#ifdef __cplusplus
extern "C" {
#endif

extern void memprof_register(struct memprof_tag *tag, const char *file, int l);
extern void *memprof_alloc(struct memprof_tag *tag, size_t size, int zero);
extern void *memprof_realloc(struct memprof_tag *tag, void *ptr, size_t size);
extern void free(void *ptr);
extern void memprof_report(void);

#ifdef __cplusplus
} /* extern "C" */
#endif 

#define MEMPROF_ALLOC(_size, _zero) ({                      \
            static struct memprof_tag _tag = {0};           \
            memprof_register(&_tag, __func__, __LINE__);    \
            memprof_alloc(&_tag, _size, _zero);             \
        })

#define MEMPROF_REALLOC(_ptr, _size) ({                     \
            static struct memprof_tag _tag = {0};           \
            memprof_register(&_tag, __func__, __LINE__);    \
            memprof_realloc(&_tag, _ptr, _size);            \
        })

#define malloc(_sz)                 MEMPROF_ALLOC(_sz, 0)
#define calloc(_n, _sz)             MEMPROF_ALLOC((_n)*(_sz), 1)
#define realloc(_ptr, _sz)          MEMPROF_REALLOC(_ptr,_sz)

/*! \mainpage memprof


<center><h2>memprof: profiling simple de la mmoire</h2></center>

<p>
<div align="center">
<hr>
\ref intro_sec &nbsp;|&nbsp;
\ref sec_step1 &nbsp;|&nbsp;
\ref sec_step2 &nbsp;|&nbsp;
\ref sec_step3
<hr>
</div>

\section intro_sec Introduction

Le package CYGPKG_PROFILE_MEMPROF collecte des informations sur l'allocation
dynamique de mmoire. Il permet notamment de dtecter facilement
certaines fuites mmoires, dbordements de buffer, etc. Il permet galement
de connatre prcisment la quantit de mmoire consomme par le systme au
cours du temps.

La mthode employe consiste  redfinir les appels standards
<tt>malloc()</tt>, <tt>free()</tt>, <tt>calloc()</tt> et <tt>realloc()</tt>.
En particulier, les fonctions <tt>malloc()</tt> et <tt>calloc()</tt> sont
remplaces par des macros, ce qui peut poser des problmes de compilation pour
un code dans lequel on trouve:

<ul>
<li> une dclaration des prototypes de ces fonctions (autre que l'inclusion
de <tt>stdlib.h</tt>).
<li> une rfrence explicite  l'adresse de ces fonctions.
</ul>

\section sec_step1 Installation

Pour installer memprof, il vous suffit d'ajouter le package
CYGPKG_PROFILE_MEMPROF, en acceptant les rsolutions de conflit proposes par
l'outil <b>configtool</b>.

\section sec_step2 Utilisation

Ds lors que le package est ajout, il est automatiquement activ. il n'y a
aucune initialisation  effectuer. Le package instrumente les allocations/
librations de mmoire et effectue les oprations suivantes:

<ul>
<li>Mise en place de marqueurs avant et aprs chaque buffer allou, et
vrification de l'intgrit des marqueurs lors de la libration (dtection
de dbordement).

<li>Dtection des doubles librations (appels conscutifs  <tt>free()</tt>
avec la mme adresse) et des librations de pointeurs invalides.

<li>Collecte de statistiques sur les nombres d'allocations et librations
effectus  chaque emplacement du code source.
</ul>

Lors de la dtection d'une corruption ou d'une opration invalide, la thread
courante affiche un message de diagnostic et entre dans une boucle infinie
pour permettre le debug.
<p>
Pour exploiter les informations collectes en temps rel, l'API disponible est
la suivante:

\code
void memprof_report(void);
\endcode

Cette fonction affiche sur la sortie diagnostic un rapport sous la
forme suivante:

<p>
<div><img alt="memprof report" src="memprof.png"></div>

La premire ligne donne l'tat courant de l'allocateur mmoire:

<ul>
<li><b>total</b>: Taille total du heap.
<li><b>free</b>: Quantit de mmoire disponible pour allocation.
<li><b>alloc</b>: Quantit de mmoire couramment alloue.
<li><b>peak</b>: Quantit de mmoire maximale alloue depuis le dmarrage du
systme.
<li><b>ovrhd</b>: La diffrence <b>total-(free+alloc)</b>,
i.e. quantit de mmoire gaspille (overhead) par l'allocateur et le package
memprof. A noter que le package memprof consomme 10 octets d'overhead par
allocation.
</ul>

Les lignes suivantes dtaillent, pour chaque allocation prsente dans le
code source, le nombre total d'allocations effectues, et le nombre
d'allocations non libres:

<table border=0>
<tr>
<td>
 <table border=0>

    <tr>
    <td valign=top align=right width="25%">
    <span style="background: #b1d185">Function name():line:</span>
    <td>Ce champ dcrit l'emplacement dans le code source d'un appel 
    <tt>malloc()</tt> ou <tt>calloc()</tt>.

    <tr>
    <td valign=top align=right width="25%">
    <span style="background: #e192a8">alloced (%):</span>
    <td>Quantit de mmoire courramment alloue.

    <tr>
    <td valign=top align=right width="25%">
    <span style="background: #b3bee4">avg sz:</span>
    <td>Taille moyenne des allocations en octets.

    <tr>
    <td valign=top align=right width="25%">
    <span style="background: #e4dc56">unfree/total:</span>
    <td>Ce champ dcrit le nombre d'allocations effectues depuis le
    dmarrage du systme (<tt>total</tt>), ainsi que le nombre d'allocations
    non libres (<tt>unfree</tt>). Le champ <tt>unfree</tt> est donc compris
    entre 0 et la valeur du champ <tt>total</tt>.

    <tr>
    <td valign=top align=right width="25%"><span style="background: #b3e4d9">
    Unfree minima history:</span>
    <td>Historique des valeurs minimales du champ <tt>unfree</tt>.

</table> 

\section sec_step3 Exploitation des rsultats

\subsection sec_step3_1 Historique des valeurs minimales du nombre d'allocations

L'historique propos dans le rapport permet d'avoir rapidement une ide du
comportement de l'allocation.
Si l'on reprsente le nombre d'allocations non libres (le champ
<tt>unfree</tt>) en fonction d'une
pseudo-variable <tt>t</tt> incrmente  chaque appel  <tt>malloc()</tt> et
<tt>free()</tt>, on obtient une courbe comme celle reprsente ci-dessous:

<p>
<div><img alt="unfree graph" src="unfree.png"></div>

Pour rechercher d'ventuelles fuites mmoires, il faut s'intresser 
l'volution des valeurs minimales sur cette courbe. Dans cet exemple, on
peut constater qu'il existe des cycles d'allocations suivies de librations,
avec un retour  une valeur nominale qui semble croissante; si la valeur
nominale attendue tait 0 ou une constante, la courbe montre probablement la
prsence d'une fuite mmoire.
<p>
La colonne <tt>Unfree minima history</tt> du rapport
liste une suite d'chantillons (entre 1 et 8), qui correspondent aux valeurs
de la courbe verte de la figure ci-dessus. Les chantillons sont les
valeurs minimales du champ <tt>unfree</tt>, calcules sur des intervalles de
<tt>t</tt> de longueur gale (sauf pour le dernier chantillon).
Cette longueur d'intervalle crot au cours du temps en prenant les valeurs
suivantes: 1, puis 2, puis 4, 8, 16, etc.

\subsection sec_step3_3 Dtection des fuites mmoires (memory leaks)

Au fur et  mesure que le temps passe, l'historique prsente des minima
d'intervalles de plus en plus longs. Quand la longueur de l'intervalle devient
suffisante pour couvrir un cycle complet d'allocation (dans l'hypothse ou
l'allocation est cyclique), alors les chantillons prennent normalement la
valeur nominale attendue.
<p>
Pour rsumer, il faut retenir 2 points:

<ul>
<li>Plus le nombre total d'allocations est important, plus l'historique est
pertinent.
<li>Une suite strictement croissante dans l'historique indique une possible
fuite mmoire, surtout si l'allocation prsente un caractre cyclique.
</ul>

Pour dtecter les fuites mmoires, l'idal est de produire des rapports
sur des dures d'excution assez longues, et en des points o l'application
est cense avoir termin un cycle d'allocation.

\subsection sec_step3_2 Utilisation avec RTmon

TODO.

*/

#endif /* CYGONCE_MEMPROF_H */
