//==========================================================================
// ####ECOSGPLCOPYRIGHTBEGIN####                                            
// -------------------------------------------                              
// This file is part of eCos, the Embedded Configurable Operating System.   
//
// eCos is free software; you can redistribute it and/or modify it under    
// the terms of the GNU General Public License as published by the Free     
// Software Foundation; either version 2 or (at your option) any later      
// version.                                                                 
//
// eCos is distributed in the hope that it will be useful, but WITHOUT      
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or    
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License    
// for more details.                                                        
//
// You should have received a copy of the GNU General Public License        
// along with eCos; if not, write to the Free Software Foundation, Inc.,    
// 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.            
//
// As a special exception, if other files instantiate templates or use      
// macros or inline functions from this file, or you compile this file      
// and link it with other works to produce a work based on this file,       
// this file does not by itself cause the resulting work to be covered by   
// the GNU General Public License. However the source code for this file    
// must still be made available in accordance with section (3) of the GNU   
// General Public License v2.                                               
//
// This exception does not invalidate any other reasons why a work based    
// on this file might be covered by the GNU General Public License.         
// -------------------------------------------                              
// ####ECOSGPLCOPYRIGHTEND####                                              
//==========================================================================

/**
*******************************************************************************
* @file  rtd2lxt.c
* @brief Real-Time Monitor host tools
*
* This utility is used to produce LXT dumpfiles (viewable with GTKwave) from
* RTmon dump (RTD) files.emit_clock
*
* Usage: rtd2lxt [-v] [-e <exefile>] <rtdfile> [<lxtfile> <savefile>]
*
* Copyright (C) 2006 Parrot S.A.
*
* @author     ivan.djelic@parrot.com
* @date       2006-12-08
*******************************************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <errno.h>
#include <assert.h>
#include <stdarg.h>

#define _GNU_SOURCE
#include <getopt.h>
#include <search.h>

#include "lxt_write.h"
#include "ecos_types.h"
#include "rtmon_common.h"
#include "posix_crc.h"
#include "rtd2lxt.h"

#define PFX             "rtd2lxt: "

/* Exported variables */
char            thd_tag[THREAD_ID_MAX][RTMON_TAGSIZE];
event_tag_t     user_event_tag[RTMON_PKT_EVENT_MAX];
event_tag_t     user_u32_tag[RTMON_PKT_USER_U32_MAX];

int             verbose               = 0;
int             process_thread_states = 0;
unsigned int    start_time  = 0;    //unit is 100 ns max value is about 5 days
unsigned int    end_time    = 0;    //unit is 100 ns max value is about 5 days
uint32_t        first_tick  = 0;    //unit is 10 ms
uint32_t        *buf        = NULL;
uint32_t        buf_index   = 0;
uint32_t        buf_size    = 0;

timer_unit_t    timer;    //<! current timer value

struct lt_trace *lt = NULL;
struct rtmon_header hdr;

const char      *exefile    = NULL;
FILE            *prof_file;



/* static variables */
static const int overhead_corr = 0;


/** Initialize profile report
 */
static FILE* profile_init(const char * name)
{
    FILE *file;

    file = fopen( name, "wb" );
    if ( file == NULL ) 
    {
        fprintf( stderr, PFX "FAILED\n");
    }
    else
    {
        fprintf( file, "positions: instr line\n" );
        fprintf( file, "events:  ticks\n" );
        fprintf( file, "ob=%s\n\n", exefile );
    }
    return file;
}


/** Log Profiling stop
 */
static void profile_stop( FILE *pfile )
{
    fclose(pfile);
}



static void enqueue_addr( uint32_t addr, int flush ) 
{
    static char cmdbuf[2*LINEBUF_MAX + ADDR2LINE_MAX*11];
    static char funcname[LINEBUF_MAX];
    static char basename[LINEBUF_MAX];
    static char tag[LINEBUF_MAX];

    static int index = 0;
    static uint32_t addrbuf[ADDR2LINE_MAX];

    int i;
    char hex32buf[12];
    uint32_t key[2];
    FILE *fp;

    ENTRY item, *rentry;

    if ( !flush ) {
        // check that this address is not already queued
        for ( i = 0; i < index; i++ ) {
            if ( addrbuf[i] == addr ) {
                // already in
                return;
            }
        }
        addrbuf[index++] = addr;
    }

    if (( index >= ADDR2LINE_MAX )||( flush && index )) {

        // prepare command string
        snprintf( cmdbuf, LINEBUF_MAX, "arm-elf-addr2line -C -f -s --exe=%s ",
                  exefile );
        // pass arguments
        assert( index <= ADDR2LINE_MAX );

        for ( i = 0; i < index; i++ ) {
            snprintf( hex32buf, 12, "0x%08x ", (unsigned int)addrbuf[i] );
            strncat( cmdbuf, hex32buf, 11 );
        }

        // call addr2line now
        fp = popen( cmdbuf, "r" );
        if ( fp ) {

            for ( i = 0; i < index; i++ ) {
                char *ret = NULL;

                if ( fgets( funcname, LINEBUF_MAX, fp ) &&
                     fgets( basename, LINEBUF_MAX, fp )) {
                    char *str;
                    // remove newline characters
                    str = strchr( funcname, '\n' );
                    if ( str ) {
                        *str = '\0';
                    }
                    str = strchr( basename, '\n' );
                    if ( str ) {
                        *str = '\0';
                    }
                    snprintf( tag, LINEBUF_MAX, "%s() [%s]", funcname,
                              basename );
                    ret = strdup(tag);
                }

                if ( !ret ) {
                    fprintf( stderr, PFX "no symbol for addr 0x%08x: %s\n",
                             (unsigned int)addrbuf[i], strerror(errno));
                }
                else {
                    //  add entry into hash table
                    key[0] = addrbuf[i];
                    key[1] = 0;
                    item.key = strdup((char *)key);
                    item.data = (void *)ret;
                    rentry = hsearch( item, ENTER );
                    assert(rentry);
                }
            }
            pclose(fp);
        }
        index = 0;
    }
}



/** Get a symbol name from an address
 */
char *get_srctag( uint32_t addr ) 
{
    char *ret = NULL;
    uint32_t key[2];
    ENTRY item, *fitem;

    if ( exefile ) {
        key[0] = addr;
        key[1] = 0;
        // lookup address info in hash table
        item.key = (char *)key;
        fitem = hsearch( item, FIND );
        if ( fitem ) {
            ret = (char *)fitem->data;
        }
    }
    return ret;
}

/** Load dump file into memory and check its integrity
 */
static int load_dumpfile(const char *name) 
{
    int     err     = 0;
    size_t  rcount  = 0;
    FILE    *fp;

    buf        = NULL;

    fprintf( stderr, PFX "loading dump file '%s'...\n", name );

    fp = fopen( name, "rb" );
    if ( !fp ) {
        fprintf( stderr, PFX "error: %s\n", strerror(errno));
        err = -1;
    }

    if ( !err ) 
    {
        rcount = fread( &hdr, sizeof(struct rtmon_header), 1, fp );
        if ( rcount != 1 ) 
        {
            fprintf( stderr, PFX "could not read header\n" );
            if ( errno ) 
            {
                fprintf( stderr, PFX "error: %s\n", strerror(errno));
            }
            err = -1;
        }
    }

    if ( verbose ) 
    {
        fprintf( stderr, PFX "HEADER:\n");
        fprintf( stderr, PFX "\tData buffer size     = %d bytes\n",(int)hdr.size);
        fprintf( stderr, PFX "\tBuffer crc           = %08x\n", (unsigned int)hdr.crc32);
        fprintf( stderr, PFX "\tRTMon dump version   = %d\n",(int)hdr.version);
        fprintf( stderr, PFX "\ttickperiod           = %d\n",(int)hdr.tickperiod);
        fprintf( stderr, PFX "\ttimefact_mult        = %d\n",(int)hdr.timefact_mult);
        fprintf( stderr, PFX "\ttimefact_div         = %d\n",(int)hdr.timefact_div);
        fprintf( stderr, PFX "\ttimescale            = %d\n",(int)hdr.timescale);
        fprintf( stderr, PFX "\tNb thread labels     = %d\n",(int)hdr.nb_thread);
        fprintf( stderr, PFX "\tNb user event labels = %d\n",(int)hdr.nb_user_event_tag);
        fprintf( stderr, PFX "\tNb user u32 labels   = %d\n",(int)hdr.nb_user_u32_tag);


         /*if ( verbose ){
         int i_irq;
         for(i_irq=0; i_irq<32; i_irq++)
            fprintf( stderr, PFX "Irq(%02d): %s\n",  i_irq, hdr.irq_tag[i_irq] );
         }*/
    }

    if (( !err )&&( hdr.magic != RTMON_MAGIC )) 
    {
        fprintf( stderr, PFX "this does not look like an RTmon dump\n" );
        err = -1;
    }

    if (( !err )&&( hdr.version != RTMON_SUPPORTED_VERSION )) 
    {
        fprintf( stderr,
           PFX "SORRY, dump version is v.%d but this tool only support v.%d \n"
           PFX "please upgrade rtd2lxt and getrtd !\n", 
           (int)(hdr.version), RTMON_SUPPORTED_VERSION );
        err = -1;
    }

   /*--------------------------------------------------------------------------
    * load thread tag buffer
    *------------------------------------------------------------------------*/
   if ( !err )
   {
      rcount = fread( thd_tag, 1, hdr.nb_thread*RTMON_TAGSIZE, fp );
      if ( rcount != (hdr.nb_thread*RTMON_TAGSIZE) ) 
      {
         fprintf( stderr, PFX "could not read thread tag buffer (%d threads)\n",
                  (int)(hdr.nb_thread));
         if ( errno ) {
            fprintf( stderr, PFX "error: %s\n", strerror(errno));
            err = -1;
         }
      }
    }

   /*--------------------------------------------------------------------------
    * load user event tag buffer
    *------------------------------------------------------------------------*/
   if ( !err )
   {
      rcount = fread( user_event_tag, 1, sizeof(event_tag_t)*hdr.nb_user_event_tag, fp );
      if ( rcount != (sizeof(event_tag_t)*hdr.nb_user_event_tag) ) 
      {
         fprintf( stderr, PFX "could not read user event tag buffer (%d events)\n",
                     (int)(hdr.nb_user_event_tag));
         if ( errno ) {
            fprintf( stderr, PFX "error: %s\n", strerror(errno));
            err = -1;
         }
      }
   }

   /*--------------------------------------------------------------------------
    * load user u32 tag buffer
    *------------------------------------------------------------------------*/
   if ( !err )
   {
      rcount = fread( user_u32_tag, 1, sizeof(event_tag_t)*hdr.nb_user_u32_tag, fp );
      if ( rcount != (sizeof(event_tag_t)*hdr.nb_user_u32_tag) ) 
      {
         fprintf( stderr, PFX "could not read user u32 tag buffer (%d u32)\n",
                  (int)(hdr.nb_user_u32_tag));
         if ( errno ) {
            fprintf( stderr, PFX "error: %s\n", strerror(errno));
            err = -1;
         }
      }
    }


    // load buffer
    if ( !err ) {
        buf = (uint32_t *)malloc(hdr.size);
        assert(buf);
        buf_size = hdr.size/sizeof(cyg_uint32);
        rcount = fread( buf, 1, hdr.size, fp );
        if ( rcount != hdr.size ) {
            fprintf( stderr, PFX "Read only (%d bytes) could not read buffer (%d bytes)\n",
                     (int)(rcount), (int)(hdr.size));
            if ( errno ) {
                fprintf( stderr, PFX "error: %s\n", strerror(errno));
            }
            err = -1;
        }
    }

    // check data integrity
    if ( !err ) 
    {
        uint32_t crc = posix_crc( (uint8_t *)buf, hdr.size );

        if ( crc != hdr.crc32 ) 
       {
            fprintf( stderr, PFX "bad CRC (0x%08x, should be 0x%08x) !\n",
                     (unsigned int)crc, (unsigned int)hdr.crc32 );
            fprintf( stderr, PFX "buffer first words \n" );
            fprintf( stderr, PFX "\n 0x%08x 0x%08x 0x%08x 0x%08x\n",
                     (unsigned int)buf[0], (unsigned int)buf[1],
                     (unsigned int)buf[2], (unsigned int)buf[3] );
            err = -1;
        }
    }

    // clean-up
    if ( fp ) {
        fclose(fp);
    }

    if ( err && buf ) {
        free(buf);
        buf = NULL;
    }

    return err;
}

/** Free dump buffer
 *
 */
static void close_dumpfile(void) 
{
    if ( buf ) 
    {
        free(buf);
        buf = NULL;
    }
}





//! Process packets with a code address and build hash table of tags
void process_addr_tags(void) 
{
    uint32_t func, type, step, next_update;
    int i;

    assert(exefile);

    fprintf( stderr, PFX "converting addresses to symbols : [00%%]" );

    step = buf_size/100;
    if ( !step ) {
        step = 1;
    }
    next_update = 0;

    for ( i = 0; i < buf_size; i++ ) 
    {
        type = RTMON_PKT_TYPE(buf[i]);

        if (   ((( type >= RTMON_PKT_SLOCK_0 )&&( type < RTMON_PKT_MAX ))
               ||(type == RTMON_PKT_CALLER)
               ||(type == RTMON_PKT_PROFILE)
               ||(type == RTMON_PKT_PC)
               ||(type == RTMON_PKT_AAI_PC)       )&&
             ( i+1 < buf_size )&&
             ( RTMON_PKT_TYPE(buf[i+1]) == RTMON_PKT_DATA24 )) 
        {
            // retrieve caller address
            func = (RTMON_PKT_DATA(buf[i]) << 24)|RTMON_PKT_DATA24(buf[i+1]);

            // enqueue address for processing if it is unknown
            if ( !get_srctag(func) ) 
            {
                enqueue_addr( func, 0 );
            }
            i++;
            if ( i >= next_update ) 
            {
                fprintf( stderr, "\b\b\b\b\b[%02d%%]", (int)(i*100/buf_size));
                next_update = i+step;
                if ( next_update > buf_size ) 
                {
                    next_update = buf_size;
                }
            }
        }
    }
    // make sure all addresses have been processed
    enqueue_addr( 0, 1 );

    fprintf( stderr, "\b\b\b\b\b\b done !\n" );
}

static void usage(void) {
    fprintf ( stderr, "\nUsage: rtd2lxt [-v] [-e <exefile>] [--start-time <n>]"
              " [--end-time <n>] <dumpfile> [<lxtfile> <savefile>]\n"
              " -h\t\t\t: Display this help\n"
              " -e <filename>\t\t: ELF file corresponding to this <dumpfile>\n"
              " --start-time  <n>\t: dump will be processed from time stamp <n>(usec) \n"
              " --end-time    <n>\t: dump will be processed until time stamp <n>(usec) \n"
              " --thread-state   \t: process thread state informations \n" );
    exit(1);
}


int main(int argc, char **argv) 
{
    int c, err;
    const char *dumpfile;
    char *lxtfile;
    char *savefile;
    char *proffile;
    char *statfile;
    uint32_t step, next_update;
    uint32_t pkt;

    static struct option long_options[] ={
        /* These options set a flag. */
        {"start-time"  , required_argument, 0,  0 },
        {"end-time"    , required_argument, 0,  1 },
        {"thread-state", no_argument      , 0, 's'},
        {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    // first parse options
    while (( c = getopt_long (argc, argv, "e:h:v",
                            long_options, &option_index) ) != -1 )
    {
        switch (c) 
        {
        case 0: /* start-time option */
            if (optarg){
                sscanf( optarg, "%u", &start_time );
                fprintf( stdout, PFX "Buffer start time is %u usec\n", start_time);
                start_time = start_time * 10;
            }else{
                fprintf( stdout, PFX "Buffer start time value is missing\n");
            }
            break;

        case 1: /* end-time option */
            if (optarg){
                sscanf( optarg, "%u", &end_time );
                fprintf( stdout, PFX "Buffer end time is %u usec\n", end_time);
                end_time = end_time * 10;
            }else{
                fprintf( stdout, PFX "Buffer end time value is missing\n");
            }
            break;

        case 's':
            process_thread_states = 1;
            break;

        case 'e':
            exefile = optarg;
            break;

        case 'v':
            verbose = 1;
            break;

        case 'h':
        default:
            usage();
            break;
        }
    }


/*    if (optind < argc)
    {
        printf ("non-option ARGV-elements: ");
        while (optind < argc)
            printf ("%s ", argv[optind++]);

            printf ("\n");
            usage();
     }
*/
    if (( optind != argc-1 )&&( argc != argc-3 )) {
        usage();
    }


    dumpfile = argv[optind];

   if ( optind == argc-3 ) 
   {
      lxtfile = argv[optind+1];
      savefile = argv[optind+2];
      proffile = argv[optind+2];
      statfile = argv[optind+2];
   }
   else 
   {
      // make new names with proper extensions
      lxtfile  = (char *)malloc(strlen(dumpfile)+5);
      savefile = (char *)malloc(strlen(dumpfile)+5);
      proffile = (char *)malloc(strlen(dumpfile)+6);
      statfile = (char *)malloc(strlen(dumpfile)+6);

      assert(lxtfile && savefile && proffile && statfile);
      strcpy( lxtfile , dumpfile ); strcat( lxtfile , ".lxt" );
      strcpy( savefile, dumpfile ); strcat( savefile, ".sav" );
      strcpy( proffile, dumpfile ); strcat( proffile, ".prof" );
      strcpy( statfile, dumpfile ); strcat( statfile, ".html" );
    }

    err = load_dumpfile(dumpfile);

    if ( err == 0 ) 
    {
      // display flush reason
      fprintf( stderr, PFX "flush reason:\n" );
      fprintf( stderr, PFX "\t%s\n", hdr.flush_reason );

      if ( exefile ) 
      {
         // create a new hash table
         int res = hcreate(HASHTAB_SIZE);
         assert(res);
         // and fill it with symbol names
         process_addr_tags();
      }

      // Trace file init
      lt = lt_init(lxtfile);
      init_savefile( );
      prof_file = profile_init(proffile);
      stat_init( );

      assert(lt);

      // set time resolution
      lt_set_timescale(lt, hdr.timescale);
      lt_set_initial_value(lt, 'z');

      if(verbose)
      {
         uint32_t i;

         /*------------------------------------------------------------
            * Print thread tag table
            *------------------------------------------------------------*/
         fprintf( stderr, PFX "Thread tag table: %d threads\n", hdr.nb_thread );
         for(i=0; i<hdr.nb_thread; i++){
               fprintf( stderr, PFX "\tThread %d:'%s'\n", i, thd_tag[i] );
         }

         /*------------------------------------------------------------
            * Print user event tag table
            *------------------------------------------------------------*/
         fprintf( stderr, PFX "User event tag table: %d user events\n", hdr.nb_user_event_tag );
         for(i=0; i<hdr.nb_user_event_tag; i++){
               fprintf( stderr, PFX "\tUser Event %d: '%s'\n", 
                     (int)user_event_tag[i].id, user_event_tag[i].tag);
         }
         /*------------------------------------------------------------
            * Print user u32 tag table
            *------------------------------------------------------------*/
         fprintf( stderr, PFX "User u32 tag table: %d user u32\n", hdr.nb_user_u32_tag );
         for(i=0; i<hdr.nb_user_u32_tag; i++){
               fprintf( stderr, PFX "\tUser Event %d: '%s'\n", 
                     (int)user_u32_tag[i].id, user_u32_tag[i].tag);
         }
      }

      /*------------------------------------------------------------
      * Seek first tick packet and ignore preceding packets
      *------------------------------------------------------------*/
      do {
         pkt = PEEK_PKT();

         if ( RTMON_PKT_TYPE(pkt) != RTMON_PKT_TICK ){
               GET_PKT();
         }else{
         		// update clock
         		timer.ticks = RTMON_PKT_DATA24(pkt);
		   }
      } while ( ( RTMON_PKT_TYPE(pkt) != RTMON_PKT_TICK ) && (pkt != NO_PKT));

      if( pkt == NO_PKT )
      {
      	fprintf( stderr, PFX "There is no OS ticks in this dump, tracing should last longer : ");
		   return err; 
      }


      first_tick = timer.ticks;
      if ( verbose ) 
      {
         fprintf( stderr, PFX "FIRST TICK: %08x\n", (unsigned int)timer.ticks );
      }
      // the buffer will be reprocessed  if the first tick packet
      // is different from 0
      if(timer.ticks != 0)
      {
         timer.ticks --;
         // now that we know which tick we're in, we can process whole buffer
         if ( verbose ) 
         {
            fprintf( stderr, PFX "Re-processing buffer\n");
            fprintf( stderr, PFX "      TICK: %08x\n", (unsigned int)timer.ticks );
         }
         RESET_PKT_INDEX();
         pkt = PEEK_PKT();
      }

      step = buf_size/100;
      if ( !step ) {
         step = 1;
      }
      next_update = 0;

      /*------------------------------------------------------------
      * If start_time has been correctly set ( ie > dump beginning )
      * then time stamp is registered in lxt file.
      *------------------------------------------------------------*/ 
      if(start_time != 0) 
      {
         if( (start_time/10000) < (first_tick*10) ){
               start_time = 0;
               fprintf( stderr, PFX "Invalid start-time value : "
                                    "dump begins at %d us\n",(first_tick*10000));             
         }else{
               lt_set_time64( lt, start_time );
         }
      }
      /*------------------------------------------------------------
      * Checking that end_time param is later than first_tick
      *------------------------------------------------------------*/ 
      if((end_time != 0) &&( (end_time/10000) < (first_tick*10) ) )
      {
         fprintf( stderr, PFX "Invalid end-time value : "
                              "dump begins at %d us\n",(first_tick*10000));             
         end_time = 0;
      }
      
      fprintf( stderr, PFX "processing dump file : [00%%]" );

      // process all remaining packets
      while ( process_pkt() == 0 ) {
         if ( buf_index >= next_update ) {
               fprintf( stderr, "\b\b\b\b\b[%02d%%]",
                        (int)(buf_index*100/buf_size));
               next_update = buf_index+step;
               if ( next_update > buf_size ) 
               {
                  next_update = buf_size;
               }
         }
      }
      fprintf( stderr, "\b\b\b\b\b\b done !\n" );

      /*------------------------------------------------------------
      * If end_time has been correctly set ( ie < dump ending )
      * then time stamp is registered in lxt file.
      -------------------------------------------------------------*/ 
      if(end_time != 0) 
      {
         if( (end_time/10000) > (timer.ticks*10) ){
               end_time = 0;
               fprintf( stderr, PFX "Invalid end-time value : "
                                    "dump stops at %d us\n",(timer.ticks*10000));             
         }else{
               lt_set_time64( lt, end_time );
         }
      }

      // create a savefile for GTKwave with comments, trace ordering, etc.
      fprintf( stderr, PFX "writing GTKWave save file '%s'...\n", savefile );
      write_savefile(savefile);

      fprintf( stderr, PFX "writing LXT file '%s'...\n", lxtfile );
      lt_close(lt);

      fprintf( stderr, PFX "writing statistic file '%s'...\n", statfile );
      write_stat_file(statfile);

      fprintf( stderr, PFX "writing Kcachegrind profiling file '%s'...\n", proffile );
      profile_stop( prof_file);
   }

   close_dumpfile();


   if ( optind != argc-3 ) 
   {
      free(lxtfile );
      free(savefile);
      free(proffile);
      free(statfile);
   }

   return err;
}
