/**
*******************************************************************************
* @file  rtmon_internal.h
* @brief Real-Time Monitor
*
* This package provides a real-time monitoring tool. This file provides the
* internal API used in kernel hooks.
*
* Copyright (C) 2006 Parrot S.A.
*
* @author     ivan.djelic@parrot.fr
* @date       2006-12-01
*******************************************************************************
*/
#ifndef CYGONCE_RTMON_INTERNAL_H
#define CYGONCE_RTMON_INTERNAL_H

#include <pkgconf/rtmon.h>



#undef __externC
#ifdef __cplusplus
// FIXME: we use statement-exprs in C++ code, which is dangerous,
// inlined functions would be better...
# define __externC extern "C"
#else
# define __externC extern
#endif

#include <cyg/rtmon/rtmon.h>
#include <cyg/rtmon/rtmon_common.h>
#include <cyg/rtmon/rtmon_hal.h>


#define RTMON_VERSION         (0x0000000A)  //! RTmon version
#define RTMON_EVENT_TICK      (1 << 0)      //! Internal tick event

//! RTmon buffer size
#define RTMON_BUFSIZE         ((CYGPKG_RTMON_LOGSIZE << 10)/sizeof(cyg_uint32))
//! Maximum number of thread to monitor
#define RTMON_THD_ID_MAX      (CYGPKG_RTMON_NBTHD)

struct rtmon_info {
    cyg_uint32 ticks;
    cyg_uint32 buf_index;   //! This offset is hardcoded in FIQ routine
    cyg_uint32 * const buf; //! This offset is hardcoded in FIQ routine
    cyg_uint32 isr_mask;
    cyg_uint32 dsr_mask;
    cyg_uint32 event_mask;
    cyg_uint32 driver_mask;
    cyg_uint32 dsr_latency[32];
    cyg_uint32 prof_period;
    cyg_uint32 prof_timer_id;
    const char *logfile;
};

extern struct rtmon_info rtmon;

#define RTMON_LOG(_pkt_) {                                                    \
    rtmon.buf[rtmon.buf_index++] = _pkt_;                                     \
    if ( rtmon.buf_index >= RTMON_BUFSIZE ) {                                 \
        rtmon.buf_index = 0;                                                  \
    }                                                                         \
}

//! Log DSR enter
#define RTMON_LOG_DSR_ENTER(_vec)  rtmon_log_dsr_enter(_vec)

#define RTMON_LOG_DSR_EXIT(_vec) {                                            \
    cyg_uint32 _old;                                                          \
    HAL_DISABLE_INTERRUPTS(_old);                                             \
    if ((1 << (_vec)) & rtmon.dsr_mask ) {                                    \
        RTMON_LOG_STAMP( RTMON_PKT_DSR_EXIT, (_vec));                         \
    }                                                                         \
    HAL_RESTORE_INTERRUPTS(_old);                                             \
}

//! Log thread context switch
#define RTMON_LOG_CS(_prio_,_id_) {                                           \
    cyg_uint32 _old;                                                          \
    HAL_DISABLE_INTERRUPTS(_old);                                             \
    if ( rtmon.event_mask & RTMON_EVENT_THREAD_SWITCH )                       \
    {                                                                         \
        RTMON_LOG_STAMP( RTMON_PKT_THREAD, RTMON_THD_CS);                     \
        RTMON_LOG_DATA24( (_prio_<< 16)| _id_ );                              \
    }                                                                         \
    HAL_RESTORE_INTERRUPTS(_old);                                             \
}

//! Log thread state
#define RTMON_LOG_THD_STATE(_state_,_reason_,_id_) {                          \
    cyg_uint32 _old;                                                          \
    HAL_DISABLE_INTERRUPTS(_old);                                             \
    if ( rtmon.event_mask & RTMON_EVENT_THREAD_STATE )                        \
    {                                                                         \
        RTMON_LOG_STAMP( RTMON_PKT_THREAD, RTMON_THD_STATE);                  \
        RTMON_LOG_DATA24( (((_reason_<<5)|_state_)<< 16) | _id_ );            \
    }                                                                         \
    HAL_RESTORE_INTERRUPTS(_old);                                             \
}


//! Log thread state
#define RTMON_LOG_THD_CS2(_nextprio_,_nextid_,_nextstate_,_nextreason_,_oldstate_,_oldreason_) { \
    cyg_uint32 _old;                                                          \
    HAL_DISABLE_INTERRUPTS(_old);                                             \
    if ( rtmon.event_mask & RTMON_EVENT_THREAD_SWITCH )                       \
    {                                                                         \
        RTMON_LOG_STAMP( RTMON_PKT_THREAD, RTMON_THD_CS);                     \
        RTMON_LOG_DATA24( (_nextprio_<< 16)| _nextid_ );                      \
        RTMON_LOG_DATA24( (((_nextreason_<<5)|_nextstate_)<< 8) |             \
                          (((_oldreason_ <<5)|_oldstate_ ))     );            \
    }                                                                         \
    HAL_RESTORE_INTERRUPTS(_old);                                             \
}



//! Log begin of thread priority modification
#define RTMON_LOG_CHGPRIO_ENTER(_prio_) {                                     \
    cyg_uint32 _old;                                                          \
    HAL_DISABLE_INTERRUPTS(_old);                                             \
    if ( rtmon.event_mask & RTMON_EVENT_THREAD_SWITCH )                       \
    {                                                                         \
        RTMON_LOG_STAMP( RTMON_PKT_THREAD, RTMON_THD_CHGPRIO_ENTER);          \
        RTMON_LOG_DATA24( _prio_ );                                           \
    }                                                                         \
    HAL_RESTORE_INTERRUPTS(_old);                                             \
}

//! Log end of thread priority modification
#define RTMON_LOG_CHGPRIO_EXIT(_prio_) {                                      \
    cyg_uint32 _old;                                                          \
    HAL_DISABLE_INTERRUPTS(_old);                                             \
    if ( rtmon.event_mask & RTMON_EVENT_THREAD_SWITCH )                       \
    {                                                                         \
        RTMON_LOG_STAMP( RTMON_PKT_THREAD, RTMON_THD_CHGPRIO_EXIT);           \
        RTMON_LOG_DATA24( _prio_ );                                           \
    }                                                                         \
    HAL_RESTORE_INTERRUPTS(_old);                                             \
}

//! Update RTmon clock
#define RTMON_TICK() {                                                        \
    rtmon.ticks++;                                                            \
    if ( rtmon.event_mask & RTMON_EVENT_TICK ) {                              \
        RTMON_LOG_TICK(rtmon.ticks);                                          \
    }                                                                         \
}

#define RTMON_LOG_DATA24(_val_) \
    RTMON_LOG( (RTMON_PKT_DATA24 << 24)|((_val_) & RTMON_PKT_DATA24_MASK ))

#define RTMON_LOG_TICK(_val_) \
    RTMON_LOG( (RTMON_PKT_TICK   << 24)|((_val_) & RTMON_PKT_DATA24_MASK ))


//! Scheduler locking event
#define RTMON_LOG_SLOCK( _slock_, _func_ ) {                                  \
    cyg_uint32 _old, _f = _func_;                                             \
    HAL_DISABLE_INTERRUPTS(_old);                                             \
    if ( rtmon.event_mask & RTMON_EVENT_SCHED_LOCK ) {                        \
        RTMON_LOG_STAMP((RTMON_PKT_SLOCK_0+(_slock_))                         \
                         & RTMON_PKT_TYPE_MASK, (_f) >> 24 );                 \
        RTMON_LOG_DATA24(_f);                                                 \
    }                                                                         \
    HAL_RESTORE_INTERRUPTS(_old);                                             \
}

//! Log caller function
#define RTMON_LOG_CALLER( ) {                                                 \
    cyg_uint32 _old, _f = RTMON_GET_CALLER();                                 \
    {                                                                         \
        HAL_DISABLE_INTERRUPTS(_old);                                         \
        RTMON_LOG_STAMP((RTMON_PKT_CALLER) & RTMON_PKT_TYPE_MASK,             \
                          (_f) >> 24 );                                       \
        RTMON_LOG_DATA24(_f);                                                 \
        HAL_RESTORE_INTERRUPTS(_old);                                         \
    }                                                                         \
}

//! Log Program Counter
#define RTMON_LOG_PC( ) {                                                     \
    cyg_uint32 _old, _f = RTMON_GET_PC();                                     \
    {                                                                         \
        HAL_DISABLE_INTERRUPTS(_old);                                         \
        RTMON_LOG_STAMP((RTMON_PKT_PC) & RTMON_PKT_TYPE_MASK,                 \
                          (_f) >> 24 );                                       \
        RTMON_LOG_DATA24(_f);                                                 \
        HAL_RESTORE_INTERRUPTS(_old);                                         \
    }                                                                         \
}


//! Log AAI DSR informations
#define RTMON_LOG_AAI_DSR(_its_, _iten_, _cfg_ ) {                            \
    cyg_uint32 _old;                                                          \
    if ( rtmon.driver_mask & RTMON_DRIVER_AAI )                               \
    {                                                                         \
        HAL_DISABLE_INTERRUPTS(_old);                                         \
        RTMON_LOG_STAMP((RTMON_PKT_AAI_REG_ITS) & RTMON_PKT_TYPE_MASK,        \
                          (_its_) >> 24 );                                    \
        RTMON_LOG_DATA24(_its_);                                              \
        RTMON_LOG_STAMP((RTMON_PKT_AAI_REG_ITEN) & RTMON_PKT_TYPE_MASK,       \
                          (_iten_) >> 24 );                                   \
        RTMON_LOG_DATA24(_iten_);                                             \
        RTMON_LOG_STAMP((RTMON_PKT_AAI_REG_CFG) & RTMON_PKT_TYPE_MASK,        \
                          (_cfg_) >> 24 );                                    \
        RTMON_LOG_DATA24(_cfg_);                                              \
        HAL_RESTORE_INTERRUPTS(_old);                                         \
    }                                                                         \
}

//! Log AAI executing code
#define RTMON_LOG_AAI_PC() {                                                  \
    cyg_uint32 _old, _f = RTMON_GET_PC();                                     \
    if ( rtmon.driver_mask & RTMON_DRIVER_AAI )                               \
    {                                                                         \
        HAL_DISABLE_INTERRUPTS(_old);                                         \
        RTMON_LOG_STAMP((RTMON_PKT_AAI_PC) & RTMON_PKT_TYPE_MASK,             \
                          (_f) >> 24 );                                       \
        RTMON_LOG_DATA24(_f);                                                 \
        HAL_RESTORE_INTERRUPTS(_old);                                         \
    }                                                                         \
}



#define RTMON_LOG_CHECK_DSR_COUNT(_vec)  rtmon_log_check_dsr_count(_vec)

__externC void rtmon_log_check_dsr_count(cyg_uint32 vector);
__externC void rtmon_set_thread_tag(cyg_uint32 id,
                                    const char *name,
                                    cyg_uint32 prio  );
__externC void rtmon_log_dsr_enter(cyg_uint32 vector);




// consistency check
#if ( RTMON_IRQ_MAX != RTMON_PKT_IRQ_MAX )
#error Bad RTMON_IRQ_MAX value !
#endif

#endif /* CYGONCE_RTMON_INTERNAL_H */
